@@ -1,6 +1,6 @@
 /*
  * WPA Supplicant / Configuration backend: text file
- * Copyright (c) 2003-2005, Jouni Malinen <jkmaline@cc.hut.fi>
+ * Copyright (c) 2003-2006, Jouni Malinen <jkmaline@cc.hut.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
@@ -16,18 +16,18 @@
  * described in the sample configuration file, wpa_supplicant.conf.
  */
 
-#include <stdlib.h>
-#include <stdio.h>
-#include <string.h>
+#include "includes.h"
 
 #include "common.h"
 #include "wpa.h"
 #include "wpa_supplicant.h"
 #include "config.h"
 #include "base64.h"
+#include "eap_methods.h"
 
 
-static char * wpa_config_get_line(char *s, int size, FILE *stream, int *line)
+static char * wpa_config_get_line(char *s, int size, FILE *stream, int *line,
+				  char **_pos)
 {
 	char *pos, *end, *sstart;
 
@@ -62,9 +62,13 @@ static char * wpa_config_get_line(char *s, int size, FILE *stream, int *line)
 		if (*pos == '\0')
 			continue;
 
+		if (_pos)
+			*_pos = pos;
 		return pos;
 	}
 
+	if (_pos)
+		*_pos = NULL;
 	return NULL;
 }
 
@@ -77,15 +81,14 @@ static struct wpa_ssid * wpa_config_read_network(FILE *f, int *line, int id)
 
 	wpa_printf(MSG_MSGDUMP, "Line: %d - start of a new network block",
 		   *line);
-	ssid = (struct wpa_ssid *) malloc(sizeof(*ssid));
+	ssid = wpa_zalloc(sizeof(*ssid));
 	if (ssid == NULL)
 		return NULL;
-	memset(ssid, 0, sizeof(*ssid));
 	ssid->id = id;
 
 	wpa_config_set_network_defaults(ssid);
 
-	while ((pos = wpa_config_get_line(buf, sizeof(buf), f, line))) {
+	while (wpa_config_get_line(buf, sizeof(buf), f, line, &pos)) {
 		if (strcmp(pos, "}") == 0) {
 			end = 1;
 			break;
@@ -135,7 +138,8 @@ static struct wpa_ssid * wpa_config_read_network(FILE *f, int *line, int id)
 	}
 
 	if ((ssid->group_cipher & WPA_CIPHER_CCMP) &&
-	    !(ssid->pairwise_cipher & WPA_CIPHER_CCMP)) {
+	    !(ssid->pairwise_cipher & WPA_CIPHER_CCMP) &&
+	    !(ssid->pairwise_cipher & WPA_CIPHER_NONE)) {
 		/* Group cipher cannot be stronger than the pairwise cipher. */
 		wpa_printf(MSG_DEBUG, "Line %d: removed CCMP from group cipher"
 			   " list since it was not allowed for pairwise "
@@ -144,7 +148,7 @@ static struct wpa_ssid * wpa_config_read_network(FILE *f, int *line, int id)
 	}
 
 	if (errors) {
-		free(ssid);
+		wpa_config_free_ssid(ssid);
 		ssid = NULL;
 	}
 
@@ -164,7 +168,7 @@ static struct wpa_config_blob * wpa_config_read_blob(FILE *f, int *line,
 	wpa_printf(MSG_MSGDUMP, "Line: %d - start of a new named blob '%s'",
 		   *line, name);
 
-	while ((pos = wpa_config_get_line(buf, sizeof(buf), f, line))) {
+	while (wpa_config_get_line(buf, sizeof(buf), f, line, &pos)) {
 		if (strcmp(pos, "}") == 0) {
 			end = 1;
 			break;
@@ -190,12 +194,11 @@ static struct wpa_config_blob * wpa_config_read_blob(FILE *f, int *line,
 		return NULL;
 	}
 
-	blob = malloc(sizeof(*blob));
+	blob = wpa_zalloc(sizeof(*blob));
 	if (blob == NULL) {
 		free(encoded);
 		return NULL;
 	}
-	memset(blob, 0, sizeof(*blob));
 	blob->name = strdup(name);
 	blob->data = base64_decode(encoded, encoded_len, &blob->len);
 	free(encoded);
@@ -228,7 +231,7 @@ struct wpa_config * wpa_config_read(const char *name)
 		return NULL;
 	}
 
-	while ((pos = wpa_config_get_line(buf, sizeof(buf), f, &line))) {
+	while (wpa_config_get_line(buf, sizeof(buf), f, &line, &pos)) {
 		if (strcmp(pos, "network={") == 0) {
 			ssid = wpa_config_read_network(f, &line, id++);
 			if (ssid == NULL) {
@@ -368,6 +371,20 @@ struct wpa_config * wpa_config_read(const char *name)
 			config->update_config = atoi(pos + 14);
 			wpa_printf(MSG_DEBUG, "update_config=%d",
 				   config->update_config);
+		} else if (strncmp(pos, "load_dynamic_eap=", 17) == 0) {
+			char *so = pos + 17;
+			int ret;
+			wpa_printf(MSG_DEBUG, "load_dynamic_eap=%s", so);
+			ret = eap_peer_method_load(so);
+			if (ret == -2) {
+				wpa_printf(MSG_DEBUG, "This EAP type was "
+					   "already loaded - not reloading.");
+			} else if (ret) {
+				wpa_printf(MSG_ERROR, "Line %d: Failed to "
+					   "load dynamic EAP method '%s'.",
+					   line, so);
+				errors++;
+			}
 		} else {
 			wpa_printf(MSG_ERROR, "Line %d: Invalid configuration "
 				   "line '%s'.", line, pos);
@@ -518,6 +535,7 @@ static void write_auth_alg(FILE *f, struct wpa_ssid *ssid)
 }
 
 
+#ifdef IEEE8021X_EAPOL
 static void write_eap(FILE *f, struct wpa_ssid *ssid)
 {
 	char *value;
@@ -530,6 +548,7 @@ static void write_eap(FILE *f, struct wpa_ssid *ssid)
 		fprintf(f, "\teap=%s\n", value);
 	free(value);
 }
+#endif /* IEEE8021X_EAPOL */
 
 
 static void write_wep_key(FILE *f, int idx, struct wpa_ssid *ssid)
@@ -562,6 +581,7 @@ static void wpa_config_write_network(FILE *f, struct wpa_ssid *ssid)
 	write_pairwise(f, ssid);
 	write_group(f, ssid);
 	write_auth_alg(f, ssid);
+#ifdef IEEE8021X_EAPOL
 	write_eap(f, ssid);
 	STR(identity);
 	STR(anonymous_identity);
@@ -590,12 +610,16 @@ static void wpa_config_write_network(FILE *f, struct wpa_ssid *ssid)
 	STR(key_id);
 	INT(engine);
 	INT_DEF(eapol_flags, DEFAULT_EAPOL_FLAGS);
+#endif /* IEEE8021X_EAPOL */
 	for (i = 0; i < 4; i++)
 		write_wep_key(f, i, ssid);
 	INT(wep_tx_keyidx);
 	INT(priority);
+#ifdef IEEE8021X_EAPOL
 	INT_DEF(eap_workaround, DEFAULT_EAP_WORKAROUND);
 	STR(pac_file);
+	INT_DEF(fragment_size, DEFAULT_FRAGMENT_SIZE);
+#endif /* IEEE8021X_EAPOL */
 	INT(mode);
 	INT(proactive_key_caching);
 	INT(disabled);
