@@ -1,6 +1,6 @@
 /*
  * WPA Supplicant - testing driver interface
- * Copyright (c) 2004-2005, Jouni Malinen <jkmaline@cc.hut.fi>
+ * Copyright (c) 2004-2006, Jouni Malinen <jkmaline@cc.hut.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
@@ -12,12 +12,10 @@
  * See README and COPYING for more details.
  */
 
-#include <stdlib.h>
-#include <stdio.h>
-#include <string.h>
-#include <unistd.h>
-#include <sys/socket.h>
+#include "includes.h"
 #include <sys/un.h>
+#include <dirent.h>
+#include <sys/stat.h>
 
 #include "common.h"
 #include "driver.h"
@@ -26,6 +24,7 @@
 #include "eloop.h"
 #include "sha1.h"
 #include "wpa.h"
+#include "mlme.h"
 
 
 struct wpa_driver_test_data {
@@ -33,15 +32,19 @@ struct wpa_driver_test_data {
 	u8 own_addr[ETH_ALEN];
 	int test_socket;
 	struct sockaddr_un hostapd_addr;
+	int hostapd_addr_set;
 	char *own_socket_path;
+	char *test_dir;
 	u8 bssid[ETH_ALEN];
 	u8 ssid[32];
 	size_t ssid_len;
-	struct wpa_scan_result scanres;
+#define MAX_SCAN_RESULTS 30
+	struct wpa_scan_result scanres[MAX_SCAN_RESULTS];
 	size_t num_scanres;
 	int use_associnfo;
 	u8 assoc_wpa_ie[80];
 	size_t assoc_wpa_ie_len;
+	int use_mlme;
 };
 
 
@@ -59,22 +62,55 @@ static void wpa_driver_test_scan_timeout(void *eloop_ctx, void *timeout_ctx)
 }
 
 
+static void wpa_driver_scan_dir(struct wpa_driver_test_data *drv,
+				const char *path)
+{
+	struct dirent *dent;
+	DIR *dir;
+	struct sockaddr_un addr;
+
+	dir = opendir(path);
+	if (dir == NULL)
+		return;
+
+	while ((dent = readdir(dir))) {
+		if (strncmp(dent->d_name, "AP-", 3) != 0)
+			continue;
+		wpa_printf(MSG_DEBUG, "%s: SCAN %s", __func__, dent->d_name);
+
+		memset(&addr, 0, sizeof(addr));
+		addr.sun_family = AF_UNIX;
+		snprintf(addr.sun_path, sizeof(addr.sun_path), "%s/%s",
+			 path, dent->d_name);
+
+		if (sendto(drv->test_socket, "SCAN", 4, 0,
+			   (struct sockaddr *) &addr, sizeof(addr)) < 0) {
+			perror("sendto(test_socket)");
+		}
+	}
+	closedir(dir);
+}
+
+
 static int wpa_driver_test_scan(void *priv, const u8 *ssid, size_t ssid_len)
 {
 	struct wpa_driver_test_data *drv = priv;
 	wpa_printf(MSG_DEBUG, "%s: priv=%p", __func__, priv);
 
-	if (drv->test_socket >= 0) {
-		if (sendto(drv->test_socket, "SCAN", 4, 0,
-			   (struct sockaddr *) &drv->hostapd_addr,
-			   sizeof(drv->hostapd_addr)) < 0) {
-			perror("sendto(test_socket)");
-			return -1;
-		}
-	} else {
-		eloop_register_timeout(1, 0, wpa_driver_test_scan_timeout, drv,
-				       drv->ctx);
+	drv->num_scanres = 0;
+
+	if (drv->test_socket >= 0 && drv->test_dir)
+		wpa_driver_scan_dir(drv, drv->test_dir);
+
+	if (drv->test_socket >= 0 && drv->hostapd_addr_set &&
+	    sendto(drv->test_socket, "SCAN", 4, 0,
+		   (struct sockaddr *) &drv->hostapd_addr,
+		   sizeof(drv->hostapd_addr)) < 0) {
+		perror("sendto(test_socket)");
 	}
+
+	eloop_register_timeout(1, 0, wpa_driver_test_scan_timeout, drv,
+			       drv->ctx);
 	return 0;
 }
 
@@ -84,8 +120,11 @@ static int wpa_driver_test_get_scan_results(void *priv,
 					    size_t max_size)
 {
 	struct wpa_driver_test_data *drv = priv;
-	memcpy(results, &drv->scanres, sizeof(drv->scanres));
-	return drv->num_scanres;
+	size_t num = drv->num_scanres;
+	if (num > max_size)
+		num = max_size;
+	memcpy(results, &drv->scanres, num * sizeof(struct wpa_scan_result));
+	return num;
 }
 
 
@@ -137,22 +176,26 @@ static int wpa_driver_test_associate(
 	} else
 		drv->assoc_wpa_ie_len = 0;
 
-	if (drv->test_socket >= 0) {
+	if (drv->test_dir && params->bssid) {
+		memset(&drv->hostapd_addr, 0, sizeof(drv->hostapd_addr));
+		drv->hostapd_addr.sun_family = AF_UNIX;
+		snprintf(drv->hostapd_addr.sun_path,
+			 sizeof(drv->hostapd_addr.sun_path), "%s/AP-" MACSTR,
+			 drv->test_dir, MAC2STR(params->bssid));
+		drv->hostapd_addr_set = 1;
+	}
+
+	if (drv->test_socket >= 0 && drv->hostapd_addr_set) {
 		char cmd[200], *pos, *end;
-		int i;
 		end = cmd + sizeof(cmd);
 		pos = cmd;
 		pos += snprintf(pos, end - pos, "ASSOC " MACSTR " ",
 				MAC2STR(drv->own_addr));
-		for (i = 0; i < params->ssid_len; i++) {
-			pos += snprintf(pos, end - pos, "%02x",
-					params->ssid[i]);
-		}
+		pos += wpa_snprintf_hex(pos, end - pos, params->ssid,
+					params->ssid_len);
 		pos += snprintf(pos, end - pos, " ");
-		for (i = 0; i < params->wpa_ie_len; i++) {
-			pos += snprintf(pos, end - pos, "%02x",
-					params->wpa_ie[i]);
-		}
+		pos += wpa_snprintf_hex(pos, end - pos, params->wpa_ie,
+					params->wpa_ie_len);
 		if (sendto(drv->test_socket, cmd, strlen(cmd), 0,
 			   (struct sockaddr *) &drv->hostapd_addr,
 			   sizeof(drv->hostapd_addr)) < 0) {
@@ -230,12 +273,17 @@ static void wpa_driver_test_scanresp(struct wpa_driver_test_data *drv,
 	struct wpa_scan_result *res;
 	const char *pos, *pos2;
 	size_t len;
-	u8 ie[200];
+	u8 ie[200], *ipos, *end;
 
 	wpa_printf(MSG_DEBUG, "test_driver: SCANRESP %s", data);
+	if (drv->num_scanres >= MAX_SCAN_RESULTS) {
+		wpa_printf(MSG_DEBUG, "test_driver: No room for the new scan "
+			   "result");
+		return;
+	}
+
 	/* SCANRESP BSSID SSID IEs */
-	res = &drv->scanres;
-	drv->num_scanres = 0;
+	res = &drv->scanres[drv->num_scanres];
 
 	memset(res, 0, sizeof(*res));
 	if (hwaddr_aton(data, res->bssid)) {
@@ -276,18 +324,24 @@ static void wpa_driver_test_scanresp(struct wpa_driver_test_data *drv,
 		return;
 	}
 
-	if (len > SSID_MAX_WPA_IE_LEN)
-		len = SSID_MAX_WPA_IE_LEN;
-	if (len > 0 && ie[0] == RSN_INFO_ELEM) {
-		memcpy(res->rsn_ie, ie, len);
-		res->rsn_ie_len = len;
-	} else {
-		memcpy(res->wpa_ie, ie, len);
-		res->wpa_ie_len = len;
+	ipos = ie;
+	end = ipos + len;
+	while (ipos + 1 < end && ipos + 2 + ipos[1] <= end) {
+		len = 2 + ipos[1];
+		if (len > SSID_MAX_WPA_IE_LEN)
+			len = SSID_MAX_WPA_IE_LEN;
+		if (ipos[0] == RSN_INFO_ELEM) {
+			memcpy(res->rsn_ie, ipos, len);
+			res->rsn_ie_len = len;
+		} else if (ipos[0] == GENERIC_INFO_ELEM) {
+			memcpy(res->wpa_ie, ipos, len);
+			res->wpa_ie_len = len;
+		}
+
+		ipos += 2 + ipos[1];
 	}
 
-	drv->num_scanres = 1;
-	wpa_supplicant_event(drv->ctx, EVENT_SCAN_RESULTS, NULL);
+	drv->num_scanres++;
 }
 
 
@@ -325,12 +379,26 @@ static void wpa_driver_test_eapol(struct wpa_driver_test_data *drv,
 				  socklen_t fromlen,
 				  const u8 *data, size_t data_len)
 {
+	const u8 *src = drv->bssid;
+
 	if (data_len > 14) {
 		/* Skip Ethernet header */
+		src = data + ETH_ALEN;
 		data += 14;
 		data_len -= 14;
 	}
-	wpa_supplicant_rx_eapol(drv->ctx, drv->bssid, (u8 *) data, data_len);
+	wpa_supplicant_rx_eapol(drv->ctx, src, data, data_len);
+}
+
+
+static void wpa_driver_test_mlme(struct wpa_driver_test_data *drv,
+				 struct sockaddr_un *from,
+				 socklen_t fromlen,
+				 const u8 *data, size_t data_len)
+{
+	struct ieee80211_rx_status rx_status;
+	memset(&rx_status, 0, sizeof(rx_status));
+	ieee80211_sta_rx(drv->ctx, data, data_len, &rx_status);
 }
 
 
@@ -367,7 +435,11 @@ static void wpa_driver_test_receive_unix(int sock, void *eloop_ctx,
 	} else if (strcmp(buf, "DEAUTH") == 0) {
 		wpa_driver_test_disassoc(drv, &from, fromlen);
 	} else if (strncmp(buf, "EAPOL ", 6) == 0) {
-		wpa_driver_test_eapol(drv, &from, fromlen, buf + 6, res - 6);
+		wpa_driver_test_eapol(drv, &from, fromlen,
+				      (const u8 *) buf + 6, res - 6);
+	} else if (strncmp(buf, "MLME ", 5) == 0) {
+		wpa_driver_test_mlme(drv, &from, fromlen,
+				     (const u8 *) buf + 5, res - 5);
 	} else {
 		wpa_hexdump_ascii(MSG_DEBUG, "Unknown test_socket command",
 				  (u8 *) buf, res);
@@ -380,10 +452,9 @@ static void * wpa_driver_test_init(void *ctx, const char *ifname)
 {
 	struct wpa_driver_test_data *drv;
 
-	drv = malloc(sizeof(*drv));
+	drv = wpa_zalloc(sizeof(*drv));
 	if (drv == NULL)
 		return NULL;
-	memset(drv, 0, sizeof(*drv));
 	drv->ctx = ctx;
 	drv->test_socket = -1;
 
@@ -399,7 +470,7 @@ static void * wpa_driver_test_init(void *ctx, const char *ifname)
 
 	/* Generate a MAC address to help testing with multiple STAs */
 	drv->own_addr[0] = 0x02; /* locally administered */
-	sha1_prf(ifname, strlen(ifname),
+	sha1_prf((const u8 *) ifname, strlen(ifname),
 		 "wpa_supplicant test mac addr generation",
 		 NULL, 0, drv->own_addr + 1, ETH_ALEN - 1);
 
@@ -427,20 +498,34 @@ static void wpa_driver_test_deinit(void *priv)
 {
 	struct wpa_driver_test_data *drv = priv;
 	wpa_driver_test_close_test_socket(drv);
+	free(drv->test_dir);
 	free(drv);
 }
 
 
-static int wpa_driver_test_attach(struct wpa_driver_test_data *drv)
+static int wpa_driver_test_attach(struct wpa_driver_test_data *drv,
+				  const char *dir)
 {
 	static unsigned int counter = 0;
 	struct sockaddr_un addr;
+	size_t len;
 
-	drv->own_socket_path = malloc(100);
-	if (drv->own_socket_path == NULL)
-		return -1;
-	snprintf(drv->own_socket_path, 100, "/tmp/wpa_supplicant_test-%d-%d",
-		 getpid(), counter++);
+	free(drv->own_socket_path);
+	if (dir) {
+		len = strlen(dir) + 30;
+		drv->own_socket_path = malloc(len);
+		if (drv->own_socket_path == NULL)
+			return -1;
+		snprintf(drv->own_socket_path, len, "%s/STA-" MACSTR,
+			 dir, MAC2STR(drv->own_addr));
+	} else {
+		drv->own_socket_path = malloc(100);
+		if (drv->own_socket_path == NULL)
+			return -1;
+		snprintf(drv->own_socket_path, 100,
+			 "/tmp/wpa_supplicant_test-%d-%d",
+			 getpid(), counter++);
+	}
 
 	drv->test_socket = socket(PF_UNIX, SOCK_DGRAM, 0);
 	if (drv->test_socket < 0) {
@@ -494,14 +579,35 @@ static int wpa_driver_test_set_param(void *priv, const char *param)
 		memset(&drv->hostapd_addr, 0, sizeof(drv->hostapd_addr));
 		drv->hostapd_addr.sun_family = AF_UNIX;
 		memcpy(drv->hostapd_addr.sun_path, pos, len);
-		wpa_driver_test_attach(drv);
+		drv->hostapd_addr_set = 1;
 	}
 
+	pos = strstr(param, "test_dir=");
+	if (pos) {
+		char *end;
+		free(drv->test_dir);
+		drv->test_dir = strdup(pos + 9);
+		if (drv->test_dir == NULL)
+			return -1;
+		end = strchr(drv->test_dir, ' ');
+		if (end)
+			*end = '\0';
+		wpa_driver_test_attach(drv, drv->test_dir);
+	} else
+		wpa_driver_test_attach(drv, NULL);
+
 	if (strstr(param, "use_associnfo=1")) {
 		wpa_printf(MSG_DEBUG, "test_driver: Use AssocInfo events");
 		drv->use_associnfo = 1;
 	}
 
+#ifdef CONFIG_CLIENT_MLME
+	if (strstr(param, "use_mlme=1")) {
+		wpa_printf(MSG_DEBUG, "test_driver: Use internal MLME");
+		drv->use_mlme = 1;
+	}
+#endif /* CONFIG_CLIENT_MLME */
+
 	return 0;
 }
 
@@ -521,6 +627,7 @@ static int wpa_driver_test_send_eapol(void *priv, const u8 *dest, u16 proto,
 	struct msghdr msg;
 	struct iovec io[3];
 	struct l2_ethhdr eth;
+	struct sockaddr_un addr;
 
 	wpa_hexdump(MSG_MSGDUMP, "test_send_eapol TX frame", data, data_len);
 
@@ -539,8 +646,24 @@ static int wpa_driver_test_send_eapol(void *priv, const u8 *dest, u16 proto,
 	memset(&msg, 0, sizeof(msg));
 	msg.msg_iov = io;
 	msg.msg_iovlen = 3;
-	msg.msg_name = &drv->hostapd_addr;
-	msg.msg_namelen = sizeof(drv->hostapd_addr);
+	if (memcmp(dest, drv->bssid, ETH_ALEN) == 0 || drv->test_dir == NULL) {
+		msg.msg_name = &drv->hostapd_addr;
+		msg.msg_namelen = sizeof(drv->hostapd_addr);
+	} else {
+		struct stat st;
+		memset(&addr, 0, sizeof(addr));
+		addr.sun_family = AF_UNIX;
+		snprintf(addr.sun_path, sizeof(addr.sun_path),
+			 "%s/STA-" MACSTR, drv->test_dir, MAC2STR(dest));
+		if (stat(addr.sun_path, &st) < 0) {
+			snprintf(addr.sun_path, sizeof(addr.sun_path),
+				 "%s/AP-" MACSTR,
+				 drv->test_dir, MAC2STR(dest));
+		}
+		msg.msg_name = &addr;
+		msg.msg_namelen = sizeof(addr);
+	}
+
 	if (sendmsg(drv->test_socket, &msg, 0) < 0) {
 		perror("sendmsg(test_socket)");
 		return -1;
@@ -550,21 +673,242 @@ static int wpa_driver_test_send_eapol(void *priv, const u8 *dest, u16 proto,
 }
 
 
+static int wpa_driver_test_get_capa(void *priv, struct wpa_driver_capa *capa)
+{
+	struct wpa_driver_test_data *drv = priv;
+	memset(capa, 0, sizeof(*capa));
+	capa->key_mgmt = WPA_DRIVER_CAPA_KEY_MGMT_WPA |
+		WPA_DRIVER_CAPA_KEY_MGMT_WPA2 |
+		WPA_DRIVER_CAPA_KEY_MGMT_WPA_PSK |
+		WPA_DRIVER_CAPA_KEY_MGMT_WPA2_PSK |
+		WPA_DRIVER_CAPA_KEY_MGMT_WPA_NONE;
+	capa->enc = WPA_DRIVER_CAPA_ENC_WEP40 |
+		WPA_DRIVER_CAPA_ENC_WEP104 |
+		WPA_DRIVER_CAPA_ENC_TKIP |
+		WPA_DRIVER_CAPA_ENC_CCMP;
+	capa->auth = WPA_DRIVER_AUTH_OPEN |
+		WPA_DRIVER_AUTH_SHARED |
+		WPA_DRIVER_AUTH_LEAP;
+	if (drv->use_mlme)
+		capa->flags |= WPA_DRIVER_FLAGS_USER_SPACE_MLME;
+
+	return 0;
+}
+
+
+static int wpa_driver_test_mlme_setprotection(void *priv, const u8 *addr,
+					      int protect_type,
+					      int key_type)
+{
+	wpa_printf(MSG_DEBUG, "%s: protect_type=%d key_type=%d",
+		   __func__, protect_type, key_type);
+
+	if (addr) {
+		wpa_printf(MSG_DEBUG, "%s: addr=" MACSTR,
+			   __func__, MAC2STR(addr));
+	}
+
+	return 0;
+}
+
+
+#ifdef CONFIG_CLIENT_MLME
+static struct wpa_hw_modes *
+wpa_driver_test_get_hw_feature_data(void *priv, u16 *num_modes, u16 *flags)
+{
+	struct wpa_hw_modes *modes;
+
+	*num_modes = 1;
+	*flags = 0;
+	modes = wpa_zalloc(*num_modes * sizeof(struct wpa_hw_modes));
+	if (modes == NULL)
+		return NULL;
+	modes[0].mode = WPA_MODE_IEEE80211G;
+	modes[0].num_channels = 1;
+	modes[0].num_rates = 1;
+	modes[0].channels = wpa_zalloc(sizeof(struct wpa_channel_data));
+	modes[0].rates = wpa_zalloc(sizeof(struct wpa_rate_data));
+	if (modes[0].channels == NULL || modes[0].rates == NULL) {
+		ieee80211_sta_free_hw_features(modes, *num_modes);
+		return NULL;
+	}
+	modes[0].channels[0].chan = 1;
+	modes[0].channels[0].freq = 2412;
+	modes[0].channels[0].flag = WPA_CHAN_W_SCAN | WPA_CHAN_W_ACTIVE_SCAN;
+	modes[0].rates[0].rate = 10;
+	modes[0].rates[0].flags = WPA_RATE_BASIC | WPA_RATE_SUPPORTED |
+		WPA_RATE_CCK | WPA_RATE_MANDATORY;
+
+	return modes;
+}
+
+
+int wpa_driver_test_set_channel(void *priv, wpa_hw_mode phymode, int chan,
+				int freq)
+{
+	wpa_printf(MSG_DEBUG, "%s: phymode=%d chan=%d freq=%d",
+		   __func__, phymode, chan, freq);
+	return 0;
+}
+
+
+static int wpa_driver_test_send_mlme(void *priv, const u8 *data,
+				     size_t data_len)
+{
+	struct wpa_driver_test_data *drv = priv;
+	struct msghdr msg;
+	struct iovec io[2];
+	struct sockaddr_un addr;
+	const u8 *dest;
+	struct dirent *dent;
+	DIR *dir;
+
+	wpa_hexdump(MSG_MSGDUMP, "test_send_mlme", data, data_len);
+	if (data_len < 10)
+		return -1;
+	dest = data + 4;
+
+	io[0].iov_base = "MLME ";
+	io[0].iov_len = 5;
+	io[1].iov_base = (u8 *) data;
+	io[1].iov_len = data_len;
+
+	memset(&msg, 0, sizeof(msg));
+	msg.msg_iov = io;
+	msg.msg_iovlen = 2;
+	if (memcmp(dest, drv->bssid, ETH_ALEN) == 0 || drv->test_dir == NULL) {
+		msg.msg_name = &drv->hostapd_addr;
+		msg.msg_namelen = sizeof(drv->hostapd_addr);
+	} else if (memcmp(dest, "\xff\xff\xff\xff\xff\xff", ETH_ALEN) == 0) {
+		dir = opendir(drv->test_dir);
+		if (dir == NULL)
+			return -1;
+		while ((dent = readdir(dir))) {
+#ifdef _DIRENT_HAVE_D_TYPE
+			/* Skip the file if it is not a socket.
+			 * Also accept DT_UNKNOWN (0) in case
+			 * the C library or underlying file
+			 * system does not support d_type. */
+			if (dent->d_type != DT_SOCK &&
+			    dent->d_type != DT_UNKNOWN)
+				continue;
+#endif /* _DIRENT_HAVE_D_TYPE */
+			if (strcmp(dent->d_name, ".") == 0 ||
+			    strcmp(dent->d_name, "..") == 0)
+				continue;
+			wpa_printf(MSG_DEBUG, "%s: Send broadcast MLME to %s",
+				   __func__, dent->d_name);
+			memset(&addr, 0, sizeof(addr));
+			addr.sun_family = AF_UNIX;
+			snprintf(addr.sun_path, sizeof(addr.sun_path), "%s/%s",
+				 drv->test_dir, dent->d_name);
+
+			msg.msg_name = &addr;
+			msg.msg_namelen = sizeof(addr);
+
+			if (sendmsg(drv->test_socket, &msg, 0) < 0)
+				perror("sendmsg(test_socket)");
+		}
+		closedir(dir);
+		return 0;
+	} else {
+		struct stat st;
+		memset(&addr, 0, sizeof(addr));
+		addr.sun_family = AF_UNIX;
+		snprintf(addr.sun_path, sizeof(addr.sun_path),
+			 "%s/AP-" MACSTR, drv->test_dir, MAC2STR(dest));
+		if (stat(addr.sun_path, &st) < 0) {
+			snprintf(addr.sun_path, sizeof(addr.sun_path),
+				 "%s/STA-" MACSTR,
+				 drv->test_dir, MAC2STR(dest));
+		}
+		msg.msg_name = &addr;
+		msg.msg_namelen = sizeof(addr);
+	}
+
+	if (sendmsg(drv->test_socket, &msg, 0) < 0) {
+		perror("sendmsg(test_socket)");
+		return -1;
+	}
+
+	return 0;
+}
+
+
+static int wpa_driver_test_mlme_add_sta(void *priv, const u8 *addr,
+					const u8 *supp_rates,
+					size_t supp_rates_len)
+{
+	wpa_printf(MSG_DEBUG, "%s: addr=" MACSTR, __func__, MAC2STR(addr));
+	return 0;
+}
+
+
+static int wpa_driver_test_mlme_remove_sta(void *priv, const u8 *addr)
+{
+	wpa_printf(MSG_DEBUG, "%s: addr=" MACSTR, __func__, MAC2STR(addr));
+	return 0;
+}
+
+
+int wpa_driver_test_set_ssid(void *priv, const u8 *ssid, size_t ssid_len)
+{
+	wpa_printf(MSG_DEBUG, "%s", __func__);
+	return 0;
+}
+
+
+int wpa_driver_test_set_bssid(void *priv, const u8 *bssid)
+{
+	wpa_printf(MSG_DEBUG, "%s: bssid=" MACSTR, __func__, MAC2STR(bssid));
+	return 0;
+}
+#endif /* CONFIG_CLIENT_MLME */
+
+
 const struct wpa_driver_ops wpa_driver_test_ops = {
-	.name = "test",
-	.desc = "wpa_supplicant test driver",
-	.init = wpa_driver_test_init,
-	.deinit = wpa_driver_test_deinit,
-	.set_param = wpa_driver_test_set_param,
-	.set_wpa = wpa_driver_test_set_wpa,
-	.scan = wpa_driver_test_scan,
-	.get_scan_results = wpa_driver_test_get_scan_results,
-	.set_key = wpa_driver_test_set_key,
-	.associate = wpa_driver_test_associate,
-	.get_bssid = wpa_driver_test_get_bssid,
-	.get_ssid = wpa_driver_test_get_ssid,
-	.get_mac_addr = wpa_driver_test_get_mac_addr,
-	.send_eapol = wpa_driver_test_send_eapol,
-	.deauthenticate = wpa_driver_test_deauthenticate,
-	.disassociate = wpa_driver_test_disassociate,
+	"test",
+	"wpa_supplicant test driver",
+	wpa_driver_test_get_bssid,
+	wpa_driver_test_get_ssid,
+	wpa_driver_test_set_wpa,
+	wpa_driver_test_set_key,
+	wpa_driver_test_init,
+	wpa_driver_test_deinit,
+	wpa_driver_test_set_param,
+	NULL /* set_countermeasures */,
+	NULL /* set_drop_unencrypted */,
+	wpa_driver_test_scan,
+	wpa_driver_test_get_scan_results,
+	wpa_driver_test_deauthenticate,
+	wpa_driver_test_disassociate,
+	wpa_driver_test_associate,
+	NULL /* set_auth_alg */,
+	NULL /* add_pmkid */,
+	NULL /* remove_pmkid */,
+	NULL /* flush_pmkid */,
+	wpa_driver_test_get_capa,
+	NULL /* poll */,
+	NULL /* get_ifname */,
+	wpa_driver_test_get_mac_addr,
+	wpa_driver_test_send_eapol,
+	NULL /* set_operstate */,
+	wpa_driver_test_mlme_setprotection,
+#ifdef CONFIG_CLIENT_MLME
+	wpa_driver_test_get_hw_feature_data,
+	wpa_driver_test_set_channel,
+	wpa_driver_test_set_ssid,
+	wpa_driver_test_set_bssid,
+	wpa_driver_test_send_mlme,
+	wpa_driver_test_mlme_add_sta,
+	wpa_driver_test_mlme_remove_sta
+#else /* CONFIG_CLIENT_MLME */
+	NULL /* get_hw_feature_data */,
+	NULL /* set_channel */,
+	NULL /* set_ssid */,
+	NULL /* set_bssid */,
+	NULL /* send_mlme */,
+	NULL /* mlme_add_sta */,
+	NULL /* mlme_remove_sta */
+#endif /* CONFIG_CLIENT_MLME */
 };
