@@ -18,12 +18,8 @@
  * ioctls, but require private ioctls for some of the functionality.
  */
 
-#include <stdlib.h>
-#include <stdio.h>
-#include <unistd.h>
-#include <string.h>
+#include "includes.h"
 #include <sys/ioctl.h>
-#include <errno.h>
 #include <net/if_arp.h>
 
 #include "wireless_copy.h"
@@ -36,11 +32,24 @@
 #include "driver_wext.h"
 #include "wpa.h"
 
+#ifdef CONFIG_CLIENT_MLME
+#include <netpacket/packet.h>
+#include <hostapd_ioctl.h>
+#include <net/d80211_common.h>
+#include <net/d80211_shared.h>
+#include "mlme.h"
+
+#ifndef ETH_P_ALL
+#define ETH_P_ALL 0x0003
+#endif
+#endif /* CONFIG_CLIENT_MLME */
+
 
 struct wpa_driver_wext_data {
 	void *ctx;
 	int event_sock;
 	int ioctl_sock;
+	int mlme_sock;
 	char ifname[IFNAMSIZ + 1];
 	int ifindex;
 	int ifindex2;
@@ -51,12 +60,76 @@ struct wpa_driver_wext_data {
 	struct wpa_driver_capa capa;
 	int has_capability;
 	int we_version_compiled;
+
+	/* for set_auth_alg fallback */
+	int use_crypt;
+	int auth_alg_fallback;
+
+	int operstate;
+
+	char mlmedev[IFNAMSIZ + 1];
 };
 
 
 static int wpa_driver_wext_flush_pmkid(void *priv);
 static int wpa_driver_wext_get_range(void *priv);
 
+static int wpa_driver_wext_send_oper_ifla(struct wpa_driver_wext_data *drv,
+					  int linkmode, int operstate)
+{
+	struct {
+		struct nlmsghdr hdr;
+		struct ifinfomsg ifinfo;
+		char opts[16];
+	} req;
+	struct rtattr *rta;
+	static int nl_seq;
+	ssize_t ret;
+
+	req.hdr.nlmsg_len = NLMSG_LENGTH(sizeof(struct ifinfomsg));
+	req.hdr.nlmsg_type = RTM_SETLINK;
+	req.hdr.nlmsg_flags = NLM_F_REQUEST;
+	req.hdr.nlmsg_seq = ++nl_seq;
+	req.hdr.nlmsg_pid = 0;
+
+	req.ifinfo.ifi_family = AF_UNSPEC;
+	req.ifinfo.ifi_type = 0;
+	req.ifinfo.ifi_index = drv->ifindex;
+	req.ifinfo.ifi_flags = 0;
+	req.ifinfo.ifi_change = 0;
+
+	if (linkmode != -1) {
+		rta = (struct rtattr *)
+			((char *) &req + NLMSG_ALIGN(req.hdr.nlmsg_len));
+		rta->rta_type = IFLA_LINKMODE;
+		rta->rta_len = RTA_LENGTH(sizeof(char));
+		*((char *) RTA_DATA(rta)) = linkmode;
+		req.hdr.nlmsg_len = NLMSG_ALIGN(req.hdr.nlmsg_len) +
+			RTA_LENGTH(sizeof(char));
+	}
+	if (operstate != -1) {
+		rta = (struct rtattr *)
+			((char *) &req + NLMSG_ALIGN(req.hdr.nlmsg_len));
+		rta->rta_type = IFLA_OPERSTATE;
+		rta->rta_len = RTA_LENGTH(sizeof(char));
+		*((char *) RTA_DATA(rta)) = operstate;
+		req.hdr.nlmsg_len = NLMSG_ALIGN(req.hdr.nlmsg_len) +
+			RTA_LENGTH(sizeof(char));
+	}
+
+	wpa_printf(MSG_DEBUG, "WEXT: Operstate: linkmode=%d, operstate=%d",
+		   linkmode, operstate);
+
+	ret = send(drv->event_sock, &req, req.hdr.nlmsg_len, 0);
+	if (ret < 0) {
+		wpa_printf(MSG_DEBUG, "WEXT: Sending operstate IFLA failed: "
+			   "%s (assume operstate is not supported)",
+			   strerror(errno));
+	}
+
+	return ret < 0 ? -1 : 0;
+}
+
 
 static int wpa_driver_wext_set_auth_param(struct wpa_driver_wext_data *drv,
 					  int idx, u32 value)
@@ -73,7 +146,7 @@ static int wpa_driver_wext_set_auth_param(struct wpa_driver_wext_data *drv,
 		perror("ioctl[SIOCSIWAUTH]");
 		fprintf(stderr, "WEXT auth param %d value 0x%x - ",
 			idx, value);
-		ret = -1;
+		ret = errno == EOPNOTSUPP ? -2 : -1;
 	}
 
 	return ret;
@@ -154,8 +227,18 @@ int wpa_driver_wext_get_ssid(void *priv, u8 *ssid)
 	if (ioctl(drv->ioctl_sock, SIOCGIWESSID, &iwr) < 0) {
 		perror("ioctl[SIOCGIWESSID]");
 		ret = -1;
-	} else
+	} else {
 		ret = iwr.u.essid.length;
+		if (ret > 32)
+			ret = 32;
+		/* Some drivers include nul termination in the SSID, so let's
+		 * remove it here before further processing. WE-21 changes this
+		 * to explicitly require the length _not_ to include nul
+		 * termination. */
+		if (ret > 0 && ssid[ret - 1] == '\0' &&
+		    drv->we_version_compiled < 21)
+			ret--;
+	}
 
 	return ret;
 }
@@ -185,12 +268,19 @@ int wpa_driver_wext_set_ssid(void *priv, const u8 *ssid, size_t ssid_len)
 	memset(buf, 0, sizeof(buf));
 	memcpy(buf, ssid, ssid_len);
 	iwr.u.essid.pointer = (caddr_t) buf;
-	/* For historic reasons, set SSID length to include one extra
-	 * character, C string nul termination, even though SSID is really an
-	 * octet string that should not be presented as a C string. Some Linux
-	 * drivers decrement the length by one and can thus end up missing the
-	 * last octet of the SSID if the length is not incremented here. */
-	iwr.u.essid.length = ssid_len ? ssid_len + 1 : 0;
+	if (drv->we_version_compiled < 21) {
+		/* For historic reasons, set SSID length to include one extra
+		 * character, C string nul termination, even though SSID is
+		 * really an octet string that should not be presented as a C
+		 * string. Some Linux drivers decrement the length by one and
+		 * can thus end up missing the last octet of the SSID if the
+		 * length is not incremented here. WE-21 changes this to
+		 * explicitly require the length _not_ to include nul
+		 * termination. */
+		if (ssid_len)
+			ssid_len++;
+	}
+	iwr.u.essid.length = ssid_len;
 
 	if (ioctl(drv->ioctl_sock, SIOCSIWESSID, &iwr) < 0) {
 		perror("ioctl[SIOCSIWESSID]");
@@ -228,8 +318,7 @@ int wpa_driver_wext_set_freq(void *priv, int freq)
 
 
 static void
-wpa_driver_wext_event_wireless_custom(struct wpa_driver_wext_data *drv,
-				      void *ctx, char *custom)
+wpa_driver_wext_event_wireless_custom(void *ctx, char *custom)
 {
 	union wpa_event_data data;
 
@@ -286,12 +375,30 @@ wpa_driver_wext_event_wireless_custom(struct wpa_driver_wext_data *drv,
 	done:
 		free(data.assoc_info.resp_ies);
 		free(data.assoc_info.req_ies);
+#ifdef CONFIG_STAKEY
+	} else if (strncmp(custom, "STAKey-Request=", 15) == 0) {
+		if (hwaddr_aton(custom + 15, data.stakey_request.peer)) {
+			wpa_printf(MSG_DEBUG, "WEXT: unrecognized "
+				   "STAKey-Request '%s'", custom + 15);
+			return;
+		}
+		wpa_supplicant_event(ctx, EVENT_STAKEY_REQUEST, &data);
+#endif /* CONFIG_STAKEY */
+#ifdef CONFIG_PEERKEY
+	} else if (strncmp(custom, "STKSTART.request=", 17) == 0) {
+		if (hwaddr_aton(custom + 17, data.stkstart.peer)) {
+			wpa_printf(MSG_DEBUG, "WEXT: unrecognized "
+				   "STKSTART.request '%s'", custom + 17);
+			return;
+		}
+		wpa_supplicant_event(ctx, EVENT_STKSTART, &data);
+#endif /* CONFIG_PEERKEY */
 	}
 }
 
 
 static int wpa_driver_wext_event_wireless_michaelmicfailure(
-	struct wpa_driver_wext_data *drv, void *ctx, const char *ev, int len)
+	void *ctx, const char *ev, size_t len)
 {
 	const struct iw_michaelmicfailure *mic;
 	union wpa_event_data data;
@@ -314,7 +421,7 @@ static int wpa_driver_wext_event_wireless_michaelmicfailure(
 
 
 static int wpa_driver_wext_event_wireless_pmkidcand(
-	struct wpa_driver_wext_data *drv, void *ctx, const char *ev, int len)
+	struct wpa_driver_wext_data *drv, const char *ev, size_t len)
 {
 	const struct iw_pmkid_cand *cand;
 	union wpa_event_data data;
@@ -341,7 +448,7 @@ static int wpa_driver_wext_event_wireless_pmkidcand(
 
 
 static int wpa_driver_wext_event_wireless_assocreqie(
-	struct wpa_driver_wext_data *drv, void *ctx, const char *ev, int len)
+	struct wpa_driver_wext_data *drv, const char *ev, int len)
 {
 	if (len < 0)
 		return -1;
@@ -362,7 +469,7 @@ static int wpa_driver_wext_event_wireless_assocreqie(
 
 
 static int wpa_driver_wext_event_wireless_assocrespie(
-	struct wpa_driver_wext_data *drv, void *ctx, const char *ev, int len)
+	struct wpa_driver_wext_data *drv, const char *ev, int len)
 {
 	if (len < 0)
 		return -1;
@@ -468,7 +575,7 @@ static void wpa_driver_wext_event_wireless(struct wpa_driver_wext_data *drv,
 			break;
 		case IWEVMICHAELMICFAILURE:
 			wpa_driver_wext_event_wireless_michaelmicfailure(
-				drv, ctx, custom, iwe->u.data.length);
+				ctx, custom, iwe->u.data.length);
 			break;
 		case IWEVCUSTOM:
 			if (custom + iwe->u.data.length > end)
@@ -478,7 +585,7 @@ static void wpa_driver_wext_event_wireless(struct wpa_driver_wext_data *drv,
 				return;
 			memcpy(buf, custom, iwe->u.data.length);
 			buf[iwe->u.data.length] = '\0';
-			wpa_driver_wext_event_wireless_custom(drv, ctx, buf);
+			wpa_driver_wext_event_wireless_custom(ctx, buf);
 			free(buf);
 			break;
 		case SIOCGIWSCAN:
@@ -488,15 +595,15 @@ static void wpa_driver_wext_event_wireless(struct wpa_driver_wext_data *drv,
 			break;
 		case IWEVASSOCREQIE:
 			wpa_driver_wext_event_wireless_assocreqie(
-				drv, ctx, custom, iwe->u.data.length);
+				drv, custom, iwe->u.data.length);
 			break;
 		case IWEVASSOCRESPIE:
 			wpa_driver_wext_event_wireless_assocrespie(
-				drv, ctx, custom, iwe->u.data.length);
+				drv, custom, iwe->u.data.length);
 			break;
 		case IWEVPMKIDCAND:
 			wpa_driver_wext_event_wireless_pmkidcand(
-				drv, ctx, custom, iwe->u.data.length);
+				drv, custom, iwe->u.data.length);
 			break;
 		}
 
@@ -505,8 +612,7 @@ static void wpa_driver_wext_event_wireless(struct wpa_driver_wext_data *drv,
 }
 
 
-static void wpa_driver_wext_event_link(struct wpa_driver_wext_data *drv,
-				       void *ctx, char *buf, size_t len,
+static void wpa_driver_wext_event_link(void *ctx, char *buf, size_t len,
 				       int del)
 {
 	union wpa_event_data event;
@@ -529,7 +635,7 @@ static void wpa_driver_wext_event_link(struct wpa_driver_wext_data *drv,
 
 static void wpa_driver_wext_event_rtm_newlink(struct wpa_driver_wext_data *drv,
 					      void *ctx, struct nlmsghdr *h,
-					      int len)
+					      size_t len)
 {
 	struct ifinfomsg *ifi;
 	int attrlen, nlmsg_len, rta_len;
@@ -547,6 +653,17 @@ static void wpa_driver_wext_event_rtm_newlink(struct wpa_driver_wext_data *drv,
 		return;
 	}
 
+	/*
+	 * Some drivers send the association event before the operup event--in
+	 * this case, lifting operstate in wpa_driver_wext_set_operstate()
+	 * fails. This will hit us when wpa_supplicant does not need to do
+	 * IEEE 802.1X authentication
+	 */
+	if (drv->ifindex == ifi->ifi_index && drv->operstate == 1 &&
+	    (ifi->ifi_flags & (IFF_LOWER_UP | IFF_DORMANT)) == IFF_LOWER_UP &&
+	    !(ifi->ifi_flags & IFF_RUNNING))
+		wpa_driver_wext_send_oper_ifla(drv, -1, IF_OPER_UP);
+
 	nlmsg_len = NLMSG_ALIGN(sizeof(struct ifinfomsg));
 
 	attrlen = h->nlmsg_len - nlmsg_len;
@@ -562,7 +679,7 @@ static void wpa_driver_wext_event_rtm_newlink(struct wpa_driver_wext_data *drv,
 				drv, ctx, ((char *) attr) + rta_len,
 				attr->rta_len - rta_len);
 		} else if (attr->rta_type == IFLA_IFNAME) {
-			wpa_driver_wext_event_link(drv, ctx,
+			wpa_driver_wext_event_link(ctx,
 						   ((char *) attr) + rta_len,
 						   attr->rta_len - rta_len, 0);
 		}
@@ -573,7 +690,7 @@ static void wpa_driver_wext_event_rtm_newlink(struct wpa_driver_wext_data *drv,
 
 static void wpa_driver_wext_event_rtm_dellink(struct wpa_driver_wext_data *drv,
 					      void *ctx, struct nlmsghdr *h,
-					      int len)
+					      size_t len)
 {
 	struct ifinfomsg *ifi;
 	int attrlen, nlmsg_len, rta_len;
@@ -595,7 +712,7 @@ static void wpa_driver_wext_event_rtm_dellink(struct wpa_driver_wext_data *drv,
 	rta_len = RTA_ALIGN(sizeof(struct rtattr));
 	while (RTA_OK(attr, attrlen)) {
 		if (attr->rta_type == IFLA_IFNAME) {
-			wpa_driver_wext_event_link(drv, ctx,
+			wpa_driver_wext_event_link(ctx,
 						   ((char *) attr) + rta_len,
 						   attr->rta_len - rta_len, 1);
 		}
@@ -623,7 +740,7 @@ static void wpa_driver_wext_event_receive(int sock, void *eloop_ctx,
 	}
 
 	h = (struct nlmsghdr *) buf;
-	while (left >= sizeof(*h)) {
+	while (left >= (int) sizeof(*h)) {
 		int len, plen;
 
 		len = h->nlmsg_len;
@@ -658,6 +775,22 @@ static void wpa_driver_wext_event_receive(int sock, void *eloop_ctx,
 }
 
 
+static int wpa_driver_wext_get_ifflags_ifname(struct wpa_driver_wext_data *drv,
+					      const char *ifname, int *flags)
+{
+	struct ifreq ifr;
+
+	memset(&ifr, 0, sizeof(ifr));
+	strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
+	if (ioctl(drv->ioctl_sock, SIOCGIFFLAGS, (caddr_t) &ifr) < 0) {
+		perror("ioctl[SIOCGIFFLAGS]");
+		return -1;
+	}
+	*flags = ifr.ifr_flags & 0xffff;
+	return 0;
+}
+
+
 /**
  * wpa_driver_wext_get_ifflags - Get interface flags (SIOCGIFFLAGS)
  * @drv: driver_wext private data
@@ -665,16 +798,23 @@ static void wpa_driver_wext_event_receive(int sock, void *eloop_ctx,
  * Returns: 0 on success, -1 on failure
  */
 int wpa_driver_wext_get_ifflags(struct wpa_driver_wext_data *drv, int *flags)
+{
+	return wpa_driver_wext_get_ifflags_ifname(drv, drv->ifname, flags);
+}
+
+
+static int wpa_driver_wext_set_ifflags_ifname(struct wpa_driver_wext_data *drv,
+					      const char *ifname, int flags)
 {
 	struct ifreq ifr;
 
 	memset(&ifr, 0, sizeof(ifr));
-	strncpy(ifr.ifr_name, drv->ifname, IFNAMSIZ);
-	if (ioctl(drv->ioctl_sock, SIOCGIFFLAGS, (caddr_t) &ifr) < 0) {
-		perror("ioctl[SIOCGIFFLAGS]");
+	strncpy(ifr.ifr_name, ifname, IFNAMSIZ);
+	ifr.ifr_flags = flags & 0xffff;
+	if (ioctl(drv->ioctl_sock, SIOCSIFFLAGS, (caddr_t) &ifr) < 0) {
+		perror("SIOCSIFFLAGS");
 		return -1;
 	}
-	*flags = ifr.ifr_flags & 0xffff;
 	return 0;
 }
 
@@ -687,16 +827,7 @@ int wpa_driver_wext_get_ifflags(struct wpa_driver_wext_data *drv, int *flags)
  */
 int wpa_driver_wext_set_ifflags(struct wpa_driver_wext_data *drv, int flags)
 {
-	struct ifreq ifr;
-
-	memset(&ifr, 0, sizeof(ifr));
-	strncpy(ifr.ifr_name, drv->ifname, IFNAMSIZ);
-	ifr.ifr_flags = flags & 0xffff;
-	if (ioctl(drv->ioctl_sock, SIOCSIFFLAGS, (caddr_t) &ifr) < 0) {
-		perror("SIOCSIFFLAGS");
-		return -1;
-	}
-	return 0;
+	return wpa_driver_wext_set_ifflags_ifname(drv, drv->ifname, flags);
 }
 
 
@@ -713,10 +844,9 @@ void * wpa_driver_wext_init(void *ctx, const char *ifname)
 	struct sockaddr_nl local;
 	struct wpa_driver_wext_data *drv;
 
-	drv = malloc(sizeof(*drv));
+	drv = wpa_zalloc(sizeof(*drv));
 	if (drv == NULL)
 		return NULL;
-	memset(drv, 0, sizeof(*drv));
 	drv->ctx = ctx;
 	strncpy(drv->ifname, ifname, sizeof(drv->ifname));
 
@@ -749,6 +879,8 @@ void * wpa_driver_wext_init(void *ctx, const char *ifname)
 	eloop_register_read_sock(s, wpa_driver_wext_event_receive, drv, ctx);
 	drv->event_sock = s;
 
+	drv->mlme_sock = -1;
+
 	/*
 	 * Make sure that the driver does not have any obsolete PMKID entries.
 	 */
@@ -782,6 +914,8 @@ void * wpa_driver_wext_init(void *ctx, const char *ifname)
 		wpa_driver_wext_alternative_ifindex(drv, ifname2);
 	}
 
+	wpa_driver_wext_send_oper_ifla(drv, 1, IF_OPER_DORMANT);
+
 	return drv;
 }
 
@@ -804,13 +938,24 @@ void wpa_driver_wext_deinit(void *priv)
 	 */
 	wpa_driver_wext_set_bssid(drv, (u8 *) "\x00\x00\x00\x00\x00\x00");
 
+	wpa_driver_wext_send_oper_ifla(priv, 0, IF_OPER_UP);
+
 	eloop_unregister_read_sock(drv->event_sock);
 
 	if (wpa_driver_wext_get_ifflags(drv, &flags) == 0)
 		(void) wpa_driver_wext_set_ifflags(drv, flags & ~IFF_UP);
 
+#ifdef CONFIG_CLIENT_MLME
+	if (drv->mlmedev[0] &&
+	    wpa_driver_wext_get_ifflags_ifname(drv, drv->mlmedev, &flags) == 0)
+		(void) wpa_driver_wext_set_ifflags_ifname(drv, drv->mlmedev,
+							  flags & ~IFF_UP);
+#endif /* CONFIG_CLIENT_MLME */
+
 	close(drv->event_sock);
 	close(drv->ioctl_sock);
+	if (drv->mlme_sock >= 0)
+		close(drv->mlme_sock);
 	free(drv->assoc_req_ies);
 	free(drv->assoc_resp_ies);
 	free(drv);
@@ -911,7 +1056,11 @@ static int wpa_scan_result_compar(const void *a, const void *b)
 
 	/* use freq for channel preference */
 
-	/* all things being equal, use signal level */
+	/* all things being equal, use signal level; if signal levels are
+	 * identical, use quality values since some drivers may only report
+	 * that value and leave the signal level zero */
+	if (wb->level == wa->level)
+		return wb->qual - wa->qual;
 	return wb->level - wa->level;
 }
 
@@ -934,7 +1083,8 @@ int wpa_driver_wext_get_scan_results(void *priv,
 {
 	struct wpa_driver_wext_data *drv = priv;
 	struct iwreq iwr;
-	int ap_num = 0, first, maxrate;
+	size_t ap_num = 0;
+	int first, maxrate;
 	u8 *res_buf;
 	struct iw_event iwe_buf, *iwe = &iwe_buf;
 	char *pos, *end, *custom, *genie, *gpos, *gend;
@@ -1069,7 +1219,8 @@ int wpa_driver_wext_get_scan_results(void *priv,
 			if (custom + clen > end)
 				break;
 			maxrate = 0;
-			while (((ssize_t) clen) >= sizeof(struct iw_param)) {
+			while (((ssize_t) clen) >=
+			       (ssize_t) sizeof(struct iw_param)) {
 				/* Note: may be misaligned, make a local,
 				 * aligned copy */
 				memcpy(&p, custom, sizeof(struct iw_param));
@@ -1166,16 +1317,17 @@ int wpa_driver_wext_get_scan_results(void *priv,
 	if (!first)
 		ap_num++;
 	if (ap_num > max_size) {
-		wpa_printf(MSG_DEBUG, "Too small scan result buffer - %d BSSes"
-			   " but room only for %lu", ap_num,
+		wpa_printf(MSG_DEBUG, "Too small scan result buffer - "
+			   "%lu BSSes but room only for %lu",
+			   (unsigned long) ap_num,
 			   (unsigned long) max_size);
 		ap_num = max_size;
 	}
 	qsort(results, ap_num, sizeof(struct wpa_scan_result),
 	      wpa_scan_result_compar);
 
-	wpa_printf(MSG_DEBUG, "Received %lu bytes of scan results (%d BSSes)",
-		   (unsigned long) len, ap_num);
+	wpa_printf(MSG_DEBUG, "Received %lu bytes of scan results (%lu BSSes)",
+		   (unsigned long) len, (unsigned long) ap_num);
 
 	return ap_num;
 }
@@ -1194,10 +1346,9 @@ static int wpa_driver_wext_get_range(void *priv)
 	 * structure to grow in the future.
 	 */
 	buflen = sizeof(struct iw_range) + 500;
-	range = malloc(buflen);
+	range = wpa_zalloc(buflen);
 	if (range == NULL)
 		return -1;
-	memset(range, 0, buflen);
 
 	memset(&iwr, 0, sizeof(iwr));
 	strncpy(iwr.ifr_name, drv->ifname, IFNAMSIZ);
@@ -1273,10 +1424,9 @@ static int wpa_driver_wext_set_key_ext(void *priv, wpa_alg alg,
 		return -1;
 	}
 
-	ext = malloc(sizeof(*ext) + key_len);
+	ext = wpa_zalloc(sizeof(*ext) + key_len);
 	if (ext == NULL)
 		return -1;
-	memset(ext, 0, sizeof(*ext) + key_len);
 	memset(&iwr, 0, sizeof(iwr));
 	strncpy(iwr.ifr_name, drv->ifname, IFNAMSIZ);
 	iwr.u.encoding.flags = key_idx + 1;
@@ -1443,6 +1593,7 @@ static int wpa_driver_wext_set_drop_unencrypted(void *priv,
 {
 	struct wpa_driver_wext_data *drv = priv;
 	wpa_printf(MSG_DEBUG, "%s", __FUNCTION__);
+	drv->use_crypt = enabled;
 	return wpa_driver_wext_set_auth_param(drv, IW_AUTH_DROP_UNENCRYPTED,
 					      enabled);
 }
@@ -1546,6 +1697,50 @@ static int wpa_driver_wext_keymgmt2wext(int keymgmt)
 }
 
 
+static int
+wpa_driver_wext_auth_alg_fallback(struct wpa_driver_wext_data *drv,
+				  struct wpa_driver_associate_params *params)
+{
+	struct iwreq iwr;
+	int ret = 0;
+
+	wpa_printf(MSG_DEBUG, "WEXT: Driver did not support "
+		   "SIOCSIWAUTH for AUTH_ALG, trying SIOCSIWENCODE");
+
+	memset(&iwr, 0, sizeof(iwr));
+	strncpy(iwr.ifr_name, drv->ifname, IFNAMSIZ);
+	/* Just changing mode, not actual keys */
+	iwr.u.encoding.flags = 0;
+	iwr.u.encoding.pointer = (caddr_t) NULL;
+	iwr.u.encoding.length = 0;
+
+	/*
+	 * Note: IW_ENCODE_{OPEN,RESTRICTED} can be interpreted to mean two
+	 * different things. Here they are used to indicate Open System vs.
+	 * Shared Key authentication algorithm. However, some drivers may use
+	 * them to select between open/restricted WEP encrypted (open = allow
+	 * both unencrypted and encrypted frames; restricted = only allow
+	 * encrypted frames).
+	 */
+
+	if (!drv->use_crypt) {
+		iwr.u.encoding.flags |= IW_ENCODE_DISABLED;
+	} else {
+		if (params->auth_alg & AUTH_ALG_OPEN_SYSTEM)
+			iwr.u.encoding.flags |= IW_ENCODE_OPEN;
+		if (params->auth_alg & AUTH_ALG_SHARED_KEY)
+			iwr.u.encoding.flags |= IW_ENCODE_RESTRICTED;
+	}
+
+	if (ioctl(drv->ioctl_sock, SIOCSIWENCODE, &iwr) < 0) {
+		perror("ioctl[SIOCSIWENCODE]");
+		ret = -1;
+	}
+
+	return ret;
+}
+
+
 static int
 wpa_driver_wext_associate(void *priv,
 			  struct wpa_driver_associate_params *params)
@@ -1557,6 +1752,14 @@ wpa_driver_wext_associate(void *priv,
 
 	wpa_printf(MSG_DEBUG, "%s", __FUNCTION__);
 
+	/*
+	 * If the driver did not support SIOCSIWAUTH, fallback to
+	 * SIOCSIWENCODE here.
+	 */
+	if (drv->auth_alg_fallback &&
+	    wpa_driver_wext_auth_alg_fallback(drv, params) < 0)
+		ret = -1;
+
 	if (!params->bssid &&
 	    wpa_driver_wext_set_bssid(drv, NULL) < 0)
 		ret = -1;
@@ -1590,7 +1793,10 @@ wpa_driver_wext_associate(void *priv,
 	if (wpa_driver_wext_set_auth_param(drv,
 					   IW_AUTH_KEY_MGMT, value) < 0)
 		ret = -1;
-	value = params->key_mgmt_suite != KEY_MGMT_NONE;
+	value = params->key_mgmt_suite != KEY_MGMT_NONE ||
+		params->pairwise_suite != CIPHER_NONE ||
+		params->group_suite != CIPHER_NONE ||
+		params->wpa_ie_len;
 	if (wpa_driver_wext_set_auth_param(drv,
 					   IW_AUTH_PRIVACY_INVOKED, value) < 0)
 		ret = -1;
@@ -1623,7 +1829,7 @@ wpa_driver_wext_associate(void *priv,
 static int wpa_driver_wext_set_auth_alg(void *priv, int auth_alg)
 {
 	struct wpa_driver_wext_data *drv = priv;
-	int algs = 0;
+	int algs = 0, res;
 
 	if (auth_alg & AUTH_ALG_OPEN_SYSTEM)
 		algs |= IW_AUTH_ALG_OPEN_SYSTEM;
@@ -1636,8 +1842,10 @@ static int wpa_driver_wext_set_auth_alg(void *priv, int auth_alg)
 		algs = IW_AUTH_ALG_OPEN_SYSTEM;
 	}
 
-	return wpa_driver_wext_set_auth_param(drv,
-					      IW_AUTH_80211_AUTH_ALG, algs);
+	res = wpa_driver_wext_set_auth_param(drv, IW_AUTH_80211_AUTH_ALG,
+					     algs);
+	drv->auth_alg_fallback = res == -2;
+	return res;
 }
 
 
@@ -1747,6 +1955,318 @@ int wpa_driver_wext_alternative_ifindex(struct wpa_driver_wext_data *drv,
 }
 
 
+int wpa_driver_wext_set_operstate(void *priv, int state)
+{
+	struct wpa_driver_wext_data *drv = priv;
+
+	drv->operstate = state;
+	return wpa_driver_wext_send_oper_ifla(
+		drv, -1, state ? IF_OPER_UP : IF_OPER_DORMANT);
+}
+
+
+#ifdef CONFIG_CLIENT_MLME
+static int hostapd_ioctl(struct wpa_driver_wext_data *drv,
+			 struct prism2_hostapd_param *param, int len)
+{
+	struct iwreq iwr;
+
+	memset(&iwr, 0, sizeof(iwr));
+	strncpy(iwr.ifr_name, drv->ifname, IFNAMSIZ);
+	iwr.u.data.pointer = (caddr_t) param;
+	iwr.u.data.length = len;
+
+	if (ioctl(drv->ioctl_sock, PRISM2_IOCTL_HOSTAPD, &iwr) < 0) {
+		perror("ioctl[PRISM2_IOCTL_HOSTAPD]");
+		return -1;
+	}
+
+	return 0;
+}
+
+
+static struct wpa_hw_modes *
+wpa_driver_wext_get_hw_feature_data(void *priv, u16 *num_modes, u16 *flags)
+{
+	struct wpa_driver_wext_data *drv = priv;
+	struct prism2_hostapd_param *param;
+	u8 *pos, *end;
+	struct wpa_hw_modes *modes = NULL;
+	int i;
+
+	param = wpa_zalloc(PRISM2_HOSTAPD_MAX_BUF_SIZE);
+	if (param == NULL)
+		return NULL;
+	param->cmd = PRISM2_HOSTAPD_GET_HW_FEATURES;
+
+	if (hostapd_ioctl(drv, param, PRISM2_HOSTAPD_MAX_BUF_SIZE) < 0) {
+		perror("ioctl[PRISM2_IOCTL_HOSTAPD]");
+		goto out;
+	}
+
+	*num_modes = param->u.hw_features.num_modes;
+	*flags = param->u.hw_features.flags;
+
+	pos = param->u.hw_features.data;
+	end = pos + PRISM2_HOSTAPD_MAX_BUF_SIZE -
+		(param->u.hw_features.data - (u8 *) param);
+
+	modes = wpa_zalloc(*num_modes * sizeof(struct wpa_hw_modes));
+	if (modes == NULL)
+		goto out;
+
+	for (i = 0; i < *num_modes; i++) {
+		struct hostapd_ioctl_hw_modes_hdr *hdr;
+		struct wpa_hw_modes *feature;
+		int clen, rlen;
+
+		hdr = (struct hostapd_ioctl_hw_modes_hdr *) pos;
+		pos = (u8 *) (hdr + 1);
+		clen = hdr->num_channels * sizeof(struct wpa_channel_data);
+		rlen = hdr->num_rates * sizeof(struct wpa_rate_data);
+
+		feature = &modes[i];
+		switch (hdr->mode) {
+		case MODE_IEEE80211A:
+			feature->mode = WPA_MODE_IEEE80211A;
+			break;
+		case MODE_IEEE80211B:
+			feature->mode = WPA_MODE_IEEE80211B;
+			break;
+		case MODE_IEEE80211G:
+			feature->mode = WPA_MODE_IEEE80211G;
+			break;
+		case MODE_ATHEROS_TURBO:
+		case MODE_ATHEROS_TURBOG:
+		case MODE_ATHEROS_PRIME:
+		case MODE_ATHEROS_PRIMEG:
+		case MODE_ATHEROS_XR:
+			wpa_printf(MSG_ERROR, "Skip unsupported hw_mode=%d in "
+				   "get_hw_features data", hdr->mode);
+			pos += clen + rlen;
+			continue;
+		default:
+			wpa_printf(MSG_ERROR, "Unknown hw_mode=%d in "
+				   "get_hw_features data", hdr->mode);
+			ieee80211_sta_free_hw_features(modes, *num_modes);
+			modes = NULL;
+			break;
+		}
+		feature->num_channels = hdr->num_channels;
+		feature->num_rates = hdr->num_rates;
+
+		feature->channels = malloc(clen);
+		feature->rates = malloc(rlen);
+		if (!feature->channels || !feature->rates) {
+			ieee80211_sta_free_hw_features(modes, *num_modes);
+			modes = NULL;
+			break;
+		}
+
+		memcpy(feature->channels, pos, clen);
+		pos += clen;
+		memcpy(feature->rates, pos, rlen);
+		pos += rlen;
+	}
+
+out:
+	free(param);
+	return modes;
+}
+
+
+int wpa_driver_wext_set_channel(void *priv, wpa_hw_mode phymode, int chan,
+				int freq)
+{
+	return wpa_driver_wext_set_freq(priv, freq);
+}
+
+
+static void wpa_driver_wext_mlme_read(int sock, void *eloop_ctx,
+				      void *sock_ctx)
+{
+	struct wpa_driver_wext_data *drv = eloop_ctx;
+	int len;
+	unsigned char buf[3000];
+	struct ieee80211_frame_info *fi;
+	struct ieee80211_rx_status rx_status;
+
+	len = recv(sock, buf, sizeof(buf), 0);
+	if (len < 0) {
+		perror("recv[MLME]");
+		return;
+	}
+
+	if (len < (int) sizeof(struct ieee80211_frame_info)) {
+		wpa_printf(MSG_DEBUG, "WEXT: Too short MLME frame (len=%d)",
+			   len);
+		return;
+	}
+
+	fi = (struct ieee80211_frame_info *) buf;
+	if (ntohl(fi->version) != IEEE80211_FI_VERSION) {
+		wpa_printf(MSG_DEBUG, "WEXT: Invalid MLME frame info version "
+			   "0x%x", ntohl(fi->version));
+		return;
+	}
+
+	memset(&rx_status, 0, sizeof(rx_status));
+	rx_status.ssi = ntohl(fi->ssi_signal);
+	rx_status.channel = ntohl(fi->channel);
+
+	ieee80211_sta_rx(drv->ctx, buf + sizeof(struct ieee80211_frame_info),
+			 len - sizeof(struct ieee80211_frame_info),
+			 &rx_status);
+}
+
+
+static int wpa_driver_wext_open_mlme(struct wpa_driver_wext_data *drv)
+{
+	int flags, ifindex, s, *i;
+	struct sockaddr_ll addr;
+	struct iwreq iwr;
+
+	memset(&iwr, 0, sizeof(iwr));
+	strncpy(iwr.ifr_name, drv->ifname, IFNAMSIZ);
+	i = (int *) iwr.u.name;
+	*i++ = PRISM2_PARAM_USER_SPACE_MLME;
+	*i++ = 1;
+
+	if (ioctl(drv->ioctl_sock, PRISM2_IOCTL_PRISM2_PARAM, &iwr) < 0) {
+		wpa_printf(MSG_ERROR, "WEXT: Failed to configure driver to "
+			   "use user space MLME");
+		return -1;
+	}
+
+	ifindex = if_nametoindex(drv->mlmedev);
+	if (ifindex == 0) {
+		wpa_printf(MSG_ERROR, "WEXT: mlmedev='%s' not found",
+			   drv->mlmedev);
+		return -1;
+	}
+
+	if (wpa_driver_wext_get_ifflags_ifname(drv, drv->mlmedev, &flags) != 0
+	    || wpa_driver_wext_set_ifflags_ifname(drv, drv->mlmedev,
+						  flags | IFF_UP) != 0) {
+		wpa_printf(MSG_ERROR, "WEXT: Could not set interface "
+			   "'%s' UP", drv->mlmedev);
+		return -1;
+	}
+
+	s = socket(PF_PACKET, SOCK_RAW, htons(ETH_P_ALL));
+	if (s < 0) {
+		perror("socket[PF_PACKET,SOCK_RAW]");
+		return -1;
+	}
+
+	memset(&addr, 0, sizeof(addr));
+	addr.sll_family = AF_PACKET;
+	addr.sll_ifindex = ifindex;
+
+	if (bind(s, (struct sockaddr *) &addr, sizeof(addr)) < 0) {
+		perror("bind(MLME)");
+		return -1;
+	}
+
+	if (eloop_register_read_sock(s, wpa_driver_wext_mlme_read, drv, NULL))
+	{
+		wpa_printf(MSG_ERROR, "WEXT: Could not register MLME read "
+			   "socket");
+		close(s);
+		return -1;
+	}
+
+	return s;
+}
+
+
+static int wpa_driver_wext_send_mlme(void *priv, const u8 *data,
+				     size_t data_len)
+{
+	struct wpa_driver_wext_data *drv = priv;
+	int ret;
+
+	ret = send(drv->mlme_sock, data, data_len, 0);
+	if (ret < 0) {
+		perror("send[MLME]");
+		return -1;
+	}
+
+	return 0;
+}
+
+
+static int wpa_driver_wext_mlme_add_sta(void *priv, const u8 *addr,
+					const u8 *supp_rates,
+					size_t supp_rates_len)
+{
+	struct wpa_driver_wext_data *drv = priv;
+	struct prism2_hostapd_param param;
+	size_t len;
+
+	memset(&param, 0, sizeof(param));
+	param.cmd = PRISM2_HOSTAPD_ADD_STA;
+	memcpy(param.sta_addr, addr, ETH_ALEN);
+	len = supp_rates_len;
+	if (len > sizeof(param.u.add_sta.supp_rates))
+		len = sizeof(param.u.add_sta.supp_rates);
+	memcpy(param.u.add_sta.supp_rates, supp_rates, len);
+	return hostapd_ioctl(drv, &param, sizeof(param));
+}
+
+
+static int wpa_driver_wext_mlme_remove_sta(void *priv, const u8 *addr)
+{
+	struct wpa_driver_wext_data *drv = priv;
+	struct prism2_hostapd_param param;
+
+	memset(&param, 0, sizeof(param));
+	param.cmd = PRISM2_HOSTAPD_REMOVE_STA;
+	memcpy(param.sta_addr, addr, ETH_ALEN);
+	return hostapd_ioctl(drv, &param, sizeof(param));
+}
+
+#endif /* CONFIG_CLIENT_MLME */
+
+
+static int wpa_driver_wext_set_param(void *priv, const char *param)
+{
+#ifdef CONFIG_CLIENT_MLME
+	struct wpa_driver_wext_data *drv = priv;
+	const char *pos, *pos2;
+	size_t len;
+
+	if (param == NULL)
+		return 0;
+
+	wpa_printf(MSG_DEBUG, "%s: param='%s'", __func__, param);
+
+	pos = strstr(param, "mlmedev=");
+	if (pos) {
+		pos += 8;
+		pos2 = strchr(pos, ' ');
+		if (pos2)
+			len = pos2 - pos;
+		else
+			len = strlen(pos);
+		if (len + 1 > sizeof(drv->mlmedev))
+			return -1;
+		memcpy(drv->mlmedev, pos, len);
+		drv->mlmedev[len] = '\0';
+		wpa_printf(MSG_DEBUG, "WEXT: Using user space MLME with "
+			   "mlmedev='%s'", drv->mlmedev);
+		drv->capa.flags |= WPA_DRIVER_FLAGS_USER_SPACE_MLME;
+
+		drv->mlme_sock = wpa_driver_wext_open_mlme(drv);
+		if (drv->mlme_sock < 0)
+			return -1;
+	}
+#endif /* CONFIG_CLIENT_MLME */
+
+	return 0;
+}
+
+
 const struct wpa_driver_ops wpa_driver_wext_ops = {
 	.name = "wext",
 	.desc = "Linux wireless extensions (generic)",
@@ -1764,8 +2284,19 @@ const struct wpa_driver_ops wpa_driver_wext_ops = {
 	.set_auth_alg = wpa_driver_wext_set_auth_alg,
 	.init = wpa_driver_wext_init,
 	.deinit = wpa_driver_wext_deinit,
+	.set_param = wpa_driver_wext_set_param,
 	.add_pmkid = wpa_driver_wext_add_pmkid,
 	.remove_pmkid = wpa_driver_wext_remove_pmkid,
 	.flush_pmkid = wpa_driver_wext_flush_pmkid,
 	.get_capa = wpa_driver_wext_get_capa,
+	.set_operstate = wpa_driver_wext_set_operstate,
+#ifdef CONFIG_CLIENT_MLME
+	.get_hw_feature_data = wpa_driver_wext_get_hw_feature_data,
+	.set_channel = wpa_driver_wext_set_channel,
+	.set_ssid = wpa_driver_wext_set_ssid,
+	.set_bssid = wpa_driver_wext_set_bssid,
+	.send_mlme = wpa_driver_wext_send_mlme,
+	.mlme_add_sta = wpa_driver_wext_mlme_add_sta,
+	.mlme_remove_sta = wpa_driver_wext_mlme_remove_sta,
+#endif /* CONFIG_CLIENT_MLME */
 };
