@@ -1,6 +1,6 @@
 /*
- * WPA Supplicant / EAP-FAST (draft-cam-winget-eap-fast-00.txt)
- * Copyright (c) 2004-2005, Jouni Malinen <jkmaline@cc.hut.fi>
+ * EAP peer method: EAP-FAST (draft-cam-winget-eap-fast-03.txt)
+ * Copyright (c) 2004-2006, Jouni Malinen <jkmaline@cc.hut.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
@@ -12,9 +12,7 @@
  * See README and COPYING for more details.
  */
 
-#include <stdlib.h>
-#include <stdio.h>
-#include <string.h>
+#include "includes.h"
 
 #include "common.h"
 #include "eap_i.h"
@@ -24,6 +22,7 @@
 #include "tls.h"
 #include "eap_tlv.h"
 #include "sha1.h"
+#include "config.h"
 
 /* TODO:
  * - encrypt PAC-Key in the PAC file
@@ -37,7 +36,8 @@
 
 #define TLS_EXT_PAC_OPAQUE 35
 
-static char *pac_file_hdr = "wpa_supplicant EAP-FAST PAC file - version 1";
+static const char *pac_file_hdr =
+	"wpa_supplicant EAP-FAST PAC file - version 1";
 
 
 static void eap_fast_deinit(struct eap_sm *sm, void *priv);
@@ -59,28 +59,18 @@ struct pac_tlv_hdr {
 };
 
 
-/* draft-cam-winget-eap-fast-00.txt, 6.2 EAP-FAST Authentication Phase 1 */
+/* draft-cam-winget-eap-fast-02.txt:
+ * 6.2 EAP-FAST Authentication Phase 1: Key Derivations */
 struct eap_fast_key_block_auth {
-	/* RC4-128-SHA */
-	u8 client_write_MAC_secret[20];
-	u8 server_write_MAC_secret[20];
-	u8 client_write_key[16];
-	u8 server_write_key[16];
-	/* u8 client_write_IV[0]; */
-	/* u8 server_write_IV[0]; */
+	/* Extra key material after TLS key_block */
 	u8 session_key_seed[40];
 };
 
 
-/* draft-cam-winget-eap-fast-00.txt, 7.3 EAP-FAST Provisioning Exchange */
+/* draft-cam-winget-eap-fast-provisioning-01.txt:
+ * 3.4 Key Derivations Used in the EAP-FAST Provisioning Exchange */
 struct eap_fast_key_block_provisioning {
-	/* AES128-SHA */
-	u8 client_write_MAC_secret[20];
-	u8 server_write_MAC_secret[20];
-	u8 client_write_key[16];
-	u8 server_write_key[16];
-	u8 client_write_IV[16];
-	u8 server_write_IV[16];
+	/* Extra key material after TLS key_block */
 	u8 session_key_seed[40];
 	u8 server_challenge[16];
 	u8 client_challenge[16];
@@ -113,8 +103,8 @@ struct eap_fast_data {
 	void *phase2_priv;
 	int phase2_success;
 
-	u8 phase2_type;
-	u8 *phase2_types;
+	struct eap_method_type phase2_type;
+	struct eap_method_type *phase2_types;
 	size_t num_phase2_types;
 	int resuming; /* starting a resumed session */
 	struct eap_fast_key_block_auth *key_block_a;
@@ -188,11 +178,10 @@ static int eap_fast_add_pac(struct eap_fast_data *data,
 	}
 
 	/* Allocate a new entry and add it to the list of PACs. */
-	pac = malloc(sizeof(*pac));
+	pac = wpa_zalloc(sizeof(*pac));
 	if (pac == NULL)
 		return -1;
 
-	memset(pac, 0, sizeof(*pac));
 	memcpy(pac->pac_key, entry->pac_key, EAP_FAST_PAC_KEY_LEN);
 	if (entry->pac_opaque) {
 		pac->pac_opaque = malloc(entry->pac_opaque_len);
@@ -250,12 +239,34 @@ static int eap_fast_add_pac(struct eap_fast_data *data,
 }
 
 
-static int eap_fast_read_line(FILE *f, char *buf, size_t buf_len)
+struct eap_fast_read_ctx {
+	FILE *f;
+	const char *pos;
+	const char *end;
+};
+
+static int eap_fast_read_line(struct eap_fast_read_ctx *rc, char *buf,
+			      size_t buf_len)
 {
 	char *pos;
 
-	if (fgets(buf, buf_len, f) == NULL) {
-		return -1;
+	if (rc->f) {
+		if (fgets(buf, buf_len, rc->f) == NULL)
+			return -1;
+	} else {
+		const char *l_end;
+		size_t len;
+		if (rc->pos >= rc->end)
+			return -1;
+		l_end = rc->pos;
+		while (l_end < rc->end && *l_end != '\n')
+			l_end++;
+		len = l_end - rc->pos;
+		if (len >= buf_len)
+			len = buf_len - 1;
+		memcpy(buf, rc->pos, len);
+		buf[len] = '\0';
+		rc->pos = l_end + 1;
 	}
 
 	buf[buf_len - 1] = '\0';
@@ -294,9 +305,10 @@ static u8 * eap_fast_parse_hex(const char *value, size_t *len)
 }
 
 
-static int eap_fast_load_pac(struct eap_fast_data *data, const char *pac_file)
+static int eap_fast_load_pac(struct eap_sm *sm, struct eap_fast_data *data,
+			     const char *pac_file)
 {
-	FILE *f;
+	struct eap_fast_read_ctx rc;
 	struct eap_fast_pac *pac = NULL;
 	int count = 0;
 	char *buf, *pos;
@@ -306,11 +318,27 @@ static int eap_fast_load_pac(struct eap_fast_data *data, const char *pac_file)
 	if (pac_file == NULL)
 		return -1;
 
-	f = fopen(pac_file, "r");
-	if (f == NULL) {
-		wpa_printf(MSG_INFO, "EAP-FAST: No PAC file '%s' - assume no "
-			   "PAC entries have been provisioned", pac_file);
-		return 0;
+	memset(&rc, 0, sizeof(rc));
+
+	if (strncmp(pac_file, "blob://", 7) == 0) {
+		const struct wpa_config_blob *blob;
+		blob = eap_get_config_blob(sm, pac_file + 7);
+		if (blob == NULL) {
+			wpa_printf(MSG_INFO, "EAP-FAST: No PAC blob '%s' - "
+				   "assume no PAC entries have been "
+				   "provisioned", pac_file + 7);
+			return 0;
+		}
+		rc.pos = (char *) blob->data;
+		rc.end = (char *) blob->data + blob->len;
+	} else {
+		rc.f = fopen(pac_file, "r");
+		if (rc.f == NULL) {
+			wpa_printf(MSG_INFO, "EAP-FAST: No PAC file '%s' - "
+				   "assume no PAC entries have been "
+				   "provisioned", pac_file);
+			return 0;
+		}
 	}
 
 	buf = malloc(buf_len);
@@ -319,16 +347,17 @@ static int eap_fast_load_pac(struct eap_fast_data *data, const char *pac_file)
 	}
 
 	line++;
-	if (eap_fast_read_line(f, buf, buf_len) < 0 ||
+	if (eap_fast_read_line(&rc, buf, buf_len) < 0 ||
 	    strcmp(pac_file_hdr, buf) != 0) {
 		wpa_printf(MSG_INFO, "EAP-FAST: Unrecognized header line in "
 			   "PAC file '%s'", pac_file);
 		free(buf);
-		fclose(f);
+		if (rc.f)
+			fclose(rc.f);
 		return -1;
 	}
 
-	while (eap_fast_read_line(f, buf, buf_len) == 0) {
+	while (eap_fast_read_line(&rc, buf, buf_len) == 0) {
 		line++;
 		pos = strchr(buf, '=');
 		if (pos) {
@@ -343,14 +372,13 @@ static int eap_fast_load_pac(struct eap_fast_data *data, const char *pac_file)
 				ret = -1;
 				break;
 			}
-			pac = malloc(sizeof(*pac));
+			pac = wpa_zalloc(sizeof(*pac));
 			if (pac == NULL) {
 				wpa_printf(MSG_INFO, "EAP-FAST: No memory for "
 					   "PAC entry");
 				ret = -1;
 				break;
 			}
-			memset(pac, 0, sizeof(*pac));
 		} else if (strcmp(buf, "END") == 0) {
 			if (pac == NULL) {
 				wpa_printf(MSG_INFO, "EAP-FAST: END line "
@@ -424,7 +452,8 @@ static int eap_fast_load_pac(struct eap_fast_data *data, const char *pac_file)
 	}
 
 	free(buf);
-	fclose(f);
+	if (rc.f)
+		fclose(rc.f);
 
 	if (ret == 0) {
 		wpa_printf(MSG_DEBUG, "EAP-FAST: read %d PAC entries from "
@@ -435,67 +464,119 @@ static int eap_fast_load_pac(struct eap_fast_data *data, const char *pac_file)
 }
 
 
-static void eap_fast_write(FILE *f, const char *field, const u8 *data,
+static void eap_fast_write(char **buf, char **pos, size_t *buf_len,
+			   const char *field, const u8 *data,
 			   size_t len, int txt)
 {
-	int i;
+	size_t i, need;
 
-	if (data == NULL)
+	if (data == NULL || *buf == NULL)
 		return;
 
-	fprintf(f, "%s=", field);
-	for (i = 0; i < len; i++) {
-		fprintf(f, "%02x", data[i]);
+	need = strlen(field) + len * 2 + 30;
+	if (txt)
+		need += strlen(field) + len + 20;
+
+	if (*pos - *buf + need > *buf_len) {
+		char *nbuf = realloc(*buf, *buf_len + need);
+		if (nbuf == NULL) {
+			free(*buf);
+			*buf = NULL;
+			return;
+		}
+		*buf = nbuf;
+		*buf_len += need;
 	}
-	fprintf(f, "\n");
+
+	*pos += snprintf(*pos, *buf + *buf_len - *pos, "%s=", field);
+	*pos += wpa_snprintf_hex(*pos, *buf + *buf_len - *pos, data, len);
+	*pos += snprintf(*pos, *buf + *buf_len - *pos, "\n");
 
 	if (txt) {
-		fprintf(f, "%s-txt=", field);
+		*pos += snprintf(*pos, *buf + *buf_len - *pos,
+				 "%s-txt=", field);
 		for (i = 0; i < len; i++) {
-			fprintf(f, "%c", data[i]);
+			*pos += snprintf(*pos, *buf + *buf_len - *pos,
+					 "%c", data[i]);
 		}
-		fprintf(f, "\n");
+		*pos += snprintf(*pos, *buf + *buf_len - *pos, "\n");
 	}
 }
 
 
-static int eap_fast_save_pac(struct eap_fast_data *data, const char *pac_file)
+static int eap_fast_save_pac(struct eap_sm *sm, struct eap_fast_data *data,
+			     const char *pac_file)
 {
 	FILE *f;
 	struct eap_fast_pac *pac;
 	int count = 0;
+	char *buf, *pos;
+	size_t buf_len;
 
 	if (pac_file == NULL)
 		return -1;
 
-	f = fopen(pac_file, "w");
-	if (f == NULL) {
-		wpa_printf(MSG_INFO, "EAP-FAST: Failed to open PAC file '%s' "
-			   "for writing", pac_file);
+	buf_len = 1024;
+	pos = buf = malloc(buf_len);
+	if (buf == NULL)
 		return -1;
-	}
 
-	fprintf(f, "%s\n", pac_file_hdr);
+	pos += snprintf(pos, buf + buf_len - pos, "%s\n", pac_file_hdr);
 
 	pac = data->pac;
 	while (pac) {
-		fprintf(f, "START\n");
-		eap_fast_write(f, "PAC-Key", pac->pac_key,
+		pos += snprintf(pos, buf + buf_len - pos, "START\n");
+		eap_fast_write(&buf, &pos, &buf_len, "PAC-Key", pac->pac_key,
 			       EAP_FAST_PAC_KEY_LEN, 0);
-		eap_fast_write(f, "PAC-Opaque", pac->pac_opaque,
-			       pac->pac_opaque_len, 0);
-		eap_fast_write(f, "PAC-Info", pac->pac_info,
+		eap_fast_write(&buf, &pos, &buf_len, "PAC-Opaque",
+			       pac->pac_opaque, pac->pac_opaque_len, 0);
+		eap_fast_write(&buf, &pos, &buf_len, "PAC-Info", pac->pac_info,
 			       pac->pac_info_len, 0);
-		eap_fast_write(f, "A-ID", pac->a_id, pac->a_id_len, 0);
-		eap_fast_write(f, "I-ID", pac->i_id, pac->i_id_len, 1);
-		eap_fast_write(f, "A-ID-Info", pac->a_id_info,
-			       pac->a_id_info_len, 1);
-		fprintf(f, "END\n");
+		eap_fast_write(&buf, &pos, &buf_len, "A-ID", pac->a_id,
+			       pac->a_id_len, 0);
+		eap_fast_write(&buf, &pos, &buf_len, "I-ID", pac->i_id,
+			       pac->i_id_len, 1);
+		eap_fast_write(&buf, &pos, &buf_len, "A-ID-Info",
+			       pac->a_id_info, pac->a_id_info_len, 1);
+		pos += snprintf(pos, buf + buf_len - pos, "END\n");
 		count++;
 		pac = pac->next;
+
+		if (buf == NULL) {
+			wpa_printf(MSG_DEBUG, "EAP-FAST: No memory for PAC "
+				   "data");
+			return -1;
+		}
 	}
 
-	fclose(f);
+	if (strncmp(pac_file, "blob://", 7) == 0) {
+		struct wpa_config_blob *blob;
+		blob = wpa_zalloc(sizeof(*blob));
+		if (blob == NULL) {
+			free(buf);
+			return -1;
+		}
+		blob->data = (u8 *) buf;
+		blob->len = pos - buf;
+		buf = NULL;
+		blob->name = strdup(pac_file + 7);
+		if (blob->name == NULL) {
+			wpa_config_free_blob(blob);
+			return -1;
+		}
+		eap_set_config_blob(sm, blob);
+	} else {
+		f = fopen(pac_file, "w");
+		if (f == NULL) {
+			wpa_printf(MSG_INFO, "EAP-FAST: Failed to open PAC "
+				   "file '%s' for writing", pac_file);
+			free(buf);
+			return -1;
+		}
+		fprintf(f, "%s", buf);
+		free(buf);
+		fclose(f);
+	}
 
 	wpa_printf(MSG_DEBUG, "EAP-FAST: wrote %d PAC entries into '%s'",
 		   count, pac_file);
@@ -509,10 +590,9 @@ static void * eap_fast_init(struct eap_sm *sm)
 	struct eap_fast_data *data;
 	struct wpa_ssid *config = eap_get_config(sm);
 
-	data = malloc(sizeof(*data));
+	data = wpa_zalloc(sizeof(*data));
 	if (data == NULL)
 		return NULL;
-	memset(data, 0, sizeof(*data));
 	data->fast_version = EAP_FAST_VERSION;
 
 	if (config && config->phase1) {
@@ -525,7 +605,8 @@ static void * eap_fast_init(struct eap_sm *sm)
 
 	if (config && config->phase2) {
 		char *start, *pos, *buf;
-		u8 method, *methods = NULL, *_methods;
+		struct eap_method_type *methods = NULL, *_methods;
+		u8 method;
 		size_t num_methods = 0;
 		start = buf = strdup(config->phase2);
 		if (buf == NULL) {
@@ -533,6 +614,7 @@ static void * eap_fast_init(struct eap_sm *sm)
 			return NULL;
 		}
 		while (start && *start != '\0') {
+			int vendor;
 			pos = strstr(start, "auth=");
 			if (pos == NULL)
 				break;
@@ -545,20 +627,25 @@ static void * eap_fast_init(struct eap_sm *sm)
 			pos = strchr(start, ' ');
 			if (pos)
 				*pos++ = '\0';
-			method = eap_get_phase2_type(start);
-			if (method == EAP_TYPE_NONE) {
+			method = eap_get_phase2_type(start, &vendor);
+			if (vendor == EAP_VENDOR_IETF &&
+			    method == EAP_TYPE_NONE) {
 				wpa_printf(MSG_ERROR, "EAP-FAST: Unsupported "
 					   "Phase2 method '%s'", start);
 			} else {
 				num_methods++;
-				_methods = realloc(methods, num_methods);
+				_methods = realloc(
+					methods,
+					num_methods * sizeof(*methods));
 				if (_methods == NULL) {
 					free(methods);
+					free(buf);
 					eap_fast_deinit(sm, data);
 					return NULL;
 				}
 				methods = _methods;
-				methods[num_methods - 1] = method;
+				methods[num_methods - 1].vendor = vendor;
+				methods[num_methods - 1].method = method;
 			}
 
 			start = pos;
@@ -577,8 +664,10 @@ static void * eap_fast_init(struct eap_sm *sm)
 		return NULL;
 	}
 	wpa_hexdump(MSG_DEBUG, "EAP-FAST: Phase2 EAP types",
-		    data->phase2_types, data->num_phase2_types);
-	data->phase2_type = EAP_TYPE_NONE;
+		    (u8 *) data->phase2_types,
+		    data->num_phase2_types * sizeof(struct eap_method_type));
+	data->phase2_type.vendor = EAP_VENDOR_IETF;
+	data->phase2_type.method = EAP_TYPE_NONE;
 
 	if (eap_tls_ssl_init(sm, &data->ssl, config)) {
 		wpa_printf(MSG_INFO, "EAP-FAST: Failed to initialize SSL.");
@@ -591,7 +680,7 @@ static void * eap_fast_init(struct eap_sm *sm)
 	 * TODO: consider making this configurable */
 	tls_connection_enable_workaround(sm->ssl_ctx, data->ssl.conn);
 
-	if (eap_fast_load_pac(data, config->pac_file) < 0) {
+	if (eap_fast_load_pac(sm, data, config->pac_file) < 0) {
 		eap_fast_deinit(sm, data);
 		return NULL;
 	}
@@ -670,35 +759,43 @@ static int eap_fast_encrypt(struct eap_sm *sm, struct eap_fast_data *data,
 }
 
 
-static int eap_fast_phase2_nak(struct eap_sm *sm,
-			       struct eap_fast_data *data,
+static int eap_fast_phase2_nak(struct eap_fast_data *data,
 			       struct eap_hdr *hdr,
 			       u8 **resp, size_t *resp_len)
 {
 	struct eap_hdr *resp_hdr;
 	u8 *pos = (u8 *) (hdr + 1);
+	size_t i;
 
+	/* TODO: add support for expanded Nak */
 	wpa_printf(MSG_DEBUG, "EAP-FAST: Phase 2 Request: Nak type=%d", *pos);
 	wpa_hexdump(MSG_DEBUG, "EAP-FAST: Allowed Phase2 EAP types",
-		    data->phase2_types, data->num_phase2_types);
-	*resp_len = sizeof(struct eap_hdr) + 1 + data->num_phase2_types;
-	*resp = malloc(*resp_len);
+		    (u8 *) data->phase2_types,
+		    data->num_phase2_types * sizeof(struct eap_method_type));
+	*resp_len = sizeof(struct eap_hdr) + 1;
+	*resp = malloc(*resp_len + data->num_phase2_types);
 	if (*resp == NULL)
 		return -1;
 
 	resp_hdr = (struct eap_hdr *) (*resp);
 	resp_hdr->code = EAP_CODE_RESPONSE;
 	resp_hdr->identifier = hdr->identifier;
-	resp_hdr->length = host_to_be16(*resp_len);
 	pos = (u8 *) (resp_hdr + 1);
 	*pos++ = EAP_TYPE_NAK;
-	memcpy(pos, data->phase2_types, data->num_phase2_types);
+	for (i = 0; i < data->num_phase2_types; i++) {
+		if (data->phase2_types[i].vendor == EAP_VENDOR_IETF &&
+		    data->phase2_types[i].method < 256) {
+			(*resp_len)++;
+			*pos++ = data->phase2_types[i].method;
+		}
+	}
+	resp_hdr->length = host_to_be16(*resp_len);
 
 	return 0;
 }
 
 
-static int eap_fast_derive_msk(struct eap_sm *sm, struct eap_fast_data *data)
+static int eap_fast_derive_msk(struct eap_fast_data *data)
 {
 	u8 isk[32];
 	u8 imck[60];
@@ -733,7 +830,8 @@ static int eap_fast_set_tls_master_secret(struct eap_sm *sm,
 	size_t seed_len, server_random_len;
 
 	if (data->tls_master_secret_set || !data->current_pac ||
-	    tls_connection_get_keys(sm->ssl_ctx, data->ssl.conn, &keys)) {
+	    tls_connection_get_keys(sm->ssl_ctx, data->ssl.conn, &keys) ||
+	    keys.client_random == NULL) {
 		return 0;
 	}
 
@@ -767,7 +865,6 @@ static int eap_fast_set_tls_master_secret(struct eap_sm *sm,
 	wpa_hexdump(MSG_DEBUG, "EAP-FAST: server_random",
 		    server_random, server_random_len);
 
-
 	seed_len = keys.client_random_len + server_random_len;
 	seed = malloc(seed_len);
 	if (seed == NULL)
@@ -800,33 +897,48 @@ static u8 * eap_fast_derive_key(struct eap_sm *sm, struct eap_ssl_data *data,
 				char *label, size_t len)
 {
 	struct tls_keys keys;
-	u8 *random;
-	u8 *out;
+	u8 *rnd = NULL, *out;
+	int block_size;
+
+	block_size = tls_connection_get_keyblock_size(sm->ssl_ctx, data->conn);
+	if (block_size < 0)
+		return NULL;
+
+	out = malloc(block_size + len);
+	if (out == NULL)
+		return NULL;
+
+	if (tls_connection_prf(sm->ssl_ctx, data->conn, label, 1, out,
+			       block_size + len) == 0) {
+		memmove(out, out + block_size, len);
+		return out;
+	}
 
 	if (tls_connection_get_keys(sm->ssl_ctx, data->conn, &keys))
-		return NULL;
-	out = malloc(len);
-	random = malloc(keys.client_random_len + keys.server_random_len);
-	if (out == NULL || random == NULL) {
-		free(out);
-		free(random);
-		return NULL;
-	}
-	memcpy(random, keys.server_random, keys.server_random_len);
-	memcpy(random + keys.server_random_len, keys.client_random,
+		goto fail;
+
+	rnd = malloc(keys.client_random_len + keys.server_random_len);
+	if (rnd == NULL)
+		goto fail;
+
+	memcpy(rnd, keys.server_random, keys.server_random_len);
+	memcpy(rnd + keys.server_random_len, keys.client_random,
 	       keys.client_random_len);
 
 	wpa_hexdump_key(MSG_MSGDUMP, "EAP-FAST: master_secret for key "
 			"expansion", keys.master_key, keys.master_key_len);
 	if (tls_prf(keys.master_key, keys.master_key_len,
-		    label, random, keys.client_random_len +
-		    keys.server_random_len, out, len)) {
-		free(random);
-		free(out);
-		return NULL;
-	}
-	free(random);
+		    label, rnd, keys.client_random_len +
+		    keys.server_random_len, out, block_size + len))
+		goto fail;
+	free(rnd);
+	memmove(out, out + block_size, len);
 	return out;
+
+fail:
+	free(rnd);
+	free(out);
+	return NULL;
 }
 
 
@@ -837,6 +949,11 @@ static void eap_fast_derive_key_auth(struct eap_sm *sm,
 	data->key_block_a = (struct eap_fast_key_block_auth *)
 		eap_fast_derive_key(sm, &data->ssl, "key expansion",
 				    sizeof(*data->key_block_a));
+	if (data->key_block_a == NULL) {
+		wpa_printf(MSG_DEBUG, "EAP-FAST: Failed to derive "
+			   "session_key_seed");
+		return;
+	}
 	wpa_hexdump_key(MSG_DEBUG, "EAP-FAST: session_key_seed",
 			data->key_block_a->session_key_seed,
 			sizeof(data->key_block_a->session_key_seed));
@@ -899,27 +1016,37 @@ static int eap_fast_phase2_request(struct eap_sm *sm,
 		*resp = eap_sm_buildIdentity(sm, req->identifier, resp_len, 1);
 		break;
 	default:
-		if (data->phase2_type == EAP_TYPE_NONE) {
-			int i;
+		if (data->phase2_type.vendor == EAP_VENDOR_IETF &&
+		    data->phase2_type.method == EAP_TYPE_NONE) {
+			size_t i;
 			for (i = 0; i < data->num_phase2_types; i++) {
-				if (data->phase2_types[i] != *pos)
+				if (data->phase2_types[i].vendor !=
+				    EAP_VENDOR_IETF ||
+				    data->phase2_types[i].method != *pos)
 					continue;
 
-				data->phase2_type = *pos;
+				data->phase2_type.vendor =
+					data->phase2_types[i].vendor;
+				data->phase2_type.method =
+					data->phase2_types[i].method;
 				wpa_printf(MSG_DEBUG, "EAP-FAST: Selected "
-					   "Phase 2 EAP method %d",
-					   data->phase2_type);
+					   "Phase 2 EAP vendor %d method %d",
+					   data->phase2_type.vendor,
+					   data->phase2_type.method);
 				break;
 			}
 		}
-		if (*pos != data->phase2_type || *pos == EAP_TYPE_NONE) {
-			if (eap_fast_phase2_nak(sm, data, hdr, resp, resp_len))
+		if (*pos != data->phase2_type.method ||
+		    *pos == EAP_TYPE_NONE) {
+			if (eap_fast_phase2_nak(data, hdr, resp, resp_len))
 				return -1;
 			return 0;
 		}
 
 		if (data->phase2_priv == NULL) {
-			data->phase2_method = eap_sm_get_eap_methods(*pos);
+			data->phase2_method = eap_sm_get_eap_methods(
+				data->phase2_type.vendor,
+				data->phase2_type.method);
 			if (data->phase2_method) {
 				if (data->key_block_p) {
 					sm->auth_challenge =
@@ -1005,11 +1132,10 @@ static u8 * eap_fast_tlv_pac_ack(size_t *len)
 	struct eap_tlv_pac_ack_tlv *ack;
 
 	*len = sizeof(*res) + sizeof(*ack);
-	res = malloc(*len);
+	res = wpa_zalloc(*len);
 	if (res == NULL)
 		return NULL;
 
-	memset(res, 0, *len);
 	res->tlv_type = host_to_be16(EAP_TLV_RESULT_TLV |
 				     EAP_TLV_TYPE_MANDATORY);
 	res->length = host_to_be16(sizeof(*res) - sizeof(struct eap_tlv_hdr));
@@ -1157,10 +1283,9 @@ static u8 * eap_fast_process_crypto_binding(
 	}
 
 	*resp_len = sizeof(*rresult) + sizeof(*rbind);
-	resp = malloc(*resp_len);
+	resp = wpa_zalloc(*resp_len);
 	if (resp == NULL)
 		return NULL;
-	memset(resp, 0, *resp_len);
 
 	/* Both intermediate and final Result TLVs are identical, so ok to use
 	 * the same structure definition for them. */
@@ -1172,7 +1297,7 @@ static u8 * eap_fast_process_crypto_binding(
 	rresult->status = host_to_be16(EAP_TLV_RESULT_SUCCESS);
 
 	if (!data->provisioning && data->phase2_success &&
-	    eap_fast_derive_msk(sm, data) < 0) {
+	    eap_fast_derive_msk(data) < 0) {
 		wpa_printf(MSG_INFO, "EAP-FAST: Failed to generate MSK");
 		ret->methodState = METHOD_DONE;
 		ret->decision = DECISION_FAIL;
@@ -1334,7 +1459,7 @@ static u8 * eap_fast_process_pac(struct eap_sm *sm, struct eap_fast_data *data,
 	}
 
 	eap_fast_add_pac(data, &entry);
-	eap_fast_save_pac(data, config->pac_file);
+	eap_fast_save_pac(sm, data, config->pac_file);
 
 	if (data->provisioning) {
 		/* EAP-FAST provisioning does not provide keying material and
@@ -1363,10 +1488,10 @@ static int eap_fast_decrypt(struct eap_sm *sm, struct eap_fast_data *data,
 			    u8 **out_data, size_t *out_len)
 {
 	u8 *in_decrypted, *pos, *end;
-	int buf_len, len_decrypted, len;
+	int len_decrypted, len;
 	struct eap_hdr *hdr;
 	u8 *resp = NULL;
-	size_t resp_len;
+	size_t buf_len, resp_len;
 	int mandatory, tlv_type;
 	u8 *eap_payload_tlv = NULL, *pac = NULL;
 	size_t eap_payload_tlv_len = 0, pac_len = 0;
@@ -1825,7 +1950,15 @@ static int eap_fast_get_status(struct eap_sm *sm, void *priv, char *buf,
 			       size_t buflen, int verbose)
 {
 	struct eap_fast_data *data = priv;
-	return eap_tls_status(sm, &data->ssl, buf, buflen, verbose);
+	int len;
+
+	len = eap_tls_status(sm, &data->ssl, buf, buflen, verbose);
+	if (data->phase2_method) {
+		len += snprintf(buf + len, buflen - len,
+				"EAP-FAST Phase2 method=%s\n",
+				data->phase2_method->name);
+	}
+	return len;
 }
 
 
@@ -1855,19 +1988,30 @@ static u8 * eap_fast_getKey(struct eap_sm *sm, void *priv, size_t *len)
 }
 
 
-const struct eap_method eap_method_fast =
+int eap_peer_fast_register(void)
 {
-	.method = EAP_TYPE_FAST,
-	.name = "FAST",
-	.init = eap_fast_init,
-	.deinit = eap_fast_deinit,
-	.process = eap_fast_process,
-	.isKeyAvailable = eap_fast_isKeyAvailable,
-	.getKey = eap_fast_getKey,
-	.get_status = eap_fast_get_status,
+	struct eap_method *eap;
+	int ret;
+
+	eap = eap_peer_method_alloc(EAP_PEER_METHOD_INTERFACE_VERSION,
+				    EAP_VENDOR_IETF, EAP_TYPE_FAST, "FAST");
+	if (eap == NULL)
+		return -1;
+
+	eap->init = eap_fast_init;
+	eap->deinit = eap_fast_deinit;
+	eap->process = eap_fast_process;
+	eap->isKeyAvailable = eap_fast_isKeyAvailable;
+	eap->getKey = eap_fast_getKey;
+	eap->get_status = eap_fast_get_status;
 #if 0
-	.has_reauth_data = eap_fast_has_reauth_data,
-	.deinit_for_reauth = eap_fast_deinit_for_reauth,
-	.init_for_reauth = eap_fast_init_for_reauth,
+	eap->has_reauth_data = eap_fast_has_reauth_data;
+	eap->deinit_for_reauth = eap_fast_deinit_for_reauth;
+	eap->init_for_reauth = eap_fast_init_for_reauth;
 #endif
-};
+
+	ret = eap_peer_method_register(eap);
+	if (ret)
+		eap_peer_method_free(eap);
+	return ret;
+}
