@@ -12,9 +12,7 @@
  * See README and COPYING for more details.
  */
 
-#include <stdlib.h>
-#include <stdio.h>
-#include <string.h>
+#include "includes.h"
 
 #include "common.h"
 #include "eapol_sm.h"
@@ -24,6 +22,10 @@
 #include "wpa.h"
 #include "md5.h"
 #include "rc4.h"
+#include "state_machine.h"
+
+#define STATE_MACHINE_DATA struct eapol_sm
+#define STATE_MACHINE_DEBUG_PREFIX "EAPOL"
 
 
 /* IEEE 802.1X-2004 - Supplicant - EAPOL state machines */
@@ -150,6 +152,10 @@ struct eapol_sm {
 #define IEEE8021X_KEY_INDEX_FLAG 0x80
 #define IEEE8021X_KEY_INDEX_MASK 0x03
 
+#ifdef _MSC_VER
+#pragma pack(push, 1)
+#endif /* _MSC_VER */
+
 struct ieee802_1x_eapol_key {
 	u8 type;
 	/* Note: key_length is unaligned */
@@ -172,7 +178,11 @@ struct ieee802_1x_eapol_key {
 	 * represents the number of least significant octets from
 	 * MS-MPPE-Send-Key attribute to be used as the keying material;
 	 * RC4 key used in encryption = Key-IV + MS-MPPE-Recv-Key */
-} __attribute__ ((packed));
+} STRUCT_PACKED;
+
+#ifdef _MSC_VER
+#pragma pack(pop)
+#endif /* _MSC_VER */
 
 
 static void eapol_sm_txLogoff(struct eapol_sm *sm);
@@ -185,29 +195,6 @@ static void eapol_sm_abort_cached(struct eapol_sm *sm);
 static void eapol_sm_step_timeout(void *eloop_ctx, void *timeout_ctx);
 
 
-/* Definitions for clarifying state machine implementation */
-#define SM_STATE(machine, state) \
-static void sm_ ## machine ## _ ## state ## _Enter(struct eapol_sm *sm, \
-	int global)
-
-#define SM_ENTRY(machine, state) \
-if (!global || sm->machine ## _state != machine ## _ ## state) { \
-	sm->changed = TRUE; \
-	wpa_printf(MSG_DEBUG, "EAPOL: " #machine " entering state " #state); \
-} \
-sm->machine ## _state = machine ## _ ## state;
-
-#define SM_ENTER(machine, state) \
-sm_ ## machine ## _ ## state ## _Enter(sm, 0)
-#define SM_ENTER_GLOBAL(machine, state) \
-sm_ ## machine ## _ ## state ## _Enter(sm, 1)
-
-#define SM_STEP(machine) \
-static void sm_ ## machine ## _Step(struct eapol_sm *sm)
-
-#define SM_STEP_RUN(machine) sm_ ## machine ## _Step(sm)
-
-
 /* Port Timers state machine - implemented as a function that will be called
  * once a second as a registered event loop timeout */
 static void eapol_port_timers_tick(void *eloop_ctx, void *timeout_ctx)
@@ -265,11 +252,24 @@ SM_STATE(SUPP_PAE, DISCONNECTED)
 
 SM_STATE(SUPP_PAE, CONNECTING)
 {
+	int send_start = sm->SUPP_PAE_state == SUPP_PAE_CONNECTING;
 	SM_ENTRY(SUPP_PAE, CONNECTING);
-	sm->startWhen = sm->startPeriod;
-	sm->startCount++;
+	if (send_start) {
+		sm->startWhen = sm->startPeriod;
+		sm->startCount++;
+	} else {
+		/*
+		 * Do not send EAPOL-Start immediately since in most cases,
+		 * Authenticator is going to start authentication immediately
+		 * after association and an extra EAPOL-Start is just going to
+		 * delay authentication. Use a short timeout to send the first
+		 * EAPOL-Start if Authenticator does not start authentication.
+		 */
+		sm->startWhen = 3;
+	}
 	sm->eapolEap = FALSE;
-	eapol_sm_txStart(sm);
+	if (send_start)
+		eapol_sm_txStart(sm);
 }
 
 
@@ -524,6 +524,20 @@ SM_STEP(SUPP_BE)
 	case SUPP_BE_UNKNOWN:
 		break;
 	case SUPP_BE_REQUEST:
+		/*
+		 * IEEE Std 802.1X-2004 has transitions from REQUEST to FAIL
+		 * and SUCCESS based on eapFail and eapSuccess, respectively.
+		 * However, IEEE Std 802.1X-2004 is also specifying that
+		 * eapNoResp should be set in conjuction with eapSuccess and
+		 * eapFail which would mean that more than one of the
+		 * transitions here would be activated at the same time.
+		 * Skipping RESPONSE and/or RECEIVE states in these cases can
+		 * cause problems and the direct transitions to do not seem
+		 * correct. Because of this, the conditions for these
+		 * transitions are verified only after eapNoResp. They are
+		 * unlikely to be used since eapNoResp should always be set if
+		 * either of eapSuccess or eapFail is set.
+		 */
 		if (sm->eapResp && sm->eapNoResp) {
 			wpa_printf(MSG_DEBUG, "EAPOL: SUPP_BE REQUEST: both "
 				   "eapResp and eapNoResp set?!");
@@ -532,6 +546,10 @@ SM_STEP(SUPP_BE)
 			SM_ENTER(SUPP_BE, RESPONSE);
 		else if (sm->eapNoResp)
 			SM_ENTER(SUPP_BE, RECEIVE);
+		else if (sm->eapFail)
+			SM_ENTER(SUPP_BE, FAIL);
+		else if (sm->eapSuccess)
+			SM_ENTER(SUPP_BE, SUCCESS);
 		break;
 	case SUPP_BE_RESPONSE:
 		SM_ENTER(SUPP_BE, RECEIVE);
@@ -859,6 +877,7 @@ void eapol_sm_step(struct eapol_sm *sm)
 }
 
 
+#ifdef CONFIG_CTRL_IFACE
 static const char *eapol_supp_pae_state(int state)
 {
 	switch (state) {
@@ -914,8 +933,10 @@ static const char * eapol_port_status(PortStatus status)
 	else
 		return "Unauthorized";
 }
+#endif /* CONFIG_CTRL_IFACE */
 
 
+#if defined(CONFIG_CTRL_IFACE) || !defined(CONFIG_NO_STDOUT_DEBUG)
 static const char * eapol_port_control(PortControl ctrl)
 {
 	switch (ctrl) {
@@ -929,6 +950,7 @@ static const char * eapol_port_control(PortControl ctrl)
 		return "Unknown";
 	}
 }
+#endif /* CONFIG_CTRL_IFACE || !CONFIG_NO_STDOUT_DEBUG */
 
 
 /**
@@ -958,6 +980,7 @@ void eapol_sm_configure(struct eapol_sm *sm, int heldPeriod, int authPeriod,
 }
 
 
+#ifdef CONFIG_CTRL_IFACE
 /**
  * eapol_sm_get_status - Get EAPOL state machine status
  * @sm: Pointer to EAPOL state machine allocated with eapol_sm_init()
@@ -1020,7 +1043,8 @@ int eapol_sm_get_status(struct eapol_sm *sm, char *buf, size_t buflen,
  */
 int eapol_sm_get_mib(struct eapol_sm *sm, char *buf, size_t buflen)
 {
-	int len;
+	size_t len;
+
 	if (sm == NULL)
 		return 0;
 	len = snprintf(buf, buflen,
@@ -1030,7 +1054,20 @@ int eapol_sm_get_mib(struct eapol_sm *sm, char *buf, size_t buflen)
 		       "dot1xSuppStartPeriod=%u\n"
 		       "dot1xSuppMaxStart=%u\n"
 		       "dot1xSuppSuppControlledPortStatus=%s\n"
-		       "dot1xSuppBackendPaeState=%d\n"
+		       "dot1xSuppBackendPaeState=%d\n",
+		       sm->SUPP_PAE_state,
+		       sm->heldPeriod,
+		       sm->authPeriod,
+		       sm->startPeriod,
+		       sm->maxStart,
+		       sm->suppPortStatus == Authorized ?
+		       "Authorized" : "Unauthorized",
+		       sm->SUPP_BE_state);
+
+	if (len >= buflen)
+		return len;
+
+	len += snprintf(buf + len, buflen - len,
 		       "dot1xSuppEapolFramesRx=%u\n"
 		       "dot1xSuppEapolFramesTx=%u\n"
 		       "dot1xSuppEapolStartFramesTx=%u\n"
@@ -1042,14 +1079,6 @@ int eapol_sm_get_mib(struct eapol_sm *sm, char *buf, size_t buflen)
 		       "dot1xSuppEapLengthErrorFramesRx=%u\n"
 		       "dot1xSuppLastEapolFrameVersion=%u\n"
 		       "dot1xSuppLastEapolFrameSource=" MACSTR "\n",
-		       sm->SUPP_PAE_state,
-		       sm->heldPeriod,
-		       sm->authPeriod,
-		       sm->startPeriod,
-		       sm->maxStart,
-		       sm->suppPortStatus == Authorized ?
-		       "Authorized" : "Unauthorized",
-		       sm->SUPP_BE_state,
 		       sm->dot1xSuppEapolFramesRx,
 		       sm->dot1xSuppEapolFramesTx,
 		       sm->dot1xSuppEapolStartFramesTx,
@@ -1061,8 +1090,10 @@ int eapol_sm_get_mib(struct eapol_sm *sm, char *buf, size_t buflen)
 		       sm->dot1xSuppEapLengthErrorFramesRx,
 		       sm->dot1xSuppLastEapolFrameVersion,
 		       MAC2STR(sm->dot1xSuppLastEapolFrameSource));
+
 	return len;
 }
+#endif /* CONFIG_CTRL_IFACE */
 
 
 /**
@@ -1079,8 +1110,9 @@ int eapol_sm_rx_eapol(struct eapol_sm *sm, const u8 *src, const u8 *buf,
 {
 	const struct ieee802_1x_hdr *hdr;
 	const struct ieee802_1x_eapol_key *key;
-	int plen, data_len;
+	int data_len;
 	int res = 1;
+	size_t plen;
 
 	if (sm == NULL)
 		return 0;
@@ -1289,6 +1321,7 @@ void eapol_sm_notify_config(struct eapol_sm *sm, struct wpa_ssid *config,
 	if (sm->eap) {
 		eap_set_fast_reauth(sm->eap, conf->fast_reauth);
 		eap_set_workaround(sm->eap, conf->workaround);
+		eap_set_force_disabled(sm->eap, conf->eap_disabled);
 	}
 }
 
@@ -1384,7 +1417,13 @@ static void eapol_sm_abort_cached(struct eapol_sm *sm)
 	sm->cached_pmk = FALSE;
 	sm->SUPP_PAE_state = SUPP_PAE_CONNECTING;
 	sm->suppPortStatus = Unauthorized;
-	sm->eapRestart= TRUE;
+
+	/* Make sure we do not start sending EAPOL-Start frames first, but
+	 * instead move to RESTART state to start EAPOL authentication. */
+	sm->startWhen = 3;
+
+	if (sm->ctx->aborted_cached)
+		sm->ctx->aborted_cached(sm->ctx->ctx);
 }
 
 
@@ -1605,20 +1644,42 @@ static void eapol_sm_set_int(void *ctx, enum eapol_int_var variable,
 }
 
 
+static void eapol_sm_set_config_blob(void *ctx, struct wpa_config_blob *blob)
+{
+	struct eapol_sm *sm = ctx;
+	if (sm && sm->ctx && sm->ctx->set_config_blob)
+		sm->ctx->set_config_blob(sm->ctx->ctx, blob);
+}
+
+
+static const struct wpa_config_blob *
+eapol_sm_get_config_blob(void *ctx, const char *name)
+{
+	struct eapol_sm *sm = ctx;
+	if (sm && sm->ctx && sm->ctx->get_config_blob)
+		return sm->ctx->get_config_blob(sm->ctx->ctx, name);
+	else
+		return NULL;
+}
+
+
 static struct eapol_callbacks eapol_cb =
 {
-	.get_config = eapol_sm_get_config,
-	.get_bool = eapol_sm_get_bool,
-	.set_bool = eapol_sm_set_bool,
-	.get_int = eapol_sm_get_int,
-	.set_int = eapol_sm_set_int,
-	.get_eapReqData = eapol_sm_get_eapReqData,
+	eapol_sm_get_config,
+	eapol_sm_get_bool,
+	eapol_sm_set_bool,
+	eapol_sm_get_int,
+	eapol_sm_set_int,
+	eapol_sm_get_eapReqData,
+	eapol_sm_set_config_blob,
+	eapol_sm_get_config_blob
 };
 
 
 /**
  * eapol_sm_init - Initialize EAPOL state machine
- * @ctx: Pointer to EAPOL context data
+ * @ctx: Pointer to EAPOL context data; this needs to be an allocated buffer
+ * and EAPOL state machine will free it in eapol_sm_deinit()
  * Returns: Pointer to the allocated EAPOL state machine or %NULL on failure
  *
  * Allocate and initialize an EAPOL state machine.
@@ -1627,10 +1688,9 @@ struct eapol_sm *eapol_sm_init(struct eapol_ctx *ctx)
 {
 	struct eapol_sm *sm;
 	struct eap_config conf;
-	sm = malloc(sizeof(*sm));
+	sm = wpa_zalloc(sizeof(*sm));
 	if (sm == NULL)
 		return NULL;
-	memset(sm, 0, sizeof(*sm));
 	sm->ctx = ctx;
 
 	sm->portControl = Auto;
