@@ -1,6 +1,6 @@
 /*
  * WPA Supplicant - test code
- * Copyright (c) 2003-2005, Jouni Malinen <jkmaline@cc.hut.fi>
+ * Copyright (c) 2003-2006, Jouni Malinen <jkmaline@cc.hut.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
@@ -15,20 +15,13 @@
  * Not used in production version.
  */
 
-#include <stdio.h>
-#include <stdlib.h>
-#include <stdarg.h>
-#include <unistd.h>
-#include <ctype.h>
-#include <string.h>
-#include <signal.h>
-#include <netinet/in.h>
+#include "includes.h"
 #include <assert.h>
-#include <arpa/inet.h>
 
 #include "common.h"
 #include "config.h"
 #include "eapol_sm.h"
+#include "eap.h"
 #include "eloop.h"
 #include "wpa.h"
 #include "eap_i.h"
@@ -44,6 +37,8 @@
 extern int wpa_debug_level;
 extern int wpa_debug_show_keys;
 
+struct wpa_driver_ops *wpa_supplicant_drivers[] = { NULL };
+
 
 struct eapol_test_data {
 	struct wpa_supplicant *wpa_s;
@@ -70,6 +65,9 @@ struct eapol_test_data {
 
 	u8 *eap_identity;
 	size_t eap_identity_len;
+
+	char *connect_info;
+	u8 own_addr[ETH_ALEN];
 };
 
 static struct eapol_test_data eapol_test;
@@ -78,135 +76,7 @@ static struct eapol_test_data eapol_test;
 static void send_eap_request_identity(void *eloop_ctx, void *timeout_ctx);
 
 
-void wpa_msg(struct wpa_supplicant *wpa_s, int level, char *fmt, ...)
-{
-	va_list ap;
-	char *buf;
-	const int buflen = 2048;
-	int len;
-
-	buf = malloc(buflen);
-	if (buf == NULL) {
-		printf("Failed to allocate message buffer for:\n");
-		va_start(ap, fmt);
-		vprintf(fmt, ap);
-		printf("\n");
-		va_end(ap);
-		return;
-	}
-	va_start(ap, fmt);
-	len = vsnprintf(buf, buflen, fmt, ap);
-	va_end(ap);
-	wpa_printf(level, "%s", buf);
-	wpa_supplicant_ctrl_iface_send(wpa_s, level, buf, len);
-	free(buf);
-}
-
-
-const char * wpa_supplicant_state_txt(int state)
-{
-	switch (state) {
-	case WPA_DISCONNECTED:
-		return "DISCONNECTED";
-	case WPA_SCANNING:
-		return "SCANNING";
-	case WPA_ASSOCIATING:
-		return "ASSOCIATING";
-	case WPA_ASSOCIATED:
-		return "ASSOCIATED";
-	case WPA_4WAY_HANDSHAKE:
-		return "4WAY_HANDSHAKE";
-	case WPA_GROUP_HANDSHAKE:
-		return "GROUP_HANDSHAKE";
-	case WPA_COMPLETED:
-		return "COMPLETED";
-	default:
-		return "UNKNOWN";
-	}
-}
-
-
-int rsn_preauth_init(struct wpa_sm *sm, const u8 *dst, struct wpa_ssid *config)
-{
-	return -1;
-}
-
-
-void rsn_preauth_deinit(struct wpa_sm *sm)
-{
-}
-
-
-int rsn_preauth_get_status(struct wpa_sm *sm, char *buf, size_t buflen,
-			   int verbose)
-{
-	return 0;
-}
-
-
-int pmksa_cache_list(struct wpa_supplicant *wpa_s, char *buf, size_t len)
-{
-	return 0;
-}
-
-
-int wpa_sm_get_mib(struct wpa_sm *sm, char *buf, size_t buflen)
-{
-	return 0;
-}
-
-
-int wpa_sm_set_param(struct wpa_sm *sm, enum wpa_sm_conf_params param,
-		     unsigned int value)
-{
-	return 0;
-}
-
-
-int wpa_sm_get_status(struct wpa_sm *sm, char *buf, size_t buflen,
-		      int verbose)
-{
-	return 0;
-}
-
-
-void wpa_supplicant_req_scan(struct wpa_supplicant *wpa_s, int sec, int usec)
-{
-}
-
-
-const char * wpa_ssid_txt(u8 *ssid, size_t ssid_len)
-{
-	return NULL;
-}
-
-
-int wpa_supplicant_reload_configuration(struct wpa_supplicant *wpa_s)
-{
-	return -1;
-}
-
-
-int wpa_parse_wpa_ie(const u8 *wpa_ie, size_t wpa_ie_len,
-		     struct wpa_ie_data *data)
-{
-	return -1;
-}
-
-
-void wpa_supplicant_disassociate(struct wpa_supplicant *wpa_s,
-				 int reason_code)
-{
-}
-
-
-int wpa_supplicant_get_scan_results(struct wpa_supplicant *wpa_s)
-{
-	return -1;
-}
-
-
-void hostapd_logger(void *ctx, u8 *addr, unsigned int module, int level,
+void hostapd_logger(void *ctx, const u8 *addr, unsigned int module, int level,
 		    char *fmt, ...)
 {
 	char *format;
@@ -259,12 +129,12 @@ const char * hostapd_ip_txt(const struct hostapd_ip_addr *addr, char *buf,
 
 
 static void ieee802_1x_encapsulate_radius(struct eapol_test_data *e,
-					  u8 *eap, size_t len)
+					  const u8 *eap, size_t len)
 {
 	struct radius_msg *msg;
 	char buf[128];
-	struct eap_hdr *hdr;
-	u8 *pos;
+	const struct eap_hdr *hdr;
+	const u8 *pos;
 
 	wpa_printf(MSG_DEBUG, "Encapsulating EAP message into a RADIUS "
 		   "packet");
@@ -279,8 +149,8 @@ static void ieee802_1x_encapsulate_radius(struct eapol_test_data *e,
 
 	radius_msg_make_authenticator(msg, (u8 *) e, sizeof(*e));
 
-	hdr = (struct eap_hdr *) eap;
-	pos = (u8 *) (hdr + 1);
+	hdr = (const struct eap_hdr *) eap;
+	pos = (const u8 *) (hdr + 1);
 	if (len > sizeof(*hdr) && hdr->code == EAP_CODE_RESPONSE &&
 	    pos[0] == EAP_TYPE_IDENTITY) {
 		pos++;
@@ -330,7 +200,7 @@ static void ieee802_1x_encapsulate_radius(struct eapol_test_data *e,
 		goto fail;
 	}
 
-	snprintf(buf, sizeof(buf), "CONNECT 11Mbps 802.11b");
+	snprintf(buf, sizeof(buf), "%s", e->connect_info);
 	if (!radius_msg_add_attr(msg, RADIUS_ATTR_CONNECT_INFO,
 				 (u8 *) buf, strlen(buf))) {
 		printf("Could not add Connect-Info\n");
@@ -368,7 +238,8 @@ static void ieee802_1x_encapsulate_radius(struct eapol_test_data *e,
 }
 
 
-static int eapol_test_eapol_send(void *ctx, int type, u8 *buf, size_t len)
+static int eapol_test_eapol_send(void *ctx, int type, const u8 *buf,
+				 size_t len)
 {
 	/* struct wpa_supplicant *wpa_s = ctx; */
 	printf("WPA: eapol_test_eapol_send(type=%d len=%d)\n", type, len);
@@ -380,6 +251,22 @@ static int eapol_test_eapol_send(void *ctx, int type, u8 *buf, size_t len)
 }
 
 
+static void eapol_test_set_config_blob(void *ctx,
+				       struct wpa_config_blob *blob)
+{
+	struct wpa_supplicant *wpa_s = ctx;
+	wpa_config_set_blob(wpa_s->conf, blob);
+}
+
+
+static const struct wpa_config_blob *
+eapol_test_get_config_blob(void *ctx, const char *name)
+{
+	struct wpa_supplicant *wpa_s = ctx;
+	return wpa_config_get_blob(wpa_s->conf, name);
+}
+
+
 static void eapol_test_eapol_done_cb(void *ctx)
 {
 	printf("WPA: EAPOL processing complete\n");
@@ -402,23 +289,27 @@ static int eapol_test_compare_pmk(struct eapol_test_data *e)
 
 	if (eapol_sm_get_key(e->wpa_s->eapol, pmk, PMK_LEN) == 0) {
 		wpa_hexdump(MSG_DEBUG, "PMK from EAPOL", pmk, PMK_LEN);
-		if (memcmp(pmk, e->authenticator_pmk, PMK_LEN) != 0)
+		if (memcmp(pmk, e->authenticator_pmk, PMK_LEN) != 0) {
 			printf("WARNING: PMK mismatch\n");
-		else if (e->radius_access_accept_received)
+			wpa_hexdump(MSG_DEBUG, "PMK from AS",
+				    e->authenticator_pmk, PMK_LEN);
+		} else if (e->radius_access_accept_received)
 			ret = 0;
 	} else if (e->authenticator_pmk_len == 16 &&
 		   eapol_sm_get_key(e->wpa_s->eapol, pmk, 16) == 0) {
 		wpa_hexdump(MSG_DEBUG, "LEAP PMK from EAPOL", pmk, 16);
-		if (memcmp(pmk, e->authenticator_pmk, 16) != 0)
+		if (memcmp(pmk, e->authenticator_pmk, 16) != 0) {
 			printf("WARNING: PMK mismatch\n");
-		else if (e->radius_access_accept_received)
+			wpa_hexdump(MSG_DEBUG, "PMK from AS",
+				    e->authenticator_pmk, 16);
+		} else if (e->radius_access_accept_received)
 			ret = 0;
 	} else if (e->radius_access_accept_received && e->no_mppe_keys) {
 		/* No keying material expected */
 		ret = 0;
 	}
 
-	if (ret)
+	if (ret && !e->no_mppe_keys)
 		e->num_mppe_mismatch++;
 	else if (!e->no_mppe_keys)
 		e->num_mppe_ok++;
@@ -447,12 +338,11 @@ static int test_eapol(struct eapol_test_data *e, struct wpa_supplicant *wpa_s,
 	struct eapol_config eapol_conf;
 	struct eapol_ctx *ctx;
 
-	ctx = malloc(sizeof(*ctx));
+	ctx = wpa_zalloc(sizeof(*ctx));
 	if (ctx == NULL) {
 		printf("Failed to allocate EAPOL context.\n");
 		return -1;
 	}
-	memset(ctx, 0, sizeof(*ctx));
 	ctx->ctx = wpa_s;
 	ctx->msg_ctx = wpa_s;
 	ctx->scard_ctx = wpa_s->scard;
@@ -462,6 +352,8 @@ static int test_eapol(struct eapol_test_data *e, struct wpa_supplicant *wpa_s,
 	ctx->preauth = 0;
 	ctx->eapol_done_cb = eapol_test_eapol_done_cb;
 	ctx->eapol_send = eapol_test_eapol_send;
+	ctx->set_config_blob = eapol_test_set_config_blob;
+	ctx->get_config_blob = eapol_test_get_config_blob;
 	ctx->opensc_engine_path = wpa_s->conf->opensc_engine_path;
 	ctx->pkcs11_engine_path = wpa_s->conf->pkcs11_engine_path;
 	ctx->pkcs11_module_path = wpa_s->conf->pkcs11_module_path;
@@ -511,7 +403,10 @@ static void test_eapol_clean(struct eapol_test_data *e,
 	free(e->radius_conf);
 	e->radius_conf = NULL;
 	scard_deinit(wpa_s->scard);
-	wpa_supplicant_ctrl_iface_deinit(wpa_s);
+	if (wpa_s->ctrl_iface) {
+		wpa_supplicant_ctrl_iface_deinit(wpa_s->ctrl_iface);
+		wpa_s->ctrl_iface = NULL;
+	}
 	wpa_config_free(wpa_s->conf);
 }
 
@@ -710,7 +605,7 @@ ieee802_1x_receive_auth(struct radius_msg *msg, struct radius_msg *req,
 			      "Access-Reject without Message-Authenticator "
 			      "since it does not include EAP-Message\n");
 	} else if (radius_msg_verify(msg, shared_secret, shared_secret_len,
-				     req)) {
+				     req, 1)) {
 		printf("Incoming RADIUS packet did not have correct "
 		       "Message-Authenticator - dropped\n");
 		return RADIUS_RX_UNKNOWN;
@@ -756,25 +651,6 @@ ieee802_1x_receive_auth(struct radius_msg *msg, struct radius_msg *req,
 }
 
 
-static int wpa_supplicant_scard_init(struct wpa_supplicant *wpa_s,
-				     struct wpa_ssid *ssid)
-{
-	if (ssid->pcsc == NULL || wpa_s->scard != NULL)
-		return 0;
-	wpa_printf(MSG_DEBUG, "Selected network is configured to use SIM - "
-		   "initialize PCSC");
-	wpa_s->scard = scard_init(SCARD_TRY_BOTH);
-	if (wpa_s->scard == NULL) {
-		wpa_printf(MSG_WARNING, "Failed to initialize SIM "
-			   "(pcsc-lite)");
-		return -1;
-	}
-	eapol_sm_register_scard_ctx(wpa_s->eapol, wpa_s->scard);
-
-	return 0;
-}
-
-
 static void wpa_init_conf(struct eapol_test_data *e,
 			  struct wpa_supplicant *wpa_s, const char *authsrv,
 			  int port, const char *secret)
@@ -783,18 +659,29 @@ static void wpa_init_conf(struct eapol_test_data *e,
 	int res;
 
 	wpa_s->bssid[5] = 1;
-	wpa_s->own_addr[5] = 2;
+	memcpy(wpa_s->own_addr, e->own_addr, ETH_ALEN);
 	e->own_ip_addr.s_addr = htonl((127 << 24) | 1);
 	strncpy(wpa_s->ifname, "test", sizeof(wpa_s->ifname));
 
-	e->radius_conf = malloc(sizeof(struct hostapd_radius_servers));
+	e->radius_conf = wpa_zalloc(sizeof(struct hostapd_radius_servers));
 	assert(e->radius_conf != NULL);
-	memset(e->radius_conf, 0, sizeof(struct hostapd_radius_servers));
 	e->radius_conf->num_auth_servers = 1;
-	as = malloc(sizeof(struct hostapd_radius_server));
+	as = wpa_zalloc(sizeof(struct hostapd_radius_server));
 	assert(as != NULL);
-	memset(as, 0, sizeof(*as));
+#if defined(CONFIG_NATIVE_WINDOWS) || defined(CONFIG_ANSI_C_EXTRA)
+	{
+		int a[4];
+		u8 *pos;
+		sscanf(authsrv, "%d.%d.%d.%d", &a[0], &a[1], &a[2], &a[3]);
+		pos = (u8 *) &as->addr.u.v4;
+		*pos++ = a[0];
+		*pos++ = a[1];
+		*pos++ = a[2];
+		*pos++ = a[3];
+	}
+#else /* CONFIG_NATIVE_WINDOWS or CONFIG_ANSI_C_EXTRA */
 	inet_aton(authsrv, &as->addr.u.v4);
+#endif /* CONFIG_NATIVE_WINDOWS or CONFIG_ANSI_C_EXTRA */
 	as->addr.af = AF_INET;
 	as->port = port;
 	as->shared_secret = (u8 *) strdup(secret);
@@ -826,7 +713,8 @@ static int scard_test(void)
 	unsigned char rand_[num_triplets][16];
 	unsigned char sres_[num_triplets][4];
 	unsigned char kc_[num_triplets][8];
-	int i, j, res;
+	int i, res;
+	size_t j;
 
 #define AKA_RAND_LEN 16
 #define AKA_AUTN_LEN 16
@@ -926,7 +814,8 @@ static int scard_get_triplets(int argc, char *argv[])
 	unsigned char sres[4];
 	unsigned char kc[8];
 	int num_triplets;
-	int i, j;
+	int i;
+	size_t j;
 
 	if (argc < 2 || ((num_triplets = atoi(argv[1])) <= 0)) {
 		printf("invalid parameters for sim command\n");
@@ -993,12 +882,14 @@ static void eapol_test_terminate(int sig, void *eloop_ctx,
 static void usage(void)
 {
 	printf("usage:\n"
-	       "eapol_test [-nW] -c<conf> [-a<AS IP>] [-p<AS port>] "
-	       "[-s<AS secret>] [-r<count>]\n"
+	       "eapol_test [-nWS] -c<conf> [-a<AS IP>] [-p<AS port>] "
+	       "[-s<AS secret>] \\\n"
+	       "           [-r<count>] [-t<timeout>] [-C<Connect-Info>] \\\n"
+	       "           [-M<client MAC address>]\n"
 	       "eapol_test scard\n"
 	       "eapol_test sim <PIN> <num triplets> [debug]\n"
-	       "\n"
-	       "options:\n"
+	       "\n");
+	printf("options:\n"
 	       "  -c<conf> = configuration file\n"
 	       "  -a<AS IP> = IP address of the authentication server, "
 	       "default 127.0.0.1\n"
@@ -1008,26 +899,39 @@ static void usage(void)
 	       "server, default 'radius'\n"
 	       "  -r<count> = number of re-authentications\n"
 	       "  -W = wait for a control interface monitor before starting\n"
-	       "  -n = no MPPE keys expected\n");
+	       "  -S = save configuration after authentiation\n"
+	       "  -n = no MPPE keys expected\n"
+	       "  -t<timeout> = sets timeout in seconds (default: 30 s)\n"
+	       "  -C<Connect-Info> = RADIUS Connect-Info (default: "
+	       "CONNECT 11Mbps 802.11b)\n"
+	       "  -M<client MAC address> = Set own MAC address "
+	       "(Calling-Station-Id,\n"
+	       "                           default: 02:00:00:00:00:01)\n");
 }
 
 
 int main(int argc, char *argv[])
 {
 	struct wpa_supplicant wpa_s;
-	int c, ret = 1, wait_for_monitor = 0;
+	int c, ret = 1, wait_for_monitor = 0, save_config = 0;
 	char *as_addr = "127.0.0.1";
 	int as_port = 1812;
 	char *as_secret = "radius";
 	char *conf = NULL;
+	int timeout = 30;
+
+	if (os_program_init())
+		return -1;
 
 	memset(&eapol_test, 0, sizeof(eapol_test));
+	eapol_test.connect_info = "CONNECT 11Mbps 802.11b";
+	memcpy(eapol_test.own_addr, "\x02\x00\x00\x00\x00\x01", ETH_ALEN);
 
 	wpa_debug_level = 0;
 	wpa_debug_show_keys = 1;
 
 	for (;;) {
-		c = getopt(argc, argv, "a:c:np:r:s:W");
+		c = getopt(argc, argv, "a:c:C:M:np:r:s:St:W");
 		if (c < 0)
 			break;
 		switch (c) {
@@ -1037,6 +941,15 @@ int main(int argc, char *argv[])
 		case 'c':
 			conf = optarg;
 			break;
+		case 'C':
+			eapol_test.connect_info = optarg;
+			break;
+		case 'M':
+			if (hwaddr_aton(optarg, eapol_test.own_addr)) {
+				usage();
+				return -1;
+			}
+			break;
 		case 'n':
 			eapol_test.no_mppe_keys++;
 			break;
@@ -1049,6 +962,12 @@ int main(int argc, char *argv[])
 		case 's':
 			as_secret = optarg;
 			break;
+		case 'S':
+			save_config++;
+			break;
+		case 't':
+			timeout = atoi(optarg);
+			break;
 		case 'W':
 			wait_for_monitor++;
 			break;
@@ -1073,7 +992,15 @@ int main(int argc, char *argv[])
 		return -1;
 	}
 
-	eloop_init(&wpa_s);
+	if (eap_peer_register_methods()) {
+		wpa_printf(MSG_ERROR, "Failed to register EAP methods");
+		return -1;
+	}
+
+	if (eloop_init(&wpa_s)) {
+		wpa_printf(MSG_ERROR, "Failed to initialize event loop");
+		return -1;
+	}
 
 	memset(&wpa_s, 0, sizeof(wpa_s));
 	eapol_test.wpa_s = &wpa_s;
@@ -1088,7 +1015,8 @@ int main(int argc, char *argv[])
 	}
 
 	wpa_init_conf(&eapol_test, &wpa_s, as_addr, as_port, as_secret);
-	if (wpa_supplicant_ctrl_iface_init(&wpa_s)) {
+	wpa_s.ctrl_iface = wpa_supplicant_ctrl_iface_init(&wpa_s);
+	if (wpa_s.ctrl_iface == NULL) {
 		printf("Failed to initialize control interface '%s'.\n"
 		       "You may have another eapol_test process already "
 		       "running or the file was\n"
@@ -1106,24 +1034,30 @@ int main(int argc, char *argv[])
 		return -1;
 
 	if (wait_for_monitor)
-		wpa_supplicant_ctrl_iface_wait(&wpa_s);
+		wpa_supplicant_ctrl_iface_wait(wpa_s.ctrl_iface);
 
-	eloop_register_timeout(30, 0, eapol_test_timeout, &eapol_test, NULL);
+	eloop_register_timeout(timeout, 0, eapol_test_timeout, &eapol_test,
+			       NULL);
 	eloop_register_timeout(0, 0, send_eap_request_identity, &wpa_s, NULL);
-	eloop_register_signal(SIGINT, eapol_test_terminate, NULL);
-	eloop_register_signal(SIGTERM, eapol_test_terminate, NULL);
-	eloop_register_signal(SIGHUP, eapol_test_terminate, NULL);
+	eloop_register_signal_terminate(eapol_test_terminate, NULL);
+	eloop_register_signal_reconfig(eapol_test_terminate, NULL);
 	eloop_run();
 
-	if (eapol_test_compare_pmk(&eapol_test) == 0)
+	if (eapol_test_compare_pmk(&eapol_test) == 0 ||
+	    eapol_test.no_mppe_keys)
 		ret = 0;
 	if (eapol_test.auth_timed_out)
 		ret = -2;
 	if (eapol_test.radius_access_reject_received)
 		ret = -3;
 
+	if (save_config)
+		wpa_config_write(conf, wpa_s.conf);
+
 	test_eapol_clean(&eapol_test, &wpa_s);
 
+	eap_peer_unregister_methods();
+
 	eloop_destroy();
 
 	printf("MPPE keys OK: %d  mismatch: %d\n",
@@ -1135,5 +1069,7 @@ int main(int argc, char *argv[])
 	else
 		printf("SUCCESS\n");
 
+	os_program_deinit();
+
 	return ret;
 }
