@@ -1,6 +1,6 @@
 /*
  * WPA Supplicant - Driver event processing
- * Copyright (c) 2003-2005, Jouni Malinen <jkmaline@cc.hut.fi>
+ * Copyright (c) 2003-2006, Jouni Malinen <jkmaline@cc.hut.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
@@ -12,11 +12,7 @@
  * See README and COPYING for more details.
  */
 
-#include <stdlib.h>
-#include <stdio.h>
-#include <string.h>
-#include <unistd.h>
-#include <time.h>
+#include "includes.h"
 
 #include "common.h"
 #include "eapol_sm.h"
@@ -28,7 +24,9 @@
 #include "wpa_supplicant_i.h"
 #include "pcsc_funcs.h"
 #include "preauth.h"
+#include "pmksa_cache.h"
 #include "wpa_ctrl.h"
+#include "eap.h"
 
 
 static int wpa_supplicant_select_config(struct wpa_supplicant *wpa_s)
@@ -88,6 +86,7 @@ static void wpa_supplicant_mark_disassoc(struct wpa_supplicant *wpa_s)
 {
 	wpa_supplicant_set_state(wpa_s, WPA_DISCONNECTED);
 	memset(wpa_s->bssid, 0, ETH_ALEN);
+	memset(wpa_s->pending_bssid, 0, ETH_ALEN);
 	eapol_sm_notify_portEnabled(wpa_s->eapol, FALSE);
 	eapol_sm_notify_portValid(wpa_s->eapol, FALSE);
 	if (wpa_s->key_mgmt == WPA_KEY_MGMT_PSK)
@@ -144,6 +143,7 @@ static int wpa_supplicant_dynamic_keys(struct wpa_supplicant *wpa_s)
 	    wpa_s->key_mgmt == WPA_KEY_MGMT_WPA_NONE)
 		return 0;
 
+#ifdef IEEE8021X_EAPOL
 	if (wpa_s->key_mgmt == WPA_KEY_MGMT_IEEE8021X_NO_WPA &&
 	    wpa_s->current_ssid &&
 	    !(wpa_s->current_ssid->eapol_flags &
@@ -153,19 +153,68 @@ static int wpa_supplicant_dynamic_keys(struct wpa_supplicant *wpa_s)
 		 * plaintext or static WEP keys). */
 		return 0;
 	}
+#endif /* IEEE8021X_EAPOL */
 
 	return 1;
 }
 
 
-static int wpa_supplicant_scard_init(struct wpa_supplicant *wpa_s,
-				     struct wpa_ssid *ssid)
+/**
+ * wpa_supplicant_scard_init - Initialize SIM/USIM access with PC/SC
+ * @wpa_s: pointer to wpa_supplicant data
+ * @ssid: Configuration data for the network
+ * Returns: 0 on success, -1 on failure
+ *
+ * This function is called when starting authentication with a network that is
+ * configured to use PC/SC for SIM/USIM access (EAP-SIM or EAP-AKA).
+ */
+int wpa_supplicant_scard_init(struct wpa_supplicant *wpa_s,
+			      struct wpa_ssid *ssid)
 {
+#ifdef IEEE8021X_EAPOL
+	int aka = 0, sim = 0, type;
+
 	if (ssid->pcsc == NULL || wpa_s->scard != NULL)
 		return 0;
-	wpa_printf(MSG_DEBUG, "Selected network is configured to use SIM - "
-		   "initialize PCSC");
-	wpa_s->scard = scard_init(SCARD_TRY_BOTH);
+
+	if (ssid->eap_methods == NULL) {
+		sim = 1;
+		aka = 1;
+	} else {
+		struct eap_method_type *eap = ssid->eap_methods;
+		while (eap->vendor != EAP_VENDOR_IETF ||
+		       eap->method != EAP_TYPE_NONE) {
+			if (eap->vendor == EAP_VENDOR_IETF) {
+				if (eap->method == EAP_TYPE_SIM)
+					sim = 1;
+				else if (eap->method == EAP_TYPE_AKA)
+					aka = 1;
+			}
+			eap++;
+		}
+	}
+
+	if (eap_sm_get_eap_methods(EAP_VENDOR_IETF, EAP_TYPE_SIM) == NULL)
+		sim = 0;
+	if (eap_sm_get_eap_methods(EAP_VENDOR_IETF, EAP_TYPE_AKA) == NULL)
+		aka = 0;
+
+	if (!sim && !aka) {
+		wpa_printf(MSG_DEBUG, "Selected network is configured to use "
+			   "SIM, but neither EAP-SIM nor EAP-AKA are enabled");
+		return 0;
+	}
+
+	wpa_printf(MSG_DEBUG, "Selected network is configured to use SIM "
+		   "(sim=%d aka=%d) - initialize PCSC", sim, aka);
+	if (sim && aka)
+		type = SCARD_TRY_BOTH;
+	else if (aka)
+		type = SCARD_USIM_ONLY;
+	else
+		type = SCARD_GSM_SIM_ONLY;
+
+	wpa_s->scard = scard_init(type);
 	if (wpa_s->scard == NULL) {
 		wpa_printf(MSG_WARNING, "Failed to initialize SIM "
 			   "(pcsc-lite)");
@@ -173,6 +222,7 @@ static int wpa_supplicant_scard_init(struct wpa_supplicant *wpa_s,
 	}
 	wpa_sm_set_scard_ctx(wpa_s->wpa, wpa_s->scard);
 	eapol_sm_register_scard_ctx(wpa_s->eapol, wpa_s->scard);
+#endif /* IEEE8021X_EAPOL */
 
 	return 0;
 }
@@ -189,10 +239,12 @@ static int wpa_supplicant_match_privacy(struct wpa_scan_result *bss,
 			break;
 		}
 	}
+#ifdef IEEE8021X_EAPOL
 	if ((ssid->key_mgmt & WPA_KEY_MGMT_IEEE8021X_NO_WPA) &&
 	    ssid->eapol_flags & (EAPOL_FLAG_REQUIRE_KEY_UNICAST |
 				 EAPOL_FLAG_REQUIRE_KEY_BROADCAST))
 		privacy = 1;
+#endif /* IEEE8021X_EAPOL */
 
 	if (bss->caps & IEEE80211_CAP_PRIVACY)
 		return privacy;
@@ -200,6 +252,89 @@ static int wpa_supplicant_match_privacy(struct wpa_scan_result *bss,
 }
 
 
+static int wpa_supplicant_ssid_bss_match(struct wpa_ssid *ssid,
+					 struct wpa_scan_result *bss)
+{
+	struct wpa_ie_data ie;
+	int proto_match = 0;
+
+	while ((ssid->proto & WPA_PROTO_RSN) && bss->rsn_ie_len > 0) {
+		proto_match++;
+
+		if (wpa_parse_wpa_ie(bss->rsn_ie, bss->rsn_ie_len, &ie)) {
+			wpa_printf(MSG_DEBUG, "   skip RSN IE - parse failed");
+			break;
+		}
+		if (!(ie.proto & ssid->proto)) {
+			wpa_printf(MSG_DEBUG, "   skip RSN IE - proto "
+				   "mismatch");
+			break;
+		}
+
+		if (!(ie.pairwise_cipher & ssid->pairwise_cipher)) {
+			wpa_printf(MSG_DEBUG, "   skip RSN IE - PTK cipher "
+				   "mismatch");
+			break;
+		}
+
+		if (!(ie.group_cipher & ssid->group_cipher)) {
+			wpa_printf(MSG_DEBUG, "   skip RSN IE - GTK cipher "
+				   "mismatch");
+			break;
+		}
+
+		if (!(ie.key_mgmt & ssid->key_mgmt)) {
+			wpa_printf(MSG_DEBUG, "   skip RSN IE - key mgmt "
+				   "mismatch");
+			break;
+		}
+
+		wpa_printf(MSG_DEBUG, "   selected based on RSN IE");
+		return 1;
+	}
+
+	while ((ssid->proto & WPA_PROTO_WPA) && bss->wpa_ie_len > 0) {
+		proto_match++;
+
+		if (wpa_parse_wpa_ie(bss->wpa_ie, bss->wpa_ie_len, &ie)) {
+			wpa_printf(MSG_DEBUG, "   skip WPA IE - parse failed");
+			break;
+		}
+		if (!(ie.proto & ssid->proto)) {
+			wpa_printf(MSG_DEBUG, "   skip WPA IE - proto "
+				   "mismatch");
+			break;
+		}
+
+		if (!(ie.pairwise_cipher & ssid->pairwise_cipher)) {
+			wpa_printf(MSG_DEBUG, "   skip WPA IE - PTK cipher "
+				   "mismatch");
+			break;
+		}
+
+		if (!(ie.group_cipher & ssid->group_cipher)) {
+			wpa_printf(MSG_DEBUG, "   skip WPA IE - GTK cipher "
+				   "mismatch");
+			break;
+		}
+
+		if (!(ie.key_mgmt & ssid->key_mgmt)) {
+			wpa_printf(MSG_DEBUG, "   skip WPA IE - key mgmt "
+				   "mismatch");
+			break;
+		}
+
+		wpa_printf(MSG_DEBUG, "   selected based on WPA IE");
+		return 1;
+	}
+
+	if (proto_match == 0)
+		wpa_printf(MSG_DEBUG, "   skip - no WPA/RSN proto match");
+
+	return 0;
+}
+
+
 static struct wpa_scan_result *
 wpa_supplicant_select_bss(struct wpa_supplicant *wpa_s, struct wpa_ssid *group,
 			  struct wpa_scan_result *results, int num,
@@ -224,8 +359,8 @@ wpa_supplicant_select_bss(struct wpa_supplicant *wpa_s, struct wpa_ssid *group,
 			   wpa_ssid_txt(bss->ssid, bss->ssid_len),
 			   (unsigned long) bss->wpa_ie_len,
 			   (unsigned long) bss->rsn_ie_len, bss->caps);
-		if ((e = wpa_blacklist_get(wpa_s, bss->bssid)) &&
-		    e->count > 1) {
+		e = wpa_blacklist_get(wpa_s, bss->bssid);
+		if (e && e->count > 1) {
 			wpa_printf(MSG_DEBUG, "   skip - blacklisted");
 			continue;
 		}
@@ -236,7 +371,6 @@ wpa_supplicant_select_bss(struct wpa_supplicant *wpa_s, struct wpa_ssid *group,
 		}
 
 		for (ssid = group; ssid; ssid = ssid->pnext) {
-			struct wpa_ie_data ie;
 			if (ssid->disabled)
 				continue;
 			if (bss->ssid_len != ssid->ssid_len ||
@@ -252,41 +386,11 @@ wpa_supplicant_select_bss(struct wpa_supplicant *wpa_s, struct wpa_ssid *group,
 					   "BSSID mismatch");
 				continue;
 			}
-			if (!(((ssid->proto & WPA_PROTO_RSN) &&
-			       wpa_parse_wpa_ie(bss->rsn_ie, bss->rsn_ie_len,
-						&ie) == 0) ||
-			      ((ssid->proto & WPA_PROTO_WPA) &&
-			       wpa_parse_wpa_ie(bss->wpa_ie, bss->wpa_ie_len,
-						&ie) == 0))) {
-				wpa_printf(MSG_DEBUG, "   skip - "
-					   "could not parse WPA/RSN IE");
-				continue;
+			if (wpa_supplicant_ssid_bss_match(ssid, bss)) {
+				selected = bss;
+				*selected_ssid = ssid;
+				break;
 			}
-			if (!(ie.proto & ssid->proto)) {
-				wpa_printf(MSG_DEBUG, "   skip - "
-					   "proto mismatch");
-				continue;
-			}
-			if (!(ie.pairwise_cipher & ssid->pairwise_cipher)) {
-				wpa_printf(MSG_DEBUG, "   skip - "
-					   "PTK cipher mismatch");
-				continue;
-			}
-			if (!(ie.group_cipher & ssid->group_cipher)) {
-				wpa_printf(MSG_DEBUG, "   skip - "
-					   "GTK cipher mismatch");
-				continue;
-			}
-			if (!(ie.key_mgmt & ssid->key_mgmt)) {
-				wpa_printf(MSG_DEBUG, "   skip - "
-					   "key mgmt mismatch");
-				continue;
-			}
-
-			selected = bss;
-			*selected_ssid = ssid;
-			wpa_printf(MSG_DEBUG, "   selected");
-			break;
 		}
 	}
 
@@ -294,8 +398,8 @@ wpa_supplicant_select_bss(struct wpa_supplicant *wpa_s, struct wpa_ssid *group,
 	 * allows this. */
 	for (i = 0; i < num && !selected; i++) {
 		bss = &results[i];
-		if ((e = wpa_blacklist_get(wpa_s, bss->bssid)) &&
-		    e->count > 1) {
+		e = wpa_blacklist_get(wpa_s, bss->bssid);
+		if (e && e->count > 1) {
 			continue;
 		}
 		for (ssid = group; ssid; ssid = ssid->pnext) {
@@ -330,7 +434,7 @@ wpa_supplicant_select_bss(struct wpa_supplicant *wpa_s, struct wpa_ssid *group,
 
 static void wpa_supplicant_event_scan_results(struct wpa_supplicant *wpa_s)
 {
-	int num, prio;
+	int num, prio, timeout;
 	struct wpa_scan_result *selected = NULL;
 	struct wpa_ssid *ssid = NULL;
 	struct wpa_scan_result *results;
@@ -340,8 +444,8 @@ static void wpa_supplicant_event_scan_results(struct wpa_supplicant *wpa_s)
 			return;
 		wpa_printf(MSG_DEBUG, "Failed to get scan results - try "
 			   "scanning again");
-		wpa_supplicant_req_scan(wpa_s, 1, 0);
-		return;
+		timeout = 1;
+		goto req_scan;
 	}
 	if (wpa_s->conf->ap_scan == 2)
 		return;
@@ -367,8 +471,15 @@ static void wpa_supplicant_event_scan_results(struct wpa_supplicant *wpa_s)
 	}
 
 	if (selected) {
+		/* Do not trigger new association unless the BSSID has changed
+		 * or if reassociation is requested. If we are in process of
+		 * associating with the selected BSSID, do not trigger new
+		 * attempt. */
 		if (wpa_s->reassociate ||
-		    memcmp(selected->bssid, wpa_s->bssid, ETH_ALEN) != 0) {
+		    (memcmp(selected->bssid, wpa_s->bssid, ETH_ALEN) != 0 &&
+		     (wpa_s->wpa_state != WPA_ASSOCIATING ||
+		      memcmp(selected->bssid, wpa_s->pending_bssid,
+			     ETH_ALEN) != 0))) {
 			if (wpa_supplicant_scard_init(wpa_s, ssid)) {
 				wpa_supplicant_req_scan(wpa_s, 10, 0);
 				return;
@@ -381,8 +492,22 @@ static void wpa_supplicant_event_scan_results(struct wpa_supplicant *wpa_s)
 		rsn_preauth_scan_results(wpa_s->wpa, results, num);
 	} else {
 		wpa_printf(MSG_DEBUG, "No suitable AP found.");
-		wpa_supplicant_req_scan(wpa_s, 5, 0);
+		timeout = 5;
+		goto req_scan;
 	}
+
+	return;
+
+req_scan:
+	if (wpa_s->scan_res_tried == 1 && wpa_s->conf->ap_scan == 1) {
+		/*
+		 * Quick recovery if the initial scan results were not
+		 * complete when fetched before the first scan request.
+		 */
+		wpa_s->scan_res_tried++;
+		timeout = 0;
+	}
+	wpa_supplicant_req_scan(wpa_s, timeout, 0);
 }
 
 
@@ -393,16 +518,22 @@ static void wpa_supplicant_event_associnfo(struct wpa_supplicant *wpa_s,
 	u8 *p;
 
 	wpa_printf(MSG_DEBUG, "Association info event");
-	wpa_hexdump(MSG_DEBUG, "req_ies", data->assoc_info.req_ies,
-		    data->assoc_info.req_ies_len);
-	wpa_hexdump(MSG_DEBUG, "resp_ies", data->assoc_info.resp_ies,
-		    data->assoc_info.resp_ies_len);
+	if (data->assoc_info.req_ies)
+		wpa_hexdump(MSG_DEBUG, "req_ies", data->assoc_info.req_ies,
+			    data->assoc_info.req_ies_len);
+	if (data->assoc_info.resp_ies)
+		wpa_hexdump(MSG_DEBUG, "resp_ies", data->assoc_info.resp_ies,
+			    data->assoc_info.resp_ies_len);
+	if (data->assoc_info.beacon_ies)
+		wpa_hexdump(MSG_DEBUG, "beacon_ies",
+			    data->assoc_info.beacon_ies,
+			    data->assoc_info.beacon_ies_len);
 
 	p = data->assoc_info.req_ies;
 	l = data->assoc_info.req_ies_len;
 
 	/* Go through the IEs and make a copy of the WPA/RSN IE, if present. */
-	while (l >= 2) {
+	while (p && l >= 2) {
 		len = p[1] + 2;
 		if (len > l) {
 			wpa_hexdump(MSG_DEBUG, "Truncated IE in assoc_info",
@@ -431,7 +562,7 @@ static void wpa_supplicant_event_associnfo(struct wpa_supplicant *wpa_s,
 	/* Go through the IEs and make a copy of the WPA/RSN IEs, if present.
 	 */
 	wpa_found = rsn_found = 0;
-	while (l >= 2) {
+	while (p && l >= 2) {
 		len = p[1] + 2;
 		if (len > l) {
 			wpa_hexdump(MSG_DEBUG, "Truncated IE in beacon_ies",
@@ -462,16 +593,24 @@ static void wpa_supplicant_event_associnfo(struct wpa_supplicant *wpa_s,
 }
 
 
-static void wpa_supplicant_event_assoc(struct wpa_supplicant *wpa_s)
+static void wpa_supplicant_event_assoc(struct wpa_supplicant *wpa_s,
+				       union wpa_event_data *data)
 {
 	u8 bssid[ETH_ALEN];
 
+	if (data)
+		wpa_supplicant_event_associnfo(wpa_s, data);
+
 	wpa_supplicant_set_state(wpa_s, WPA_ASSOCIATED);
-	if (wpa_drv_get_bssid(wpa_s, bssid) >= 0 &&
-	    memcmp(bssid, wpa_s->bssid, ETH_ALEN) != 0) {
+	if (wpa_s->use_client_mlme)
+		memcpy(bssid, wpa_s->bssid, ETH_ALEN);
+	if (wpa_s->use_client_mlme ||
+	    (wpa_drv_get_bssid(wpa_s, bssid) >= 0 &&
+	     memcmp(bssid, wpa_s->bssid, ETH_ALEN) != 0)) {
 		wpa_msg(wpa_s, MSG_DEBUG, "Associated to a new BSS: BSSID="
 			MACSTR, MAC2STR(bssid));
 		memcpy(wpa_s->bssid, bssid, ETH_ALEN);
+		memset(wpa_s->pending_bssid, 0, ETH_ALEN);
 		if (wpa_supplicant_dynamic_keys(wpa_s)) {
 			wpa_clear_keys(wpa_s, bssid);
 		}
@@ -483,6 +622,13 @@ static void wpa_supplicant_event_assoc(struct wpa_supplicant *wpa_s)
 	}
 
 	wpa_msg(wpa_s, MSG_INFO, "Associated with " MACSTR, MAC2STR(bssid));
+	if (wpa_s->current_ssid) {
+		/* When using scanning (ap_scan=1), SIM PC/SC interface can be
+		 * initialized before association, but for other modes,
+		 * initialize PC/SC here, if the current configuration needs
+		 * smartcard or SIM/USIM. */
+		wpa_supplicant_scard_init(wpa_s, wpa_s->current_ssid);
+	}
 	wpa_sm_notify_assoc(wpa_s->wpa, bssid);
 	l2_packet_notify_auth_start(wpa_s->l2);
 
@@ -514,6 +660,8 @@ static void wpa_supplicant_event_assoc(struct wpa_supplicant *wpa_s)
 
 static void wpa_supplicant_event_disassoc(struct wpa_supplicant *wpa_s)
 {
+	const u8 *bssid;
+
 	if (wpa_s->key_mgmt == WPA_KEY_MGMT_WPA_NONE) {
 		/*
 		 * At least Host AP driver and a Prism3 card seemed to be
@@ -532,7 +680,10 @@ static void wpa_supplicant_event_disassoc(struct wpa_supplicant *wpa_s)
 	}
 	if (wpa_s->wpa_state >= WPA_ASSOCIATED)
 		wpa_supplicant_req_scan(wpa_s, 0, 100000);
-	wpa_blacklist_add(wpa_s, wpa_s->bssid);
+	bssid = wpa_s->bssid;
+	if (memcmp(bssid, "\x00\x00\x00\x00\x00\x00", ETH_ALEN) == 0)
+		bssid = wpa_s->pending_bssid;
+	wpa_blacklist_add(wpa_s, bssid);
 	wpa_sm_notify_disassoc(wpa_s->wpa);
 	wpa_supplicant_mark_disassoc(wpa_s);
 	wpa_msg(wpa_s, MSG_INFO, WPA_EVENT_DISCONNECTED "- Disconnect event - "
@@ -549,14 +700,14 @@ wpa_supplicant_event_michael_mic_failure(struct wpa_supplicant *wpa_s,
 					 union wpa_event_data *data)
 {
 	int pairwise;
-	time_t now;
+	struct os_time t;
 
 	wpa_msg(wpa_s, MSG_WARNING, "Michael MIC failure detected");
 	pairwise = (data && data->michael_mic_failure.unicast);
 	wpa_sm_key_request(wpa_s->wpa, 1, pairwise);
-	time(&now);
+	os_get_time(&t);
 	if (wpa_s->last_michael_mic_error &&
-	    now - wpa_s->last_michael_mic_error <= 60) {
+	    t.sec - wpa_s->last_michael_mic_error <= 60) {
 		/* initialize countermeasures */
 		wpa_s->countermeasures = 1;
 		wpa_msg(wpa_s, MSG_WARNING, "TKIP countermeasures started");
@@ -565,7 +716,7 @@ wpa_supplicant_event_michael_mic_failure(struct wpa_supplicant *wpa_s,
 		 * Need to wait for completion of request frame. We do not get
 		 * any callback for the message completion, so just wait a
 		 * short while and hope for the best. */
-		usleep(10000);
+		os_sleep(0, 10000);
 
 		wpa_drv_set_countermeasures(wpa_s, 1);
 		wpa_supplicant_deauthenticate(wpa_s,
@@ -578,7 +729,7 @@ wpa_supplicant_event_michael_mic_failure(struct wpa_supplicant *wpa_s,
 		/* TODO: mark the AP rejected for 60 second. STA is
 		 * allowed to associate with another AP.. */
 	}
-	wpa_s->last_michael_mic_error = now;
+	wpa_s->last_michael_mic_error = t.sec;
 }
 
 
@@ -611,12 +762,36 @@ wpa_supplicant_event_interface_status(struct wpa_supplicant *wpa_s,
 }
 
 
+#ifdef CONFIG_STAKEY
+static void
+wpa_supplicant_event_stakey_request(struct wpa_supplicant *wpa_s,
+				    union wpa_event_data *data)
+{
+	if (data == NULL)
+		return;
+	wpa_sm_stakey_request(wpa_s->wpa, data->stakey_request.peer);
+}
+#endif /* CONFIG_STAKEY */
+
+
+#ifdef CONFIG_PEERKEY
+static void
+wpa_supplicant_event_stkstart(struct wpa_supplicant *wpa_s,
+			      union wpa_event_data *data)
+{
+	if (data == NULL)
+		return;
+	wpa_sm_stkstart(wpa_s->wpa, data->stkstart.peer);
+}
+#endif /* CONFIG_PEERKEY */
+
+
 void wpa_supplicant_event(struct wpa_supplicant *wpa_s, wpa_event_type event,
 			  union wpa_event_data *data)
 {
 	switch (event) {
 	case EVENT_ASSOC:
-		wpa_supplicant_event_assoc(wpa_s);
+		wpa_supplicant_event_assoc(wpa_s, data);
 		break;
 	case EVENT_DISASSOC:
 		wpa_supplicant_event_disassoc(wpa_s);
@@ -636,6 +811,16 @@ void wpa_supplicant_event(struct wpa_supplicant *wpa_s, wpa_event_type event,
 	case EVENT_PMKID_CANDIDATE:
 		wpa_supplicant_event_pmkid_candidate(wpa_s, data);
 		break;
+#ifdef CONFIG_STAKEY
+	case EVENT_STAKEY_REQUEST:
+		wpa_supplicant_event_stakey_request(wpa_s, data);
+		break;
+#endif /* CONFIG_STAKEY */
+#ifdef CONFIG_PEERKEY
+	case EVENT_STKSTART:
+		wpa_supplicant_event_stkstart(wpa_s, data);
+		break;
+#endif /* CONFIG_PEERKEY */
 	default:
 		wpa_printf(MSG_INFO, "Unknown event %d", event);
 		break;
