@@ -1,6 +1,6 @@
 /*
  * WPA Supplicant / SSL/TLS interface functions for openssl
- * Copyright (c) 2004-2005, Jouni Malinen <jkmaline@cc.hut.fi>
+ * Copyright (c) 2004-2006, Jouni Malinen <jkmaline@cc.hut.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
@@ -12,9 +12,7 @@
  * See README and COPYING for more details.
  */
 
-#include <stdlib.h>
-#include <stdio.h>
-#include <string.h>
+#include "includes.h"
 
 #ifndef CONFIG_SMARTCARD
 #ifndef OPENSSL_NO_ENGINE
@@ -33,6 +31,13 @@
 #include "common.h"
 #include "tls.h"
 
+#if OPENSSL_VERSION_NUMBER >= 0x0090800fL
+#define OPENSSL_d2i_TYPE const unsigned char **
+#else
+#define OPENSSL_d2i_TYPE unsigned char **
+#endif
+
+static int tls_openssl_ref_count = 0;
 
 struct tls_connection {
 	SSL *ssl;
@@ -49,6 +54,36 @@ struct tls_connection {
 };
 
 
+#ifdef CONFIG_NO_STDOUT_DEBUG
+
+static void _tls_show_errors(void)
+{
+	unsigned long err;
+
+	while ((err = ERR_get_error())) {
+		/* Just ignore the errors, since stdout is disabled */
+	}
+}
+#define tls_show_errors(l, f, t) _tls_show_errors()
+
+#else /* CONFIG_NO_STDOUT_DEBUG */
+
+static void tls_show_errors(int level, const char *func, const char *txt)
+{
+	unsigned long err;
+
+	wpa_printf(level, "OpenSSL: %s - %s %s",
+		   func, txt, ERR_error_string(ERR_get_error(), NULL));
+
+	while ((err = ERR_get_error())) {
+		wpa_printf(MSG_INFO, "OpenSSL: pending error: %s",
+			   ERR_error_string(err, NULL));
+	}
+}
+
+#endif /* CONFIG_NO_STDOUT_DEBUG */
+
+
 #ifdef CONFIG_NATIVE_WINDOWS
 
 /* Windows CryptoAPI and access to certificate stores */
@@ -71,6 +106,11 @@ static BOOL WINAPI
 				     DWORD *pdwKeySpec, BOOL *pfCallerFreeProv)
 = NULL; /* to be loaded from crypt32.dll */
 
+static PCCERT_CONTEXT WINAPI
+(*CertEnumCertificatesInStore)(HCERTSTORE hCertStore,
+			       PCCERT_CONTEXT pPrevCertContext)
+= NULL; /* to be loaded from crypt32.dll */
+
 static int mingw_load_crypto_func(void)
 {
 	HINSTANCE dll;
@@ -96,6 +136,16 @@ static int mingw_load_crypto_func(void)
 			   "crypt32 library");
 		return -1;
 	}
+
+	CertEnumCertificatesInStore = (void *) GetProcAddress(
+		dll, "CertEnumCertificatesInStore");
+	if (CertEnumCertificatesInStore == NULL) {
+		wpa_printf(MSG_DEBUG, "CryptoAPI: Could not get "
+			   "CertEnumCertificatesInStore() address from "
+			   "crypt32 library");
+		return -1;
+	}
+
 	return 0;
 }
 
@@ -183,7 +233,7 @@ static int cryptoapi_rsa_priv_enc(int flen, const unsigned char *from,
 		goto err;
 	}
 
-	if (hash_size != flen) {
+	if ((int) hash_size != flen) {
 		wpa_printf(MSG_INFO, "CryptoAPI: Invalid hash size (%u != %d)",
 			   (unsigned) hash_size, flen);
 		RSAerr(RSA_F_RSA_EAY_PRIVATE_ENCRYPT,
@@ -307,8 +357,8 @@ static int tls_cryptoapi_cert(SSL *ssl, const char *name)
 	     strncmp(name, "hash://", 7) != 0))
 		return -1;
 
-	priv = malloc(sizeof(*priv));
-	rsa_meth = malloc(sizeof(*rsa_meth));
+	priv = wpa_zalloc(sizeof(*priv));
+	rsa_meth = wpa_zalloc(sizeof(*rsa_meth));
 	if (priv == NULL || rsa_meth == NULL) {
 		wpa_printf(MSG_WARNING, "CryptoAPI: Failed to allocate memory "
 			   "for CryptoAPI RSA method");
@@ -316,8 +366,6 @@ static int tls_cryptoapi_cert(SSL *ssl, const char *name)
 		free(rsa_meth);
 		return -1;
 	}
-	memset(priv, 0, sizeof(*priv));
-	memset(rsa_meth, 0, sizeof(*rsa_meth));
 
 	priv->cert = cryptoapi_find_cert(name, CERT_SYSTEM_STORE_CURRENT_USER);
 	if (priv->cert == NULL) {
@@ -330,8 +378,7 @@ static int tls_cryptoapi_cert(SSL *ssl, const char *name)
 		goto err;
 	}
 
-	cert = d2i_X509(NULL,
-			(const unsigned char **) &priv->cert->pbCertEncoded,
+	cert = d2i_X509(NULL, (OPENSSL_d2i_TYPE) &priv->cert->pbCertEncoded,
 			priv->cert->cbCertEncoded);
 	if (cert == NULL) {
 		wpa_printf(MSG_INFO, "CryptoAPI: Could not process X509 DER "
@@ -397,6 +444,61 @@ err:
 	return -1;
 }
 
+
+static int tls_cryptoapi_ca_cert(SSL_CTX *ssl_ctx, SSL *ssl, const char *name)
+{
+	HCERTSTORE cs;
+	PCCERT_CONTEXT ctx = NULL;
+	X509 *cert;
+	char buf[128];
+
+	if (mingw_load_crypto_func())
+		return -1;
+
+	if (name == NULL || strncmp(name, "cert_store://", 13) != 0)
+		return -1;
+
+	cs = CertOpenSystemStore(0, name + 13);
+	if (cs == NULL) {
+		wpa_printf(MSG_DEBUG, "%s: failed to open system cert store "
+			   "'%s': error=%d", __func__, name + 13,
+			   (int) GetLastError());
+		return -1;
+	}
+
+	while ((ctx = CertEnumCertificatesInStore(cs, ctx))) {
+		cert = d2i_X509(NULL, (OPENSSL_d2i_TYPE) &ctx->pbCertEncoded,
+				ctx->cbCertEncoded);
+		if (cert == NULL) {
+			wpa_printf(MSG_INFO, "CryptoAPI: Could not process "
+				   "X509 DER encoding for CA cert");
+			continue;
+		}
+
+		X509_NAME_oneline(X509_get_subject_name(cert), buf,
+				  sizeof(buf));
+		wpa_printf(MSG_DEBUG, "OpenSSL: Loaded CA certificate for "
+			   "system certificate store: subject='%s'", buf);
+
+		if (!X509_STORE_add_cert(ssl_ctx->cert_store, cert)) {
+			tls_show_errors(MSG_WARNING, __func__,
+					"Failed to add ca_cert to OpenSSL "
+					"certificate store");
+		}
+
+		X509_free(cert);
+	}
+
+	if (!CertCloseStore(cs, 0)) {
+		wpa_printf(MSG_DEBUG, "%s: failed to close system cert store "
+			   "'%s': error=%d", __func__, name + 13,
+			   (int) GetLastError());
+	}
+
+	return 0;
+}
+
+
 #else /* CONFIG_NATIVE_WINDOWS */
 
 static int tls_cryptoapi_cert(SSL *ssl, const char *name)
@@ -537,21 +639,25 @@ static int tls_engine_load_dynamic_pkcs11(const char *pkcs11_so_path,
 {
 	char *engine_id = "pkcs11";
 	const char *pre_cmd[] = {
-		"SO_PATH", pkcs11_so_path,
-		"ID", engine_id,
+		"SO_PATH", NULL /* pkcs11_so_path */,
+		"ID", NULL /* engine_id */,
 		"LIST_ADD", "1",
 		/* "NO_VCHECK", "1", */
 		"LOAD", NULL,
 		NULL, NULL
 	};
 	const char *post_cmd[] = {
-		"MODULE_PATH", pkcs11_module_path,
+		"MODULE_PATH", NULL /* pkcs11_module_path */,
 		NULL, NULL
 	};
 
 	if (!pkcs11_so_path || !pkcs11_module_path)
 		return 0;
 
+	pre_cmd[1] = pkcs11_so_path;
+	pre_cmd[3] = engine_id;
+	post_cmd[1] = pkcs11_module_path;
+
 	wpa_printf(MSG_DEBUG, "ENGINE: Loading pkcs11 Engine from %s",
 		   pkcs11_so_path);
 
@@ -567,8 +673,8 @@ static int tls_engine_load_dynamic_opensc(const char *opensc_so_path)
 {
 	char *engine_id = "opensc";
 	const char *pre_cmd[] = {
-		"SO_PATH", opensc_so_path,
-		"ID", engine_id,
+		"SO_PATH", NULL /* opensc_so_path */,
+		"ID", NULL /* engine_id */,
 		"LIST_ADD", "1",
 		"LOAD", NULL,
 		NULL, NULL
@@ -577,6 +683,9 @@ static int tls_engine_load_dynamic_opensc(const char *opensc_so_path)
 	if (!opensc_so_path)
 		return 0;
 
+	pre_cmd[1] = opensc_so_path;
+	pre_cmd[3] = engine_id;
+
 	wpa_printf(MSG_DEBUG, "ENGINE: Loading OpenSC Engine from %s",
 		   opensc_so_path);
 
@@ -589,14 +698,18 @@ void * tls_init(const struct tls_config *conf)
 {
 	SSL_CTX *ssl;
 
-	SSL_load_error_strings();
-	SSL_library_init();
-	/* TODO: if /dev/urandom is available, PRNG is seeded automatically.
-	 * If this is not the case, random data should be added here. */
+	if (tls_openssl_ref_count == 0) {
+		SSL_load_error_strings();
+		SSL_library_init();
+		/* TODO: if /dev/urandom is available, PRNG is seeded
+		 * automatically. If this is not the case, random data should
+		 * be added here. */
 
 #ifdef PKCS12_FUNCS
-	PKCS12_PBE_add();
+		PKCS12_PBE_add();
 #endif  /* PKCS12_FUNCS */
+	}
+	tls_openssl_ref_count++;
 
 	ssl = SSL_CTX_new(TLSv1_method());
 	if (ssl == NULL)
@@ -629,11 +742,15 @@ void tls_deinit(void *ssl_ctx)
 {
 	SSL_CTX *ssl = ssl_ctx;
 	SSL_CTX_free(ssl);
+
+	tls_openssl_ref_count--;
+	if (tls_openssl_ref_count == 0) {
 #ifndef OPENSSL_NO_ENGINE
-	ENGINE_cleanup();
+		ENGINE_cleanup();
 #endif /* OPENSSL_NO_ENGINE */
-	ERR_free_strings();
-	EVP_cleanup();
+		ERR_free_strings();
+		EVP_cleanup();
+	}
 }
 
 
@@ -658,30 +775,29 @@ static int tls_engine_init(struct tls_connection *conn, const char *engine_id,
 	ERR_clear_error();
 	conn->engine = ENGINE_by_id(engine_id);
 	if (!conn->engine) {
-		char *error = ERR_error_string(ERR_get_error(), NULL);
 		wpa_printf(MSG_ERROR, "ENGINE: engine %s not available [%s]",
-			   engine_id, error);
+			   engine_id, ERR_error_string(ERR_get_error(), NULL));
 		goto err;
 	}
 	if (ENGINE_init(conn->engine) != 1) {
-		char *error = ERR_error_string(ERR_get_error(), NULL);
 		wpa_printf(MSG_ERROR, "ENGINE: engine init failed "
-			   "(engine: %s) [%s]", engine_id, error);
+			   "(engine: %s) [%s]", engine_id,
+			   ERR_error_string(ERR_get_error(), NULL));
 		goto err;
 	}
 	wpa_printf(MSG_DEBUG, "ENGINE: engine initialized");
 
 	if (ENGINE_ctrl_cmd_string(conn->engine, "PIN", pin, 0) == 0) {
-		char *error = ERR_error_string(ERR_get_error(), NULL);
-		wpa_printf(MSG_ERROR, "ENGINE: can't set pin [%s]", error);
+		wpa_printf(MSG_ERROR, "ENGINE: cannot set pin [%s]",
+			   ERR_error_string(ERR_get_error(), NULL));
 		goto err;
 	}
 	conn->private_key = ENGINE_load_private_key(conn->engine,
 						    key_id, NULL, NULL);
 	if (!conn->private_key) {
-		char *error = ERR_error_string(ERR_get_error(), NULL);
-		wpa_printf(MSG_ERROR, "ENGINE: can't load private key with id"
-				" '%s' [%s]", key_id, error);
+		wpa_printf(MSG_ERROR, "ENGINE: cannot load private key with id"
+				" '%s' [%s]", key_id,
+			   ERR_error_string(ERR_get_error(), NULL));
 		ret = TLS_SET_PARAMS_ENGINE_PRV_INIT_FAILED;
 		goto err;
 	}
@@ -740,15 +856,13 @@ struct tls_connection * tls_connection_init(void *ssl_ctx)
 	SSL_CTX *ssl = ssl_ctx;
 	struct tls_connection *conn;
 
-	conn = malloc(sizeof(*conn));
+	conn = wpa_zalloc(sizeof(*conn));
 	if (conn == NULL)
 		return NULL;
-	memset(conn, 0, sizeof(*conn));
 	conn->ssl = SSL_new(ssl);
 	if (conn->ssl == NULL) {
-		wpa_printf(MSG_INFO, "TLS: Failed to initialize new SSL "
-			   "connection: %s",
-			   ERR_error_string(ERR_get_error(), NULL));
+		tls_show_errors(MSG_INFO, __func__,
+				"Failed to initialize new SSL connection");
 		free(conn);
 		return NULL;
 	}
@@ -760,9 +874,8 @@ struct tls_connection * tls_connection_init(void *ssl_ctx)
 
 	conn->ssl_in = BIO_new(BIO_s_mem());
 	if (!conn->ssl_in) {
-		wpa_printf(MSG_INFO, "SSL: Failed to create a new BIO for "
-			   "ssl_in: %s",
-			   ERR_error_string(ERR_get_error(), NULL));
+		tls_show_errors(MSG_INFO, __func__,
+				"Failed to create a new BIO for ssl_in");
 		SSL_free(conn->ssl);
 		free(conn);
 		return NULL;
@@ -770,9 +883,8 @@ struct tls_connection * tls_connection_init(void *ssl_ctx)
 
 	conn->ssl_out = BIO_new(BIO_s_mem());
 	if (!conn->ssl_out) {
-		wpa_printf(MSG_INFO, "SSL: Failed to create a new BIO for "
-			   "ssl_out: %s",
-			   ERR_error_string(ERR_get_error(), NULL));
+		tls_show_errors(MSG_INFO, __func__,
+				"Failed to create a new BIO for ssl_out");
 		SSL_free(conn->ssl);
 		BIO_free(conn->ssl_in);
 		free(conn);
@@ -912,22 +1024,102 @@ static int tls_verify_cb(int preverify_ok, X509_STORE_CTX *x509_ctx)
 }
 
 
-static int tls_connection_ca_cert(void *ssl_ctx, struct tls_connection *conn,
-				  const char *ca_cert)
+#ifndef OPENSSL_NO_STDIO
+static int tls_load_ca_der(void *_ssl_ctx, const char *ca_cert)
 {
-	if (ca_cert) {
-		if (SSL_CTX_load_verify_locations(ssl_ctx, ca_cert, NULL) != 1)
-		{
-			wpa_printf(MSG_WARNING, "TLS: Failed to load root "
-				   "certificates: %s",
-				   ERR_error_string(ERR_get_error(), NULL));
+	SSL_CTX *ssl_ctx = _ssl_ctx;
+	X509_LOOKUP *lookup;
+	int ret = 0;
+
+	lookup = X509_STORE_add_lookup(ssl_ctx->cert_store,
+				       X509_LOOKUP_file());
+	if (lookup == NULL) {
+		tls_show_errors(MSG_WARNING, __func__,
+				"Failed add lookup for X509 store");
+		return -1;
+	}
+
+	if (!X509_LOOKUP_load_file(lookup, ca_cert, X509_FILETYPE_ASN1)) {
+		unsigned long err = ERR_peek_error();
+		tls_show_errors(MSG_WARNING, __func__,
+				"Failed load CA in DER format");
+		if (ERR_GET_LIB(err) == ERR_LIB_X509 &&
+		    ERR_GET_REASON(err) == X509_R_CERT_ALREADY_IN_HASH_TABLE) {
+			wpa_printf(MSG_DEBUG, "OpenSSL: %s - ignoring "
+				   "cert already in hash table error",
+				   __func__);
+		} else
+			ret = -1;
+	}
+
+	return ret;
+}
+#endif /* OPENSSL_NO_STDIO */
+
+
+static int tls_connection_ca_cert(void *_ssl_ctx, struct tls_connection *conn,
+				  const char *ca_cert, const u8 *ca_cert_blob,
+				  size_t ca_cert_blob_len, const char *ca_path)
+{
+	SSL_CTX *ssl_ctx = _ssl_ctx;
+
+	if (ca_cert_blob) {
+		X509 *cert = d2i_X509(NULL, (OPENSSL_d2i_TYPE) &ca_cert_blob,
+				      ca_cert_blob_len);
+		if (cert == NULL) {
+			tls_show_errors(MSG_WARNING, __func__,
+					"Failed to parse ca_cert_blob");
 			return -1;
+		}
+
+		if (!X509_STORE_add_cert(ssl_ctx->cert_store, cert)) {
+			tls_show_errors(MSG_WARNING, __func__,
+					"Failed to add ca_cert_blob to "
+					"certificate store");
+			X509_free(cert);
+			return -1;
+		}
+		X509_free(cert);
+		wpa_printf(MSG_DEBUG, "OpenSSL: %s - added ca_cert_blob "
+			   "to certificate store", __func__);
+		SSL_set_verify(conn->ssl, SSL_VERIFY_PEER, tls_verify_cb);
+		return 0;
+	}
+
+#ifdef CONFIG_NATIVE_WINDOWS
+	if (ca_cert && tls_cryptoapi_ca_cert(ssl_ctx, conn->ssl, ca_cert) ==
+	    0) {
+		wpa_printf(MSG_DEBUG, "OpenSSL: Added CA certificates from "
+			   "system certificate store");
+		SSL_set_verify(conn->ssl, SSL_VERIFY_PEER, tls_verify_cb);
+		return 0;
+	}
+#endif /* CONFIG_NATIVE_WINDOWS */
+
+	if (ca_cert || ca_path) {
+#ifndef OPENSSL_NO_STDIO
+		if (SSL_CTX_load_verify_locations(ssl_ctx, ca_cert, ca_path) !=
+		    1) {
+			tls_show_errors(MSG_WARNING, __func__,
+					"Failed to load root certificates");
+			if (ca_cert &&
+			    tls_load_ca_der(ssl_ctx, ca_cert) == 0) {
+				wpa_printf(MSG_DEBUG, "OpenSSL: %s - loaded "
+					   "DER format CA certificate",
+					   __func__);
+			} else
+				return -1;
 		} else {
 			wpa_printf(MSG_DEBUG, "TLS: Trusted root "
 				   "certificate(s) loaded");
 			tls_get_errors(ssl_ctx);
 		}
 		SSL_set_verify(conn->ssl, SSL_VERIFY_PEER, tls_verify_cb);
+#else /* OPENSSL_NO_STDIO */
+		wpa_printf(MSG_DEBUG, "OpenSSL: %s - OPENSSL_NO_STDIO",
+			   __func__);
+		return -1;
+#endif /* OPENSSL_NO_STDIO */
 	} else {
 		/* No ca_cert configured - do not try to verify server
 		 * certificate */
@@ -938,24 +1130,24 @@ static int tls_connection_ca_cert(void *ssl_ctx, struct tls_connection *conn,
 }
 
 
-int tls_global_ca_cert(void *_ssl_ctx, const char *ca_cert)
+static int tls_global_ca_cert(SSL_CTX *ssl_ctx, const char *ca_cert)
 {
-	SSL_CTX *ssl_ctx = _ssl_ctx;
 	if (ca_cert) {
 		if (SSL_CTX_load_verify_locations(ssl_ctx, ca_cert, NULL) != 1)
 		{
-			wpa_printf(MSG_WARNING, "TLS: Failed to load root "
-				   "certificates: %s",
-				   ERR_error_string(ERR_get_error(), NULL));
+			tls_show_errors(MSG_WARNING, __func__,
+					"Failed to load root certificates");
 			return -1;
 		}
 
 		wpa_printf(MSG_DEBUG, "TLS: Trusted root "
 			   "certificate(s) loaded");
 
+#ifndef OPENSSL_NO_STDIO
 		/* Add the same CAs to the client certificate requests */
 		SSL_CTX_set_client_CA_list(ssl_ctx,
 					   SSL_load_client_CA_file(ca_cert));
+#endif /* OPENSSL_NO_STDIO */
 	}
 
 	return 0;
@@ -969,9 +1161,9 @@ int tls_global_set_verify(void *ssl_ctx, int check_crl)
 	if (check_crl) {
 		X509_STORE *cs = SSL_CTX_get_cert_store(ssl_ctx);
 		if (cs == NULL) {
-			wpa_printf(MSG_INFO, "TLS: Failed to get certificate "
-				   "store when enabling check_crl: %s",
-				   ERR_error_string(ERR_get_error(), NULL));
+			tls_show_errors(MSG_INFO, __func__, "Failed to get "
+					"certificate store when enabling "
+					"check_crl");
 			return -1;
 		}
 		flags = X509_V_FLAG_CRL_CHECK;
@@ -983,8 +1175,7 @@ int tls_global_set_verify(void *ssl_ctx, int check_crl)
 }
 
 
-static int tls_connection_set_subject_match(void *ssl_ctx,
-					    struct tls_connection *conn,
+static int tls_connection_set_subject_match(struct tls_connection *conn,
 					    const char *subject_match,
 					    const char *altsubject_match)
 {
@@ -1028,29 +1219,59 @@ int tls_connection_set_verify(void *ssl_ctx, struct tls_connection *conn,
 }
 
 
-static int tls_connection_client_cert(void *ssl_ctx,
-				      struct tls_connection *conn,
-				      const char *client_cert)
+static int tls_connection_client_cert(struct tls_connection *conn,
+				      const char *client_cert,
+				      const u8 *client_cert_blob,
+				      size_t client_cert_blob_len)
 {
-	if (client_cert == NULL)
+	if (client_cert == NULL && client_cert_blob == NULL)
 		return 0;
 
-	if (SSL_use_certificate_file(conn->ssl, client_cert,
-				     SSL_FILETYPE_ASN1) != 1 &&
-	    SSL_use_certificate_file(conn->ssl, client_cert,
-				     SSL_FILETYPE_PEM) != 1) {
-		wpa_printf(MSG_INFO, "TLS: Failed to load client "
-			   "certificate: %s",
-			   ERR_error_string(ERR_get_error(), NULL));
-		return -1;
+	if (client_cert_blob &&
+	    SSL_use_certificate_ASN1(conn->ssl, (u8 *) client_cert_blob,
+				     client_cert_blob_len) == 1) {
+		wpa_printf(MSG_DEBUG, "OpenSSL: SSL_use_certificate_ASN1 --> "
+			   "OK");
+		return 0;
+	} else if (client_cert_blob) {
+		tls_show_errors(MSG_DEBUG, __func__,
+				"SSL_use_certificate_ASN1 failed");
 	}
-	return 0;
+
+	if (client_cert == NULL)
+		return -1;
+
+#ifndef OPENSSL_NO_STDIO
+	if (SSL_use_certificate_file(conn->ssl, client_cert,
+				     SSL_FILETYPE_ASN1) == 1) {
+		wpa_printf(MSG_DEBUG, "OpenSSL: SSL_use_certificate_file (DER)"
+			   " --> OK");
+		return 0;
+	} else {
+		tls_show_errors(MSG_DEBUG, __func__,
+				"SSL_use_certificate_file (DER) failed");
+	}
+
+	if (SSL_use_certificate_file(conn->ssl, client_cert,
+				     SSL_FILETYPE_PEM) == 1) {
+		wpa_printf(MSG_DEBUG, "OpenSSL: SSL_use_certificate_file (PEM)"
+			   " --> OK");
+		return 0;
+	} else {
+		tls_show_errors(MSG_DEBUG, __func__,
+				"SSL_use_certificate_file (PEM) failed");
+	}
+#else /* OPENSSL_NO_STDIO */
+	wpa_printf(MSG_DEBUG, "OpenSSL: %s - OPENSSL_NO_STDIO", __func__);
+#endif /* OPENSSL_NO_STDIO */
+
+	return -1;
 }
 
 
-int tls_global_client_cert(void *_ssl_ctx, const char *client_cert)
+static int tls_global_client_cert(SSL_CTX *ssl_ctx, const char *client_cert)
 {
-	SSL_CTX *ssl_ctx = _ssl_ctx;
+#ifndef OPENSSL_NO_STDIO
 	if (client_cert == NULL)
 		return 0;
 
@@ -1058,12 +1279,17 @@ int tls_global_client_cert(void *_ssl_ctx, const char *client_cert)
 					 SSL_FILETYPE_ASN1) != 1 &&
 	    SSL_CTX_use_certificate_file(ssl_ctx, client_cert,
 					 SSL_FILETYPE_PEM) != 1) {
-		wpa_printf(MSG_INFO, "TLS: Failed to load client "
-			   "certificate: %s",
-			   ERR_error_string(ERR_get_error(), NULL));
+		tls_show_errors(MSG_INFO, __func__,
+				"Failed to load client certificate");
 		return -1;
 	}
 	return 0;
+#else /* OPENSSL_NO_STDIO */
+	if (client_cert == NULL)
+		return 0;
+	wpa_printf(MSG_DEBUG, "OpenSSL: %s - OPENSSL_NO_STDIO", __func__);
+	return -1;
+#endif /* OPENSSL_NO_STDIO */
 }
 
 
@@ -1078,42 +1304,26 @@ static int tls_passwd_cb(char *buf, int size, int rwflag, void *password)
 }
 
 
-static int tls_read_pkcs12(SSL_CTX *ssl_ctx, SSL *ssl, const char *private_key,
-			   const char *passwd)
-{
 #ifdef PKCS12_FUNCS
-	FILE *f;
-	PKCS12 *p12;
+static int tls_parse_pkcs12(SSL_CTX *ssl_ctx, SSL *ssl, PKCS12 *p12,
+			    const char *passwd)
+{
 	EVP_PKEY *pkey;
 	X509 *cert;
 	STACK_OF(X509) *certs;
 	int res = 0;
 	char buf[256];
 
-	f = fopen(private_key, "r");
-	if (f == NULL)
-		return -1;
-
-	p12 = d2i_PKCS12_fp(f, NULL);
-	if (p12 == NULL) {
-		wpa_printf(MSG_DEBUG, "TLS: Failed to read PKCS12 file '%s'",
-			   private_key);
-		fclose(f);
-		return -1;
-	}
-	fclose(f);
-
 	pkey = NULL;
 	cert = NULL;
 	certs = NULL;
 	if (!PKCS12_parse(p12, passwd, &pkey, &cert, &certs)) {
-		wpa_printf(MSG_DEBUG, "TLS: Failed to parse PKCS12 file '%s': "
-			   "%s", private_key,
-			   ERR_error_string(ERR_get_error(), NULL));
+		tls_show_errors(MSG_DEBUG, __func__,
+				"Failed to parse PKCS12 file");
+		PKCS12_free(p12);
 		return -1;
 	}
-	wpa_printf(MSG_DEBUG, "TLS: Successfully parsed PKCS12 file '%s'",
-		   private_key);
+	wpa_printf(MSG_DEBUG, "TLS: Successfully parsed PKCS12 data");
 
 	if (cert) {
 		X509_NAME_oneline(X509_get_subject_name(cert), buf,
@@ -1166,6 +1376,32 @@ static int tls_read_pkcs12(SSL_CTX *ssl_ctx, SSL *ssl, const char *private_key,
 		tls_get_errors(ssl_ctx);
 
 	return res;
+}
+#endif  /* PKCS12_FUNCS */
+
+
+static int tls_read_pkcs12(SSL_CTX *ssl_ctx, SSL *ssl, const char *private_key,
+			   const char *passwd)
+{
+#ifdef PKCS12_FUNCS
+	FILE *f;
+	PKCS12 *p12;
+
+	f = fopen(private_key, "r");
+	if (f == NULL)
+		return -1;
+
+	p12 = d2i_PKCS12_fp(f, NULL);
+	fclose(f);
+
+	if (p12 == NULL) {
+		tls_show_errors(MSG_INFO, __func__,
+				"Failed to use PKCS#12 file");
+		return -1;
+	}
+
+	return tls_parse_pkcs12(ssl_ctx, ssl, p12, passwd);
+
 #else /* PKCS12_FUNCS */
 	wpa_printf(MSG_INFO, "TLS: PKCS12 support disabled - cannot read "
 		   "p12/pfx files");
@@ -1174,19 +1410,40 @@ static int tls_read_pkcs12(SSL_CTX *ssl_ctx, SSL *ssl, const char *private_key,
 }
 
 
-static int tls_connection_engine_private_key(void *_ssl_ctx,
-					     struct tls_connection *conn)
+static int tls_read_pkcs12_blob(SSL_CTX *ssl_ctx, SSL *ssl,
+				const u8 *blob, size_t len, const char *passwd)
+{
+#ifdef PKCS12_FUNCS
+	PKCS12 *p12;
+
+	p12 = d2i_PKCS12(NULL, (OPENSSL_d2i_TYPE) &blob, len);
+	if (p12 == NULL) {
+		tls_show_errors(MSG_INFO, __func__,
+				"Failed to use PKCS#12 blob");
+		return -1;
+	}
+
+	return tls_parse_pkcs12(ssl_ctx, ssl, p12, passwd);
+
+#else /* PKCS12_FUNCS */
+	wpa_printf(MSG_INFO, "TLS: PKCS12 support disabled - cannot parse "
+		   "p12/pfx blobs");
+	return -1;
+#endif  /* PKCS12_FUNCS */
+}
+
+
+static int tls_connection_engine_private_key(struct tls_connection *conn)
 {
 #ifndef OPENSSL_NO_ENGINE
 	if (SSL_use_PrivateKey(conn->ssl, conn->private_key) != 1) {
-		char *error = ERR_error_string(ERR_get_error(), NULL);
-		wpa_printf(MSG_ERROR, "ENGINE: can't use private key for TLS "
-			   "[%s]", error);
+		tls_show_errors(MSG_ERROR, __func__,
+				"ENGINE: cannot use private key for TLS");
 		return -1;
 	}
 	if (!SSL_check_private_key(conn->ssl)) {
-		wpa_printf(MSG_INFO, "SSL: Private key failed verification: "
-			   "%s", ERR_error_string(ERR_get_error(), NULL));
+		tls_show_errors(MSG_INFO, __func__,
+				"Private key failed verification");
 		return -1;
 	}
 	return 0;
@@ -1201,12 +1458,15 @@ static int tls_connection_engine_private_key(void *_ssl_ctx,
 static int tls_connection_private_key(void *_ssl_ctx,
 				      struct tls_connection *conn,
 				      const char *private_key,
-				      const char *private_key_passwd)
+				      const char *private_key_passwd,
+				      const u8 *private_key_blob,
+				      size_t private_key_blob_len)
 {
 	SSL_CTX *ssl_ctx = _ssl_ctx;
 	char *passwd;
+	int ok;
 
-	if (private_key == NULL)
+	if (private_key == NULL && private_key_blob == NULL)
 		return 0;
 
 	if (private_key_passwd) {
@@ -1218,37 +1478,130 @@ static int tls_connection_private_key(void *_ssl_ctx,
 
 	SSL_CTX_set_default_passwd_cb(ssl_ctx, tls_passwd_cb);
 	SSL_CTX_set_default_passwd_cb_userdata(ssl_ctx, passwd);
-	if (SSL_use_PrivateKey_file(conn->ssl, private_key,
-				    SSL_FILETYPE_ASN1) != 1 &&
-	    SSL_use_PrivateKey_file(conn->ssl, private_key,
-				    SSL_FILETYPE_PEM) != 1 &&
-	    tls_read_pkcs12(ssl_ctx, conn->ssl, private_key, passwd) &&
-	    tls_cryptoapi_cert(conn->ssl, private_key)) {
-		wpa_printf(MSG_INFO, "SSL: Failed to load private key: %s",
-			   ERR_error_string(ERR_get_error(), NULL));
+
+	ok = 0;
+	while (private_key_blob) {
+		if (SSL_use_PrivateKey_ASN1(EVP_PKEY_RSA, conn->ssl,
+					    (u8 *) private_key_blob,
+					    private_key_blob_len) == 1) {
+			wpa_printf(MSG_DEBUG, "OpenSSL: SSL_use_PrivateKey_"
+				   "ASN1(EVP_PKEY_RSA) --> OK");
+			ok = 1;
+			break;
+		} else {
+			tls_show_errors(MSG_DEBUG, __func__,
+					"SSL_use_PrivateKey_ASN1(EVP_PKEY_RSA)"
+					" failed");
+		}
+
+		if (SSL_use_PrivateKey_ASN1(EVP_PKEY_DSA, conn->ssl,
+					    (u8 *) private_key_blob,
+					    private_key_blob_len) == 1) {
+			wpa_printf(MSG_DEBUG, "OpenSSL: SSL_use_PrivateKey_"
+				   "ASN1(EVP_PKEY_DSA) --> OK");
+			ok = 1;
+			break;
+		} else {
+			tls_show_errors(MSG_DEBUG, __func__,
+					"SSL_use_PrivateKey_ASN1(EVP_PKEY_DSA)"
+					" failed");
+		}
+
+		if (SSL_use_RSAPrivateKey_ASN1(conn->ssl,
+					       (u8 *) private_key_blob,
+					       private_key_blob_len) == 1) {
+			wpa_printf(MSG_DEBUG, "OpenSSL: "
+				   "SSL_use_RSAPrivateKey_ASN1 --> OK");
+			ok = 1;
+			break;
+		} else {
+			tls_show_errors(MSG_DEBUG, __func__,
+					"SSL_use_RSAPrivateKey_ASN1 failed");
+		}
+
+		if (tls_read_pkcs12_blob(ssl_ctx, conn->ssl, private_key_blob,
+					 private_key_blob_len, passwd) == 0) {
+			wpa_printf(MSG_DEBUG, "OpenSSL: PKCS#12 as blob --> "
+				   "OK");
+			ok = 1;
+			break;
+		}
+
+		break;
+	}
+
+	while (!ok && private_key) {
+#ifndef OPENSSL_NO_STDIO
+		if (SSL_use_PrivateKey_file(conn->ssl, private_key,
+					    SSL_FILETYPE_ASN1) == 1) {
+			wpa_printf(MSG_DEBUG, "OpenSSL: "
+				   "SSL_use_PrivateKey_File (DER) --> OK");
+			ok = 1;
+			break;
+		} else {
+			tls_show_errors(MSG_DEBUG, __func__,
+					"SSL_use_PrivateKey_File (DER) "
+					"failed");
+		}
+
+		if (SSL_use_PrivateKey_file(conn->ssl, private_key,
+					    SSL_FILETYPE_PEM) == 1) {
+			wpa_printf(MSG_DEBUG, "OpenSSL: "
+				   "SSL_use_PrivateKey_File (PEM) --> OK");
+			ok = 1;
+			break;
+		} else {
+			tls_show_errors(MSG_DEBUG, __func__,
+					"SSL_use_PrivateKey_File (PEM) "
+					"failed");
+		}
+#else /* OPENSSL_NO_STDIO */
+		wpa_printf(MSG_DEBUG, "OpenSSL: %s - OPENSSL_NO_STDIO",
+			   __func__);
+#endif /* OPENSSL_NO_STDIO */
+
+		if (tls_read_pkcs12(ssl_ctx, conn->ssl, private_key, passwd)
+		    == 0) {
+			wpa_printf(MSG_DEBUG, "OpenSSL: Reading PKCS#12 file "
+				   "--> OK");
+			ok = 1;
+			break;
+		}
+
+		if (tls_cryptoapi_cert(conn->ssl, private_key) == 0) {
+			wpa_printf(MSG_DEBUG, "OpenSSL: Using CryptoAPI to "
+				   "access certificate store --> OK");
+			ok = 1;
+			break;
+		}
+
+		break;
+	}
+
+	if (!ok) {
+		wpa_printf(MSG_INFO, "OpenSSL: Failed to load private key");
 		free(passwd);
 		ERR_clear_error();
 		return -1;
 	}
 	ERR_clear_error();
-	free(passwd);
 	SSL_CTX_set_default_passwd_cb(ssl_ctx, NULL);
+	free(passwd);
 	
 	if (!SSL_check_private_key(conn->ssl)) {
-		wpa_printf(MSG_INFO, "SSL: Private key failed "
-			   "verification: %s",
-			   ERR_error_string(ERR_get_error(), NULL));
+		tls_show_errors(MSG_INFO, __func__, "Private key failed "
+				"verification");
 		return -1;
 	}
 
+	wpa_printf(MSG_DEBUG, "SSL: Private key loaded successfully");
 	return 0;
 }
 
 
-int tls_global_private_key(void *_ssl_ctx, const char *private_key,
-			   const char *private_key_passwd)
+static int tls_global_private_key(SSL_CTX *ssl_ctx, const char *private_key,
+				  const char *private_key_passwd)
 {
-	SSL_CTX *ssl_ctx = _ssl_ctx;
 	char *passwd;
 
 	if (private_key == NULL)
@@ -1263,13 +1616,16 @@ int tls_global_private_key(void *_ssl_ctx, const char *private_key,
 
 	SSL_CTX_set_default_passwd_cb(ssl_ctx, tls_passwd_cb);
 	SSL_CTX_set_default_passwd_cb_userdata(ssl_ctx, passwd);
-	if (SSL_CTX_use_PrivateKey_file(ssl_ctx, private_key,
+	if (
+#ifndef OPENSSL_NO_STDIO
+	    SSL_CTX_use_PrivateKey_file(ssl_ctx, private_key,
 					SSL_FILETYPE_ASN1) != 1 &&
 	    SSL_CTX_use_PrivateKey_file(ssl_ctx, private_key,
 					SSL_FILETYPE_PEM) != 1 &&
+#endif /* OPENSSL_NO_STDIO */
 	    tls_read_pkcs12(ssl_ctx, NULL, private_key, passwd)) {
-		wpa_printf(MSG_INFO, "SSL: Failed to load private key: %s",
-			   ERR_error_string(ERR_get_error(), NULL));
+		tls_show_errors(MSG_INFO, __func__,
+				"Failed to load private key");
 		free(passwd);
 		ERR_clear_error();
 		return -1;
@@ -1279,9 +1635,8 @@ int tls_global_private_key(void *_ssl_ctx, const char *private_key,
 	SSL_CTX_set_default_passwd_cb(ssl_ctx, NULL);
 	
 	if (!SSL_CTX_check_private_key(ssl_ctx)) {
-		wpa_printf(MSG_INFO, "SSL: Private key failed "
-			   "verification: %s",
-			   ERR_error_string(ERR_get_error(), NULL));
+		tls_show_errors(MSG_INFO, __func__,
+				"Private key failed verification");
 		return -1;
 	}
 
@@ -1289,8 +1644,7 @@ int tls_global_private_key(void *_ssl_ctx, const char *private_key,
 }
 
 
-static int tls_connection_dh(void *ssl_ctx, struct tls_connection *conn,
-		      const char *dh_file)
+static int tls_connection_dh(struct tls_connection *conn, const char *dh_file)
 {
 #ifdef OPENSSL_NO_DH
 	if (dh_file == NULL)
@@ -1302,6 +1656,7 @@ static int tls_connection_dh(void *ssl_ctx, struct tls_connection *conn,
 	DH *dh;
 	BIO *bio;
 
+	/* TODO: add support for dh_blob */
 	if (dh_file == NULL)
 		return 0;
 	if (conn == NULL)
@@ -1374,6 +1729,7 @@ int tls_connection_get_keys(void *ssl_ctx, struct tls_connection *conn,
 	if (ssl == NULL || ssl->s3 == NULL || ssl->session == NULL)
 		return -1;
 
+	memset(keys, 0, sizeof(*keys));
 	keys->master_key = ssl->session->master_key;
 	keys->master_key_len = ssl->session->master_key_length;
 	keys->client_random = ssl->s3->client_random;
@@ -1385,6 +1741,14 @@ int tls_connection_get_keys(void *ssl_ctx, struct tls_connection *conn,
 }
 
 
+int tls_connection_prf(void *tls_ctx, struct tls_connection *conn,
+		       const char *label, int server_random_first,
+		       u8 *out, size_t out_len)
+{
+	return -1;
+}
+
+
 u8 * tls_connection_handshake(void *ssl_ctx, struct tls_connection *conn,
 			      const u8 *in_data, size_t in_len,
 			      size_t *out_len)
@@ -1398,8 +1762,8 @@ u8 * tls_connection_handshake(void *ssl_ctx, struct tls_connection *conn,
 	 */
 	if (in_data &&
 	    BIO_write(conn->ssl_in, in_data, in_len) < 0) {
-		wpa_printf(MSG_INFO, "TLS: Handshake failed - BIO_write: %s",
-			   ERR_error_string(ERR_get_error(), NULL));
+		tls_show_errors(MSG_INFO, __func__,
+				"Handshake failed - BIO_write");
 		return NULL;
 	}
 
@@ -1414,8 +1778,7 @@ u8 * tls_connection_handshake(void *ssl_ctx, struct tls_connection *conn,
 			wpa_printf(MSG_DEBUG, "SSL: SSL_connect - want to "
 				   "write");
 		else {
-			wpa_printf(MSG_INFO, "SSL: SSL_connect: %s",
-				   ERR_error_string(ERR_get_error(), NULL));
+			tls_show_errors(MSG_INFO, __func__, "SSL_connect");
 			conn->failed++;
 		}
 	}
@@ -1427,15 +1790,21 @@ u8 * tls_connection_handshake(void *ssl_ctx, struct tls_connection *conn,
 	if (out_data == NULL) {
 		wpa_printf(MSG_DEBUG, "SSL: Failed to allocate memory for "
 			   "handshake output (%d bytes)", res);
-		BIO_reset(conn->ssl_out);
+		if (BIO_reset(conn->ssl_out) < 0) {
+			tls_show_errors(MSG_INFO, __func__,
+					"BIO_reset failed");
+		}
 		*out_len = 0;
 		return NULL;
 	}
 	res = res == 0 ? 0 : BIO_read(conn->ssl_out, out_data, res);
 	if (res < 0) {
-		wpa_printf(MSG_INFO, "TLS: Handshake failed - BIO_read: %s",
-			   ERR_error_string(ERR_get_error(), NULL));
-		BIO_reset(conn->ssl_out);
+		tls_show_errors(MSG_INFO, __func__,
+				"Handshake failed - BIO_read");
+		if (BIO_reset(conn->ssl_out) < 0) {
+			tls_show_errors(MSG_INFO, __func__,
+					"BIO_reset failed");
+		}
 		*out_len = 0;
 		return NULL;
 	}
@@ -1455,8 +1824,8 @@ u8 * tls_connection_server_handshake(void *ssl_ctx,
 
 	if (in_data &&
 	    BIO_write(conn->ssl_in, in_data, in_len) < 0) {
-		wpa_printf(MSG_INFO, "TLS: Handshake failed - BIO_write: %s",
-			   ERR_error_string(ERR_get_error(), NULL));
+		tls_show_errors(MSG_INFO, __func__,
+				"Handshake failed - BIO_write");
 		return NULL;
 	}
 
@@ -1472,15 +1841,21 @@ u8 * tls_connection_server_handshake(void *ssl_ctx,
 	if (out_data == NULL) {
 		wpa_printf(MSG_DEBUG, "SSL: Failed to allocate memory for "
 			   "handshake output (%d bytes)", res);
-		BIO_reset(conn->ssl_out);
+		if (BIO_reset(conn->ssl_out) < 0) {
+			tls_show_errors(MSG_INFO, __func__,
+					"BIO_reset failed");
+		}
 		*out_len = 0;
 		return NULL;
 	}
 	res = res == 0 ? 0 : BIO_read(conn->ssl_out, out_data, res);
 	if (res < 0) {
-		wpa_printf(MSG_INFO, "TLS: Handshake failed - BIO_read: %s",
-			   ERR_error_string(ERR_get_error(), NULL));
-		BIO_reset(conn->ssl_out);
+		tls_show_errors(MSG_INFO, __func__,
+				"Handshake failed - BIO_read");
+		if (BIO_reset(conn->ssl_out) < 0) {
+			tls_show_errors(MSG_INFO, __func__,
+					"BIO_reset failed");
+		}
 		*out_len = 0;
 		return NULL;
 	}
@@ -1499,20 +1874,23 @@ int tls_connection_encrypt(void *ssl_ctx, struct tls_connection *conn,
 		return -1;
 
 	/* Give plaintext data for OpenSSL to encrypt into the TLS tunnel. */
-	BIO_reset(conn->ssl_in);
-	BIO_reset(conn->ssl_out);
+	if ((res = BIO_reset(conn->ssl_in)) < 0 ||
+	    (res = BIO_reset(conn->ssl_out)) < 0) {
+		tls_show_errors(MSG_INFO, __func__, "BIO_reset failed");
+		return res;
+	}
 	res = SSL_write(conn->ssl, in_data, in_len);
 	if (res < 0) {
-		wpa_printf(MSG_INFO, "TLS: Encryption failed - SSL_write: %s",
-			   ERR_error_string(ERR_get_error(), NULL));
+		tls_show_errors(MSG_INFO, __func__,
+				"Encryption failed - SSL_write");
 		return res;
 	}
 
 	/* Read encrypted data to be sent to the server */
 	res = BIO_read(conn->ssl_out, out_data, out_len);
 	if (res < 0) {
-		wpa_printf(MSG_INFO, "TLS: Encryption failed - BIO_read: %s",
-			   ERR_error_string(ERR_get_error(), NULL));
+		tls_show_errors(MSG_INFO, __func__,
+				"Encryption failed - BIO_read");
 		return res;
 	}
 
@@ -1529,17 +1907,20 @@ int tls_connection_decrypt(void *ssl_ctx, struct tls_connection *conn,
 	/* Give encrypted data from TLS tunnel for OpenSSL to decrypt. */
 	res = BIO_write(conn->ssl_in, in_data, in_len);
 	if (res < 0) {
-		wpa_printf(MSG_INFO, "TLS: Decryption failed - BIO_write: %s",
-			   ERR_error_string(ERR_get_error(), NULL));
+		tls_show_errors(MSG_INFO, __func__,
+				"Decryption failed - BIO_write");
+		return res;
+	}
+	if (BIO_reset(conn->ssl_out) < 0) {
+		tls_show_errors(MSG_INFO, __func__, "BIO_reset failed");
 		return res;
 	}
-	BIO_reset(conn->ssl_out);
 
 	/* Read decrypted data for further processing */
 	res = SSL_read(conn->ssl, out_data, out_len);
 	if (res < 0) {
-		wpa_printf(MSG_INFO, "TLS: Decryption failed - SSL_read: %s",
-			   ERR_error_string(ERR_get_error(), NULL));
+		tls_show_errors(MSG_INFO, __func__,
+				"Decryption failed - SSL_read");
 		return res;
 	}
 
@@ -1553,7 +1934,7 @@ int tls_connection_resumed(void *ssl_ctx, struct tls_connection *conn)
 }
 
 
-#ifdef EAP_FAST
+#if defined(EAP_FAST) || defined(EAP_FAST_DYNAMIC)
 /* Pre-shared secred requires a patch to openssl, so this function is
  * commented out unless explicitly needed for EAP-FAST in order to be able to
  * build this file with unmodified openssl. */
@@ -1600,7 +1981,7 @@ int tls_connection_set_master_key(void *ssl_ctx, struct tls_connection *conn,
 
 	return 0;
 }
-#endif /* EAP_FAST */
+#endif /* EAP_FAST || EAP_FAST_DYNAMIC */
 
 
 int tls_connection_set_anon_dh(void *ssl_ctx, struct tls_connection *conn)
@@ -1609,8 +1990,8 @@ int tls_connection_set_anon_dh(void *ssl_ctx, struct tls_connection *conn)
 		return -1;
 
 	if (SSL_set_cipher_list(conn->ssl, "ADH-AES128-SHA") != 1) {
-		wpa_printf(MSG_INFO, "TLS: Anon DH configuration failed - %s",
-			   ERR_error_string(ERR_get_error(), NULL));
+		tls_show_errors(MSG_INFO, __func__,
+				"Anon DH configuration failed");
 		return -1;
 	}
 
@@ -1643,7 +2024,7 @@ int tls_connection_enable_workaround(void *ssl_ctx,
 }
 
 
-#ifdef EAP_FAST
+#if defined(EAP_FAST) || defined(EAP_FAST_DYNAMIC)
 /* ClientHello TLS extensions require a patch to openssl, so this function is
  * commented out unless explicitly needed for EAP-FAST in order to be able to
  * build this file with unmodified openssl. */
@@ -1660,7 +2041,7 @@ int tls_connection_client_hello_ext(void *ssl_ctx, struct tls_connection *conn,
 
 	return 0;
 }
-#endif /* EAP_FAST */
+#endif /* EAP_FAST || EAP_FAST_DYNAMIC */
 
 
 int tls_connection_get_failed(void *ssl_ctx, struct tls_connection *conn)
@@ -1691,16 +2072,28 @@ int tls_connection_set_params(void *tls_ctx, struct tls_connection *conn,
 			      const struct tls_connection_params *params)
 {
 	int ret;
+	unsigned long err;
 
 	if (conn == NULL)
 		return -1;
-	if (tls_connection_set_subject_match(tls_ctx, conn,
+
+	while ((err = ERR_get_error())) {
+		wpa_printf(MSG_INFO, "%s: Clearing pending SSL error: %s",
+			   __func__, ERR_error_string(err, NULL));
+	}
+
+	if (tls_connection_set_subject_match(conn,
 					     params->subject_match,
 					     params->altsubject_match))
 		return -1;
-	if (tls_connection_ca_cert(tls_ctx, conn, params->ca_cert))
+	if (tls_connection_ca_cert(tls_ctx, conn, params->ca_cert,
+				   params->ca_cert_blob,
+				   params->ca_cert_blob_len,
+				   params->ca_path))
 		return -1;
-	if (tls_connection_client_cert(tls_ctx, conn, params->client_cert))
+	if (tls_connection_client_cert(conn, params->client_cert,
+				       params->client_cert_blob,
+				       params->client_cert_blob_len))
 		return -1;
 
 	if (params->engine) {
@@ -1709,21 +2102,108 @@ int tls_connection_set_params(void *tls_ctx, struct tls_connection *conn,
 				      params->key_id);
 		if (ret)
 			return ret;
-		if (tls_connection_engine_private_key(tls_ctx, conn))
+		if (tls_connection_engine_private_key(conn))
 			return TLS_SET_PARAMS_ENGINE_PRV_VERIFY_FAILED;
 	} else if (tls_connection_private_key(tls_ctx, conn,
 					      params->private_key,
-					      params->private_key_passwd)) {
+					      params->private_key_passwd,
+					      params->private_key_blob,
+					      params->private_key_blob_len)) {
 		wpa_printf(MSG_INFO, "TLS: Failed to load private key '%s'",
 			   params->private_key);
 		return -1;
 	}
 
-	if (tls_connection_dh(tls_ctx, conn, params->dh_file)) {
+	if (tls_connection_dh(conn, params->dh_file)) {
 		wpa_printf(MSG_INFO, "TLS: Failed to load DH file '%s'",
 			   params->dh_file);
 		return -1;
 	}
 
+	tls_get_errors(tls_ctx);
+
 	return 0;
 }
+
+
+int tls_global_set_params(void *tls_ctx,
+			  const struct tls_connection_params *params)
+{
+	SSL_CTX *ssl_ctx = tls_ctx;
+	unsigned long err;
+
+	while ((err = ERR_get_error())) {
+		wpa_printf(MSG_INFO, "%s: Clearing pending SSL error: %s",
+			   __func__, ERR_error_string(err, NULL));
+	}
+
+	if (tls_global_ca_cert(ssl_ctx, params->ca_cert))
+		return -1;
+
+	if (tls_global_client_cert(ssl_ctx, params->client_cert))
+		return -1;
+
+	if (tls_global_private_key(ssl_ctx, params->private_key,
+				   params->private_key_passwd))
+		return -1;
+
+	return 0;
+}
+
+
+int tls_connection_get_keyblock_size(void *tls_ctx,
+				     struct tls_connection *conn)
+{
+	const EVP_CIPHER *c;
+	const EVP_MD *h;
+
+	if (conn == NULL || conn->ssl == NULL ||
+	    conn->ssl->enc_read_ctx == NULL ||
+	    conn->ssl->enc_read_ctx->cipher == NULL ||
+	    conn->ssl->read_hash == NULL)
+		return -1;
+
+	c = conn->ssl->enc_read_ctx->cipher;
+	h = conn->ssl->read_hash;
+
+	return 2 * (EVP_CIPHER_key_length(c) +
+		    EVP_MD_size(h) +
+		    EVP_CIPHER_iv_length(c));
+}
+
+
+unsigned int tls_capabilities(void *tls_ctx)
+{
+	return 0;
+}
+
+
+int tls_connection_set_ia(void *tls_ctx, struct tls_connection *conn,
+			  int tls_ia)
+{
+	return -1;
+}
+
+
+int tls_connection_ia_send_phase_finished(void *tls_ctx,
+					  struct tls_connection *conn,
+					  int final,
+					  u8 *out_data, size_t out_len)
+{
+	return -1;
+}
+
+
+int tls_connection_ia_final_phase_finished(void *tls_ctx,
+					   struct tls_connection *conn)
+{
+	return -1;
+}
+
+
+int tls_connection_ia_permute_inner_secret(void *tls_ctx,
+					   struct tls_connection *conn,
+					   const u8 *key, size_t key_len)
+{
+	return -1;
+}
