@@ -1,6 +1,6 @@
 /*
  * WPA Supplicant - command line interface for wpa_supplicant daemon
- * Copyright (c) 2004-2005, Jouni Malinen <jkmaline@cc.hut.fi>
+ * Copyright (c) 2004-2006, Jouni Malinen <jkmaline@cc.hut.fi>
  *
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License version 2 as
@@ -12,28 +12,26 @@
  * See README and COPYING for more details.
  */
 
-#include <stdlib.h>
-#include <stdio.h>
-#include <string.h>
-#include <signal.h>
-#include <unistd.h>
+#include "includes.h"
+
+#ifdef CONFIG_CTRL_IFACE
+
+#ifndef CONFIG_CTRL_IFACE_UDP
 #include <dirent.h>
-#include <errno.h>
+#endif /* CONFIG_CTRL_IFACE_UDP */
 #ifdef CONFIG_READLINE
 #include <readline/readline.h>
 #include <readline/history.h>
 #endif /* CONFIG_READLINE */
 
 #include "wpa_ctrl.h"
-#ifdef CONFIG_NATIVE_WINDOWS
 #include "common.h"
-#endif /* CONFIG_NATIVE_WINDOWS */
 #include "version.h"
 
 
 static const char *wpa_cli_version =
 "wpa_cli v" VERSION_STR "\n"
-"Copyright (c) 2004-2005, Jouni Malinen <jkmaline@cc.hut.fi> and contributors";
+"Copyright (c) 2004-2006, Jouni Malinen <jkmaline@cc.hut.fi> and contributors";
 
 
 static const char *wpa_cli_license =
@@ -55,7 +53,7 @@ static const char *wpa_cli_full_license =
 "\n"
 "You should have received a copy of the GNU General Public License\n"
 "along with this program; if not, write to the Free Software\n"
-"Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA\n"
+"Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA\n"
 "\n"
 "Alternatively, this software may be distributed under the terms of the\n"
 "BSD license.\n"
@@ -126,12 +124,19 @@ static const char *commands_help =
 "connecting\n"
 "  scan = request new BSS scan\n"
 "  scan_results = get latest scan results\n"
+"  get_capability <eap/pairwise/group/key_mgmt/proto/auth_alg> = "
+"get capabilies\n"
+"  stakey-request <addr> = request STAKey negotiation with <addr>\n"
+"  ap_scan <value> = set ap_scan parameter\n"
+"  stkstart <addr> = request STK negotiation with <addr>\n"
 "  terminate = terminate wpa_supplicant\n"
 "  quit = exit wpa_cli\n";
 
 static struct wpa_ctrl *ctrl_conn;
 static int wpa_cli_quit = 0;
 static int wpa_cli_attached = 0;
+static int wpa_cli_connected = 0;
+static int wpa_cli_last_id = 0;
 static const char *ctrl_iface_dir = "/var/run/wpa_supplicant";
 static char *ctrl_ifname = NULL;
 static const char *pid_file = NULL;
@@ -142,7 +147,7 @@ static void usage(void)
 {
 	printf("wpa_cli [-p<path to ctrl sockets>] [-i<ifname>] [-hvB] "
 	       "[-a<action file>] \\\n"
-	       "        [-P<pid file> [command..]\n"
+	       "        [-P<pid file>] [-g<global ctrl>]  [command..]\n"
 	       "  -h = help (show this usage text)\n"
 	       "  -v = shown version information\n"
 	       "  -a = run in daemon mode executing the action file based on "
@@ -304,6 +309,7 @@ static void wpa_cli_show_variables(void)
 static int wpa_cli_cmd_set(struct wpa_ctrl *ctrl, int argc, char *argv[])
 {
 	char cmd[256];
+	int res;
 
 	if (argc == 0) {
 		wpa_cli_show_variables();
@@ -316,8 +322,8 @@ static int wpa_cli_cmd_set(struct wpa_ctrl *ctrl, int argc, char *argv[])
 		return 0;
 	}
 
-	if (snprintf(cmd, sizeof(cmd), "SET %s %s", argv[0], argv[1]) >=
-	    sizeof(cmd) - 1) {
+	res = snprintf(cmd, sizeof(cmd), "SET %s %s", argv[0], argv[1]);
+	if (res < 0 || (size_t) res >= sizeof(cmd) - 1) {
 		printf("Too long SET command.\n");
 		return 0;
 	}
@@ -348,6 +354,7 @@ static int wpa_cli_cmd_preauthenticate(struct wpa_ctrl *ctrl, int argc,
 				       char *argv[])
 {
 	char cmd[256];
+	int res;
 
 	if (argc != 1) {
 		printf("Invalid PREAUTH command: needs one argument "
@@ -355,8 +362,8 @@ static int wpa_cli_cmd_preauthenticate(struct wpa_ctrl *ctrl, int argc,
 		return 0;
 	}
 
-	if (snprintf(cmd, sizeof(cmd), "PREAUTH %s", argv[0]) >=
-	    sizeof(cmd) - 1) {
+	res = snprintf(cmd, sizeof(cmd), "PREAUTH %s", argv[0]);
+	if (res < 0 || (size_t) res >= sizeof(cmd) - 1) {
 		printf("Too long PREAUTH command.\n");
 		return 0;
 	}
@@ -364,6 +371,61 @@ static int wpa_cli_cmd_preauthenticate(struct wpa_ctrl *ctrl, int argc,
 }
 
 
+static int wpa_cli_cmd_stakey_request(struct wpa_ctrl *ctrl, int argc,
+				      char *argv[])
+{
+	char cmd[256];
+	int res;
+
+	if (argc != 1) {
+		printf("Invalid STAKEY-REQUEST command: needs one argument "
+		       "(Peer STA MAC address)\n");
+		return 0;
+	}
+
+	res = snprintf(cmd, sizeof(cmd), "STAKEY-REQUEST %s", argv[0]);
+	if (res < 0 || (size_t) res >= sizeof(cmd) - 1) {
+		printf("Too long STAKEY-REQUEST command.\n");
+		return 0;
+	}
+	return wpa_ctrl_command(ctrl, cmd);
+}
+
+
+static int wpa_cli_cmd_ap_scan(struct wpa_ctrl *ctrl, int argc, char *argv[])
+{
+	char cmd[256];
+	if (argc != 1) {
+		printf("Invalid AP_SCAN command: needs one argument (ap_scan "
+		       "value)\n");
+		return 0;
+	}
+	snprintf(cmd, sizeof(cmd), "AP_SCAN %s", argv[0]);
+	return wpa_ctrl_command(ctrl, cmd);
+}
+
+
+static int wpa_cli_cmd_stkstart(struct wpa_ctrl *ctrl, int argc,
+				char *argv[])
+{
+	char cmd[256];
+	int res;
+
+	if (argc != 1) {
+		printf("Invalid STKSTART command: needs one argument "
+		       "(Peer STA MAC address)\n");
+		return 0;
+	}
+
+	res = snprintf(cmd, sizeof(cmd), "STKSTART %s", argv[0]);
+	if (res < 0 || (size_t) res >= sizeof(cmd) - 1) {
+		printf("Too long STKSTART command.\n");
+		return 0;
+	}
+	return wpa_ctrl_command(ctrl, cmd);
+}
+
+
 static int wpa_cli_cmd_level(struct wpa_ctrl *ctrl, int argc, char *argv[])
 {
 	char cmd[256];
@@ -454,7 +516,7 @@ static int wpa_cli_cmd_pin(struct wpa_ctrl *ctrl, int argc, char *argv[])
 		       "(network id and pin)\n");
 		return 0;
 	}
-	
+
 	end = cmd + sizeof(cmd);
 	pos = cmd;
 	pos += snprintf(pos, end - pos, WPA_CTRL_RSP "PIN-%s:%s",
@@ -639,6 +701,7 @@ static int wpa_cli_cmd_set_network(struct wpa_ctrl *ctrl, int argc,
 				   char *argv[])
 {
 	char cmd[256];
+	int res;
 
 	if (argc == 0) {
 		wpa_cli_show_network_variables();
@@ -651,8 +714,9 @@ static int wpa_cli_cmd_set_network(struct wpa_ctrl *ctrl, int argc,
 		return 0;
 	}
 
-	if (snprintf(cmd, sizeof(cmd), "SET_NETWORK %s %s %s",
-		     argv[0], argv[1], argv[2]) >= sizeof(cmd) - 1) {
+	res = snprintf(cmd, sizeof(cmd), "SET_NETWORK %s %s %s",
+		       argv[0], argv[1], argv[2]);
+	if (res < 0 || (size_t) res >= sizeof(cmd) - 1) {
 		printf("Too long SET_NETWORK command.\n");
 		return 0;
 	}
@@ -664,6 +728,7 @@ static int wpa_cli_cmd_get_network(struct wpa_ctrl *ctrl, int argc,
 				   char *argv[])
 {
 	char cmd[256];
+	int res;
 
 	if (argc == 0) {
 		wpa_cli_show_network_variables();
@@ -676,8 +741,9 @@ static int wpa_cli_cmd_get_network(struct wpa_ctrl *ctrl, int argc,
 		return 0;
 	}
 
-	if (snprintf(cmd, sizeof(cmd), "GET_NETWORK %s %s",
-		     argv[0], argv[1]) >= sizeof(cmd) - 1) {
+	res = snprintf(cmd, sizeof(cmd), "GET_NETWORK %s %s",
+		       argv[0], argv[1]);
+	if (res < 0 || (size_t) res >= sizeof(cmd) - 1) {
 		printf("Too long GET_NETWORK command.\n");
 		return 0;
 	}
@@ -712,8 +778,25 @@ static int wpa_cli_cmd_scan_results(struct wpa_ctrl *ctrl, int argc,
 }
 
 
+static int wpa_cli_cmd_get_capability(struct wpa_ctrl *ctrl, int argc,
+				      char *argv[])
+{
+	char cmd[64];
+
+	if (argc != 1) {
+		printf("Invalid GET_CAPABILITY command: needs one argument\n");
+		return 0;
+	}
+
+	snprintf(cmd, sizeof(cmd), "GET_CAPABILITY %s", argv[0]);
+
+	return wpa_ctrl_command(ctrl, cmd);
+}
+
+
 static void wpa_cli_list_interfaces(struct wpa_ctrl *ctrl)
 {
+#ifndef CONFIG_CTRL_IFACE_UDP
 	struct dirent *dent;
 	DIR *dir;
 
@@ -732,6 +815,7 @@ static void wpa_cli_list_interfaces(struct wpa_ctrl *ctrl)
 		printf("%s\n", dent->d_name);
 	}
 	closedir(dir);
+#endif /* CONFIG_CTRL_IFACE_UDP */
 }
 
 
@@ -776,6 +860,48 @@ static int wpa_cli_cmd_terminate(struct wpa_ctrl *ctrl, int argc,
 }
 
 
+static int wpa_cli_cmd_interface_add(struct wpa_ctrl *ctrl, int argc,
+				     char *argv[])
+{
+	char cmd[256];
+
+	if (argc < 1) {
+		printf("Invalid INTERFACE_ADD command: needs at least one "
+		       "argument (interface name)\n"
+			"All arguments: ifname confname driver ctrl_interface "
+			"driver_param bridge_name\n");
+		return 0;
+	}
+
+	/*
+	 * INTERFACE_ADD <ifname>TAB<confname>TAB<driver>TAB<ctrl_interface>TAB
+	 * <driver_param>TAB<bridge_name>
+	 */
+	snprintf(cmd, sizeof(cmd), "INTERFACE_ADD %s\t%s\t%s\t%s\t%s\t%s",
+		 argv[0],
+		 argc > 1 ? argv[1] : "", argc > 2 ? argv[2] : "",
+		 argc > 3 ? argv[3] : "", argc > 4 ? argv[4] : "",
+		 argc > 5 ? argv[5] : "");
+	return wpa_ctrl_command(ctrl, cmd);
+}
+
+
+static int wpa_cli_cmd_interface_remove(struct wpa_ctrl *ctrl, int argc,
+					char *argv[])
+{
+	char cmd[128];
+
+	if (argc != 1) {
+		printf("Invalid INTERFACE_REMOVE command: needs one argument "
+		       "(interface name)\n");
+		return 0;
+	}
+
+	snprintf(cmd, sizeof(cmd), "INTERFACE_REMOVE %s", argv[0]);
+	return wpa_ctrl_command(ctrl, cmd);
+}
+
+
 struct wpa_cli_cmd {
 	const char *cmd;
 	int (*handler)(struct wpa_ctrl *ctrl, int argc, char *argv[]);
@@ -815,8 +941,14 @@ static struct wpa_cli_cmd wpa_cli_commands[] = {
 	{ "disconnect", wpa_cli_cmd_disconnect },
 	{ "scan", wpa_cli_cmd_scan },
 	{ "scan_results", wpa_cli_cmd_scan_results },
+	{ "get_capability", wpa_cli_cmd_get_capability },
 	{ "reconfigure", wpa_cli_cmd_reconfigure },
 	{ "terminate", wpa_cli_cmd_terminate },
+	{ "interface_add", wpa_cli_cmd_interface_add },
+	{ "interface_remove", wpa_cli_cmd_interface_remove },
+	{ "stakey-request", wpa_cli_cmd_stakey_request },
+	{ "ap_scan", wpa_cli_cmd_ap_scan },
+	{ "stkstart", wpa_cli_cmd_stkstart },
 	{ NULL, NULL }
 };
 
@@ -877,7 +1009,9 @@ static int wpa_cli_exec(const char *program, const char *arg1,
 	if (cmd == NULL)
 		return -1;
 	snprintf(cmd, len, "%s %s %s", program, arg1, arg2);
+#ifndef _WIN32_WCE
 	system(cmd);
+#endif /* _WIN32_WCE */
 	free(cmd);
 
 	return 0;
@@ -887,6 +1021,7 @@ static int wpa_cli_exec(const char *program, const char *arg1,
 static void wpa_cli_action_process(const char *msg)
 {
 	const char *pos;
+	char *copy = NULL, *id, *pos2;
 
 	pos = msg;
 	if (*pos == '<') {
@@ -899,9 +1034,46 @@ static void wpa_cli_action_process(const char *msg)
 	}
 
 	if (str_match(pos, WPA_EVENT_CONNECTED)) {
-		wpa_cli_exec(action_file, ctrl_ifname, "CONNECTED");
+		int new_id = -1;
+		os_unsetenv("WPA_ID");
+		os_unsetenv("WPA_ID_STR");
+		os_unsetenv("WPA_CTRL_DIR");
+
+		pos = strstr(pos, "[id=");
+		if (pos)
+			copy = strdup(pos + 4);
+
+		if (copy) {
+			pos2 = id = copy;
+			while (*pos2 && *pos2 != ' ')
+				pos2++;
+			*pos2++ = '\0';
+			new_id = atoi(id);
+			os_setenv("WPA_ID", id, 1);
+			while (*pos2 && *pos2 != '=')
+				pos2++;
+			if (*pos2 == '=')
+				pos2++;
+			id = pos2;
+			while (*pos2 && *pos2 != ']')
+				pos2++;
+			*pos2 = '\0';
+			os_setenv("WPA_ID_STR", id, 1);
+			free(copy);
+		}
+
+		os_setenv("WPA_CTRL_DIR", ctrl_iface_dir, 1);
+
+		if (!wpa_cli_connected || new_id != wpa_cli_last_id) {
+			wpa_cli_connected = 1;
+			wpa_cli_last_id = new_id;
+			wpa_cli_exec(action_file, ctrl_ifname, "CONNECTED");
+		}
 	} else if (str_match(pos, WPA_EVENT_DISCONNECTED)) {
-		wpa_cli_exec(action_file, ctrl_ifname, "DISCONNECTED");
+		if (wpa_cli_connected) {
+			wpa_cli_connected = 0;
+			wpa_cli_exec(action_file, ctrl_ifname, "DISCONNECTED");
+		}
 	} else if (str_match(pos, WPA_EVENT_TERMINATING)) {
 		printf("wpa_supplicant is terminating - stop monitoring\n");
 		wpa_cli_quit = 1;
@@ -909,10 +1081,12 @@ static void wpa_cli_action_process(const char *msg)
 }
 
 
+#ifndef CONFIG_ANSI_C_EXTRA
 static void wpa_cli_action_cb(char *msg, size_t len)
 {
 	wpa_cli_action_process(msg);
 }
+#endif /* CONFIG_ANSI_C_EXTRA */
 
 
 static void wpa_cli_recv_pending(struct wpa_ctrl *ctrl, int in_read,
@@ -1096,12 +1270,16 @@ static void wpa_cli_interactive(void)
 }
 
 
-static void wpa_cli_action(struct wpa_ctrl *ctrl, const char *action_file)
+static void wpa_cli_action(struct wpa_ctrl *ctrl)
 {
+#ifdef CONFIG_ANSI_C_EXTRA
+	/* TODO: ANSI C version(?) */
+	printf("Action processing not supported in ANSI C build.\n");
+#else /* CONFIG_ANSI_C_EXTRA */
 	fd_set rfds;
 	int fd, res;
 	struct timeval tv;
-	char buf[16];
+	char buf[256]; /* note: large enough to fit in unsolicited messages */
 	size_t len;
 
 	fd = wpa_ctrl_get_fd(ctrl);
@@ -1131,6 +1309,7 @@ static void wpa_cli_action(struct wpa_ctrl *ctrl, const char *action_file)
 			}
 		}
 	}
+#endif /* CONFIG_ANSI_C_EXTRA */
 }
 
 
@@ -1138,11 +1317,9 @@ static void wpa_cli_cleanup(void)
 {
 	wpa_cli_close_connection();
 	if (pid_file)
-		unlink(pid_file);
+		os_daemonize_terminate(pid_file);
 
-#ifdef CONFIG_NATIVE_WINDOWS
-	WSACleanup();
-#endif /* CONFIG_NATIVE_WINDOWS */
+	os_program_deinit();
 }
 
 static void wpa_cli_terminate(int sig)
@@ -1186,18 +1363,13 @@ int main(int argc, char *argv[])
 	int warning_displayed = 0;
 	int c;
 	int daemonize = 0;
-	FILE *f;
+	const char *global = NULL;
 
-#ifdef CONFIG_NATIVE_WINDOWS
-	WSADATA wsaData;
-	if (WSAStartup(MAKEWORD(2, 0), &wsaData)) {
-		printf("Could not find a usable WinSock.dll\n");
+	if (os_program_init())
 		return -1;
-	}
-#endif /* CONFIG_NATIVE_WINDOWS */
 
 	for (;;) {
-		c = getopt(argc, argv, "a:Bhi:p:P:v");
+		c = getopt(argc, argv, "a:Bg:hi:p:P:v");
 		if (c < 0)
 			break;
 		switch (c) {
@@ -1207,6 +1379,9 @@ int main(int argc, char *argv[])
 		case 'B':
 			daemonize = 1;
 			break;
+		case 'g':
+			global = optarg;
+			break;
 		case 'h':
 			usage();
 			return 0;
@@ -1233,12 +1408,31 @@ int main(int argc, char *argv[])
 	if (interactive)
 		printf("%s\n\n%s\n\n", wpa_cli_version, wpa_cli_license);
 
-	for (;;) {
+	if (global) {
+		ctrl_conn = wpa_ctrl_open(global);
+		if (ctrl_conn == NULL) {
+			perror("Failed to connect to wpa_supplicant - "
+			       "wpa_ctrl_open");
+			return -1;
+		}
+	}
+
+	for (; !global;) {
+#ifndef CONFIG_CTRL_IFACE_UDP
 		if (ctrl_ifname == NULL) {
 			struct dirent *dent;
 			DIR *dir = opendir(ctrl_iface_dir);
 			if (dir) {
 				while ((dent = readdir(dir))) {
+#ifdef _DIRENT_HAVE_D_TYPE
+					/* Skip the file if it is not a socket.
+					 * Also accept DT_UNKNOWN (0) in case
+					 * the C library or underlying file
+					 * system does not support d_type. */
+					if (dent->d_type != DT_SOCK &&
+					    dent->d_type != DT_UNKNOWN)
+						continue;
+#endif /* _DIRENT_HAVE_D_TYPE */
 					if (strcmp(dent->d_name, ".") == 0 ||
 					    strcmp(dent->d_name, "..") == 0)
 						continue;
@@ -1250,6 +1444,7 @@ int main(int argc, char *argv[])
 				closedir(dir);
 			}
 		}
+#endif /* CONFIG_CTRL_IFACE_UDP */
 		ctrl_conn = wpa_cli_open_connection(ctrl_ifname);
 		if (ctrl_conn) {
 			if (warning_displayed)
@@ -1268,12 +1463,14 @@ int main(int argc, char *argv[])
 			       "re-trying\n");
 			warning_displayed = 1;
 		}
-		sleep(1);
+		os_sleep(1, 0);
 		continue;
 	}
 
+#ifndef _WIN32_WCE
 	signal(SIGINT, wpa_cli_terminate);
 	signal(SIGTERM, wpa_cli_terminate);
+#endif /* _WIN32_WCE */
 #ifndef CONFIG_NATIVE_WINDOWS
 	signal(SIGALRM, wpa_cli_alarm);
 #endif /* CONFIG_NATIVE_WINDOWS */
@@ -1289,23 +1486,13 @@ int main(int argc, char *argv[])
 		}
 	}
 
-	if (daemonize && daemon(0, 0)) {
-		perror("daemon");
+	if (daemonize && os_daemonize(pid_file))
 		return -1;
-	}
-
-	if (pid_file) {
-		f = fopen(pid_file, "w");
-		if (f) {
-			fprintf(f, "%u\n", getpid());
-			fclose(f);
-		}
-	}
 
 	if (interactive)
 		wpa_cli_interactive();
 	else if (action_file)
-		wpa_cli_action(ctrl_conn, action_file);
+		wpa_cli_action(ctrl_conn);
 	else
 		wpa_request(ctrl_conn, argc - optind, &argv[optind]);
 
@@ -1314,3 +1501,11 @@ int main(int argc, char *argv[])
 
 	return 0;
 }
+
+#else /* CONFIG_CTRL_IFACE */
+int main(int argc, char *argv[])
+{
+	printf("CONFIG_CTRL_IFACE not defined - wpa_cli disabled\n");
+	return -1;
+}
+#endif /* CONFIG_CTRL_IFACE */
